use std::sync::{Arc};
use std::sync::atomic::{AtomicU64, Ordering};
use std::cell::{RefMut};
use std::ffi::CStr;
use log::*;

use ecvector::{WriteHandle, ReadHandle};
use crate::modules::Module;
use crate::components::{Component, CarDirection};
use crate::ecs::ShareableRefCell;

use shared_memory_interface::constants::*;


// Generic utils
// Logging
pub fn log_agent_run(timestamp: u64, agent: &str) {
    debug!("[{}]: Run at ({})", agent, timestamp);
}

// Unpack timestamp
pub fn timestamp_value(timestamp: Arc<AtomicU64>) -> u64 {
    return (*timestamp).load(Ordering::SeqCst);
}

pub fn unpack_module<'a, T>(module: &'a Box<dyn Module>) -> Option<&'a T>
    where T: 'static
{
    return module.as_any().downcast_ref::<T>();
}

// Handles handling
pub fn unpack_write_handle(write_handle: &Arc<ShareableRefCell<WriteHandle<u64, Component>>>)
                           -> RefMut<WriteHandle<u64, Component>> {
    return (*write_handle).borrow_mut();
}

pub fn unpack_write_handle_as_read_handle(write_handle: &Arc<ShareableRefCell<WriteHandle<u64, Component>>>)
                           -> ReadHandle<u64, Component> {
    return (*write_handle).borrow().clone();
}

pub fn unpack_read_handle(read_handles: &Vec<ReadHandle<u64, Component>>, index: usize)
                          -> ReadHandle<u64, Component> {
    return read_handles[index].clone();
}

// Construct string from raw bytes
pub fn unpack_raw_string(bytes: &[u8]) -> String {
    // Strip trailing zeros
    let mut data = vec![];
    for byte in bytes {
        if *byte != 0 {
            data.push(*byte);
        }
    }
    // Add null terminator!
    data.push(0);

    // Build string
    let str = CStr::from_bytes_with_nul(&data).unwrap().to_str().unwrap();
    return String::from(str);
}

// Construct static str safe for storing in components
pub fn string_to_static_str(s: String) -> &'static str {
    Box::leak(s.into_boxed_str())
}

// M/s to km/h
pub fn ms2kmh(value: f32) -> f32 {
    return value * 3.6;
}

// Km/h to m/s
pub fn kmh2ms(value: f32) -> f32 {
    return value / 3.6;
}

// Get turn signal state
// -1 means blinking left, 1 means blinking right and 0 means no blinking
pub fn decode_turn_signal_state(bits_from_car1: u32) -> i32 {
    let left_on = bits_from_car1 & BFC1_BLINKING_LEFT;
    let right_on = bits_from_car1 & BFC1_BLINKING_RIGHT;

    // Check the status
    if left_on != 0 {
        return -1;
    } else if right_on != 0 {
        return 1;
    } else {
        return 0;
    }
}

// Decodes bitmask into real ego direction enum
pub fn decode_ego_direction(direction: i32) -> CarDirection
{
    if direction == 1 {
        return CarDirection::Right;
    } else if direction == 2 {
        return CarDirection::Straight;
    } else if direction == 4 {
        return CarDirection::Left;
    } else {
        return CarDirection::Unknown;
    }
}
