use shared_memory_interface::shared_memory::SharedMemory;
use knowledge_graph::graph::{KnowledgeGraph};

use ordered_float::OrderedFloat;
use enum_variant_type::EnumVariantType;

// Helper types
type Of32 = OrderedFloat<f32>;


// Special component for knowledge graph
#[derive(Clone, Copy, PartialEq, Eq, Debug)]
pub struct GraphData {
    graph_ptr: *mut SharedMemory<KnowledgeGraph>
}
impl Default for GraphData {
    fn default() -> Self {
        GraphData {
            graph_ptr: std::ptr::null_mut()
        }
    }
}
impl GraphData {
    pub fn new(graph_ptr: *mut SharedMemory<KnowledgeGraph>) -> Self {
        GraphData { graph_ptr }
    }

    pub fn get_graph_ref(&self) -> &KnowledgeGraph {
        unsafe { &mut *self.graph_ptr }.get_data_ref()
    }

    pub fn get_graph_mut_ref(&mut self) -> &mut KnowledgeGraph {
        unsafe { &mut *self.graph_ptr }.get_data_mut_ref()
    }
}

// Components for visual shared memory data
#[derive(Clone, Copy, PartialEq, Eq, Default, Debug)]
pub struct EgoData {
    pub position_x: Of32,
    pub position_y: Of32,
    pub position_z: Of32,

    pub direction_x: Of32,
    pub direction_y: Of32,
    pub direction_z: Of32,
    pub direction_w: Of32,

    pub speed: Of32,
    pub speed_vector_x: Of32,
    pub speed_vector_y: Of32,
    pub speed_vector_z: Of32,

    pub steering_wheel: Of32,
    pub gear: i32,
    pub brake: Of32,
    pub clutch: Of32,
    pub throttle: Of32,

    pub rpm: Of32,
    pub torque: Of32,

    pub bits_from_car1: u32,
}

#[derive(Clone, Copy, PartialEq, Eq, Default, Debug)]
pub struct GazeData {
    pub data: u32,
}


// AGENT COMPONENTS
// Component list
#[derive(Clone, Copy, PartialEq, Eq, Debug, EnumVariantType)]
pub enum Component {
    // Implicit components i.e. data sources
    Graph(GraphData),
    EgoShared(EgoData),
    GazeShared(GazeData),
    // Custom components
    IncorrectGear(IncorrectGearData),
    IncorrectGearFeedback(IncorrectGearFeedbackData),
    Speeding (SpeedingData),
    SpeedingFeedback (SpeedingFeedbackData),
    LaneChange (LaneChangeData),
    TurnSignal (TurnSignalData),
    Overtake (OvertakeData),
    OvertakeFeedback (OvertakeFeedbackData),
    CarYielding(CarYieldingData),
    CarYieldingFeedback(CarYieldingFeedbackData),
    PedestrianYielding(PedestrianYieldingData),
    PedestrianYieldingFeedback(PedestrianYieldingFeedbackData),
}


// INCORRECT GEAR AGENT
#[derive(Clone, Copy, PartialEq, Eq, Debug)]
pub enum GearState {
    Ok,
    TooLow,
    TooHigh,
}
impl Default for GearState {
    fn default() -> Self {
        GearState::Ok
    }
}

#[derive(Clone, Copy, PartialEq, Eq, Default, Debug)]
pub struct IncorrectGearData {
    pub state: GearState,
}

#[derive(Clone, Copy, PartialEq, Eq, Default, Debug)]
pub struct IncorrectGearFeedbackData {
    pub value: &'static str,
}

// SPEEDING AGENT
#[derive(Clone, Copy, PartialEq, Eq, Debug)]
pub enum SpeedingState {
    Ok,
    SlightlyAboveLimit,
    ConsiderablyAboveLimit,
    SevereSpeeding,
}
impl Default for SpeedingState {
    fn default() -> Self {
        SpeedingState::Ok
    }
}

#[derive(Clone, Copy, PartialEq, Eq, Default, Debug)]
pub struct SpeedingData {
    pub speed_limit: usize,
    pub ego_speed: Of32,
    pub state: SpeedingState,
}

#[derive(Clone, Copy, PartialEq, Eq, Default, Debug)]
pub struct SpeedingFeedbackData {
    pub value: &'static str,
}

// LANE CHANGE AGENT
#[derive(Clone, Copy, PartialEq, Eq, Default, Debug)]
pub struct LaneChangeData {
    pub from_lane_id: usize,
    pub from_lane_name: &'static str,
    pub to_lane_id: usize,
    pub to_lane_name: &'static str,
}

// TURN SIGNAL AGENT
#[derive(Clone, Copy, PartialEq, Eq, Debug)]
pub enum TurnSignaling {
    None,
    ToLeft,
    ToRight
}
impl Default for TurnSignaling {
    fn default() -> Self {
        TurnSignaling::None
    }
}

#[derive(Clone, Copy, PartialEq, Eq, Default, Debug)]
pub struct TurnSignalData {
    pub turn_signal: TurnSignaling,
}

// OVERTAKE  AGENT
#[derive(Clone, Copy, PartialEq, Eq, Debug)]
pub enum OvertakeTurnSignal {
    None,
    FirstToLeft,
    SecondToRight,
    Both
}
impl Default for OvertakeTurnSignal {
    fn default() -> Self {
        OvertakeTurnSignal::None
    }
}
#[derive(Clone, Copy, PartialEq, Eq, Default, Debug)]
pub struct OvertakeData {
    pub overtake_timestamp: u64,
    pub turn_signals: OvertakeTurnSignal,
    pub duration: usize,
}

#[derive(Clone, Copy, PartialEq, Eq, Default, Debug)]
pub struct OvertakeFeedbackData {
    pub value: &'static str,
}

// CAR YIELDING  AGENT
#[derive(Clone, Copy, PartialEq, Eq, Debug)]
pub enum CarYieldingType {
    None,
    Uncontrolled,
    TrafficLights,
}
impl Default for CarYieldingType {
    fn default() -> Self {
        CarYieldingType::None
    }
}

#[derive(Clone, Copy, PartialEq, Eq, Debug)]
pub enum CarDirection {
    Unknown,
    Straight,
    Left,
    Right,
}
impl Default for CarDirection {
    fn default() -> Self {
        CarDirection::Unknown
    }
}

#[derive(Clone, Copy, PartialEq, Eq, Default, Debug)]
pub struct CarYieldingData {
    pub yielding_timestamp: u64,
    pub yielding_type: CarYieldingType,
}

#[derive(Clone, Copy, PartialEq, Eq, Default, Debug)]
pub struct CarYieldingFeedbackData {
    pub value: &'static str,
}

// PEDESTRIAN YIELDING AGENT
#[derive(Clone, Copy, PartialEq, Eq, Debug)]
pub enum PedestrianYieldingType {
    None,
    Crosswalk,
}
impl Default for PedestrianYieldingType {
    fn default() -> Self {
        PedestrianYieldingType::None
    }
}

#[derive(Clone, Copy, PartialEq, Eq, Default, Debug)]
pub struct PedestrianYieldingData {
    pub yielding_timestamp: u64,
    pub yielding_type: PedestrianYieldingType,
}

#[derive(Clone, Copy, PartialEq, Eq, Default, Debug)]
pub struct PedestrianYieldingFeedbackData {
    pub value: &'static str,
}