use std::sync::{Arc};
use std::cell::{RefCell, Ref, RefMut};
use std::sync::atomic::{AtomicU64};

use super::modules::Module;

use ecvector;
use ecvector::{ReadHandle, WriteHandle};


// In fact a dummy lock for lock free access
pub struct ShareableRefCell<T> {
    cell: RefCell<T>
}
impl<T> ShareableRefCell<T> {
    pub fn new(inner: T) -> Self {
        ShareableRefCell { cell: RefCell::new(inner) }
    }
    pub fn borrow(&self) -> Ref<T> {
        self.cell.borrow()
    }
    pub fn borrow_mut(&self) -> RefMut<T> {
        self.cell.borrow_mut()
    }
    pub fn get_inner(&self) -> &RefCell<T> {
        &self.cell
    }
}
unsafe impl<T> Sync for ShareableRefCell<T> {}
unsafe impl<T> Send for ShareableRefCell<T> {}


// Components
pub struct ComponentStorage<T, C>
    where
        T: Eq + Clone,
        C: Clone,
{
    pub read_handle: ReadHandle<T, C>,
    pub write_handle: Arc<ShareableRefCell<WriteHandle<T, C>>>,
}
impl<T, C> ComponentStorage<T, C>
    where
        T: Eq + Clone,
        C: Clone,
{
    pub fn new() -> Self {
        let vector = ecvector::new::<T, C>();
        ComponentStorage {
            read_handle: vector.0,
            write_handle: Arc::new(ShareableRefCell::new(vector.1)),
        }
    }

    pub fn handles(&self) -> (ReadHandle<T, C>, Arc<ShareableRefCell<WriteHandle<T, C>>>) {
        (self.read_handle.clone(), self.write_handle.clone())
    }
}

// Entity
#[derive(Clone)]
pub enum AgentType {
    Periodic (u64),
    Observing
}

#[derive(Clone)]
pub struct AgentEntity<T, C>
    where
        T: Eq + Clone,
        C: Clone,
{
    pub agent_type: AgentType,
    pub source_components: Option<Vec<usize>>,
    pub sink_component: Option<usize>,
    pub system_method: Arc<dyn Fn(Arc<AtomicU64>,
        Vec<ReadHandle<T, C>>,
        Arc<ShareableRefCell<WriteHandle<T, C>>>,
        Vec<(&'static str, Box<dyn Module>)>) -> bool + Send + Sync>
}
impl<T, C> AgentEntity<T, C>
    where
        T: Eq + Clone,
        C: Clone,
{
    pub fn new(agent_type: AgentType) -> Self {
        AgentEntity {
            agent_type,
            source_components: None,
            sink_component: None,
            system_method: Arc::new(|_timestamp, _reads, _write, _modules| { false })
        }
    }

    pub fn set_source_components(mut self, source: Vec<usize>) -> Self {
        self.source_components = Some(source);
        self
    }

    pub fn set_sink_component(mut self, sink: usize) -> Self {
        self.sink_component = Some(sink);
        self
    }

    pub fn set_system_method(mut self, method: Arc<dyn Fn(Arc<AtomicU64>,
        Vec<ReadHandle<T, C>>,
        Arc<ShareableRefCell<WriteHandle<T, C>>>,
        Vec<(&'static str, Box<dyn Module>)>) -> bool + Send + Sync>) -> Self {
        self.system_method = method;
        self
    }
}
