// Platform building blocks
pub mod platform;
use platform::*;

pub mod ecs;
use ecs::*;

pub mod components;
use components::*;

pub mod modules;
use modules::*;

pub mod agent_utils;
use agent_utils::*;

pub mod types;
use types::*;

// Imports
use knowledge_graph::graph::types::*;

use interval_algebra::*;
use pattern_query_engine::*;

use log::*;

use std::sync::Arc;
use std::collections::HashMap;

use rand::seq::SliceRandom;
use ordered_float::OrderedFloat;
use rsfuzzy::*;

use config::*;


// Big bang for the whole system
pub fn build_and_run() {
    info!("[Agent platform]: Agent platform build started!");

    let mut platform = AgentPlatform::new();
    platform.initialize();

    // Register modules
    platform.register_module("text2speech", Box::new(SpeechFeedbackModule::new())); // Index 0
    platform.register_module("fuzzy_logic", Box::new(FuzzyLogicModule::new())); // Index 1

    // Register components
    // Implicit components: Graph -> Index 0, EgoData -> Index 1, GazeData -> Index 2
    platform.register_component("incorrect_gear",  // Index 3
                                Component::IncorrectGear (IncorrectGearData::default()));
    platform.register_component("incorrect_gear_feedback",  // Index 4
                                Component::IncorrectGearFeedback (IncorrectGearFeedbackData::default()));
    platform.register_component("speeding",  // Index 5
                                Component::Speeding (SpeedingData::default()));
    platform.register_component("speeding_feedback",  // Index 6
                                Component::SpeedingFeedback (SpeedingFeedbackData::default()));
    platform.register_component("lane_change",  // Index 7
                                Component::LaneChange (LaneChangeData::default()));
    platform.register_component("turn_signal",  // Index 8
                                Component::TurnSignal (TurnSignalData::default()));
    platform.register_component("overtake",  // Index 9
                                Component::Overtake (OvertakeData::default()));
    platform.register_component("overtake_feedback",  // Index 10
                                Component::OvertakeFeedback (OvertakeFeedbackData::default()));
    platform.register_component("car_yielding",  // Index 11
                                Component::CarYielding(CarYieldingData::default()));
    platform.register_component("car_yielding_feedback",  // Index 12
                                Component::CarYieldingFeedback(CarYieldingFeedbackData::default()));
    platform.register_component("pedestrian_yielding",  // Index 13
                                Component::PedestrianYielding(PedestrianYieldingData::default()));
    platform.register_component("pedestrian_yielding_feedback",  // Index 14
                                Component::PedestrianYieldingFeedback(PedestrianYieldingFeedbackData::default()));


    // INCORRECT GEAR AGENT
    let incorrect_gear_system = |timestamp: GlobalTimestampMs,
                                 read_handles: ComponentReadHandles,
                                 write_handle: ComponentWriteHandle,
                                 _modules: Modules | -> bool {

        // Extract timestamp
        let timestamp = timestamp_value(timestamp);
        log_agent_run(timestamp, "IncorrectGearAgent");

        // Get handles and data
        let read_ego = unpack_read_handle(&read_handles, 0);
        let mut write_incorrect_gear = unpack_write_handle(&write_handle);

        let mut ego = Default::default();
        if let Component::EgoShared(data) = read_ego.last().unwrap().1 {
            ego = data;
        }

        // Run the logic
        fn judge_incorrect_gear(gear: i32, clutch: f32, speed: f32) -> bool
        {
            // We dont care about gears if neutral or clutch pressed
            if gear == 0 || clutch > 0.1 {
                return false;
            }

            // Turn on when you hit the speed
            if ms2kmh(speed) > GEAR_AGENT_ENABLE_SPEED_THRESHOLD {
                return true;
            } else {
                return false;
            }
        }

        let mut gear_state = GearState::Ok;
        let gear = ego.gear;
        let throttle = ego.throttle.into_inner();
        let clutch = ego.clutch.into_inner();
        let speed = ego.speed.into_inner();
        let rpm = ego.rpm.into_inner();

        if judge_incorrect_gear(gear, clutch, speed)
        {
            // Determine normal driving or hard acceleration
            if throttle < GEAR_AGENT_THROTTLE_THRESHOLD_ACCELERATION_MODE {
                if rpm > GEAR_AGENT_NORMAL_RPM_HIGH_BOUND && gear < GEAR_AGENT_MAX_GEAR {
                    gear_state = GearState::TooLow;
                } else if rpm < GEAR_AGENT_NORMAL_RPM_LOW_BOUND && gear > GEAR_AGENT_MIN_GEAR {
                    gear_state = GearState::TooHigh;
                }
            } else {
                if rpm > GEAR_AGENT_ACCELERATION_RPM_HIGH_BOUND && gear < GEAR_AGENT_MAX_GEAR {
                    gear_state = GearState::TooLow;
                } else if rpm < GEAR_AGENT_ACCELERATION_RPM_LOW_BOUND && gear > GEAR_AGENT_MIN_GEAR {
                    gear_state = GearState::TooHigh;
                }
            }
        }

        // Write result to the component
        if gear_state != GearState::Ok {
            let result = Component::IncorrectGear(IncorrectGearData { state: gear_state });
            write_incorrect_gear.push(timestamp, result);
            write_incorrect_gear.refresh();

            warn!("[IncorrectGearAgent]: Gear: {:?}", gear_state);
            return true;
        }

        return false;
    };

    let incorrect_gear_agent = AgentEntity::new(AgentType::Periodic(1000))
        .set_source_components(vec![1])
        .set_sink_component(3)
        .set_system_method(Arc::new(incorrect_gear_system));


    // INCORRECT GEAR FEEDBACK AGENT
    let incorrect_gear_feedback_system = |timestamp: GlobalTimestampMs,
                                          read_handles: ComponentReadHandles,
                                          write_handle: ComponentWriteHandle,
                                          modules: Modules | -> bool {

        // Extract timestamp
        let timestamp = timestamp_value(timestamp);
        log_agent_run(timestamp, "IncorrectGearFeedbackAgent");

        // Get text to speech module
        let text2speech = unpack_module::<SpeechFeedbackModule>(&modules[0].1).unwrap();

        // Get handles and data
        let read_incorrect_gear_state = unpack_read_handle(&read_handles, 0);
        let mut write_incorrect_gear_feedback = unpack_write_handle(&write_handle);

        // Check if we have any data, if not return
        if read_incorrect_gear_state.is_empty() {
            return false;
        }


        // Run the logic
        let target_length: usize = GEAR_AGENT_QUEUE_LENGTH_FOR_FEEDBACK;
        let last_feedback = write_incorrect_gear_feedback.clone().last(); // Cloning the handle gives read handle

        // If there is no feedback given yet assume zero last timestamp
        let last_feedback_timestamp;
        if last_feedback.is_some() {
            last_feedback_timestamp = last_feedback.unwrap().0;
        }
        else {
            last_feedback_timestamp = 0;
        }

        // If we encounter the same incorrect gear marker x times, send feedback
        let vec: Vec<_> = read_incorrect_gear_state.rev_take_into(|k, v| {
            (k.clone(), v.clone())
        }, target_length);

        let sum: i32 = vec.iter().fold(0, |mut accumulator, x| {
            let timestamp = x.0;
            let component = x.1;

            // Check if all records are newer then the ones we already seen
            if timestamp > last_feedback_timestamp {
                // Decode the component data
                let mut gear_state = Default::default();
                if let Component::IncorrectGear(data) = component {
                    gear_state = data.state;
                }

                if gear_state == GearState::TooLow {
                    accumulator += 1
                } else if gear_state == GearState::TooHigh {
                    accumulator -= 1
                }
            }
            return accumulator;
        });

        // Read the incorrect gear state for feedback
        let mut gear_state = Default::default();
        if let Component::IncorrectGear (data) = read_incorrect_gear_state.last().unwrap().1 {
            gear_state = data.state;
        }

        // Give feedback and write result to the component
        if sum.abs() as usize == target_length {
            // Randomly pick the feedback text
            let mut text = "";

            // All the options for gear states
            if gear_state == GearState::TooLow {
                let low: Vec<_> = GEAR_AGENT_TOO_LOW_GEAR_TEXT
                    .choose_multiple(&mut rand::thread_rng(), 1)
                    .collect();
                text = low.first().unwrap();
            } else if gear_state == GearState::TooHigh {
                let high: Vec<_> = GEAR_AGENT_TOO_HIGH_GEAR_TEXT
                    .choose_multiple(&mut rand::thread_rng(), 1)
                    .collect();
                text = high.first().unwrap();
            }

            // Log the feedback
            warn!("[IncorrectGearFeedbackAgent]: Feedback: {:?}", text);

            // Save given feedback
            let result = Component::IncorrectGearFeedback(IncorrectGearFeedbackData { value: text });
            write_incorrect_gear_feedback.push(timestamp, result);
            write_incorrect_gear_feedback.refresh();

            // Finally enqueue
            text2speech.enqueue_for_read(text);
            return true;
        }

        return false;
    };

    let incorrect_gear_feedback_agent = AgentEntity::new(AgentType::Observing)
        .set_source_components(vec![3])
        .set_sink_component(4)
        .set_system_method(Arc::new(incorrect_gear_feedback_system));


    // SPEEDING AGENT
    let speeding_system = |timestamp: GlobalTimestampMs,
                              read_handles: ComponentReadHandles,
                              write_handle: ComponentWriteHandle,
                              modules: Modules | -> bool {

        // Extract timestamp
        let timestamp = timestamp_value(timestamp);
        log_agent_run(timestamp, "SpeedingAgent");

        // Get handles
        let read_graph = unpack_read_handle(&read_handles, 0);
        let read_ego = unpack_read_handle(&read_handles, 1);
        let mut write_speeding = unpack_write_handle(&write_handle);


        // Read current speed limit from the knowledge graph
        let mut current_lane_speed_limit = 0;
        if let Component::Graph(data) = read_graph.first().unwrap().1 {
            let graph = data.get_graph_ref();

            // Search for current lane
            // Type indices
            let ego_type_index = KnowledgeGraphNode::Ego(Ego::default()).type_index();
            let lane_type_index = KnowledgeGraphNode::Lane(Lane::default()).type_index();
            let speed_limit_type_index = KnowledgeGraphNode::SpeedLimitSign(SpeedLimitSign::default()).type_index();
            let is_in_type_index = KnowledgeGraphRelation::isIn(isIn::default()).type_index();
            let belongs_to_type_index = KnowledgeGraphRelation::belongsTo(belongsTo::default()).type_index();

            // Run the lane pattern query
            let mut lane_pattern = Pattern::new();
            lane_pattern.node(Node::new(ego_type_index))
                .with_outgoing_relation(Relation::new(is_in_type_index))
                .to_node(Node::new(lane_type_index));

            // Search during the last second
            if lane_pattern.find(Interval::eternity(), &graph) {
                let result_data = lane_pattern.get_result();

                if let KnowledgeGraphNode::Lane(lane) = result_data[0]
                    .outgoing_relations[0]
                    .to_node[0]
                    .node
                {
                    current_lane_speed_limit = lane.laneSpeedLimit;
                    debug!("[SpeedingAgent]: Lane found {:?}", lane);
                }
            } else {
                // Check if it makes sense to assess speed i.e. if lane was found
                debug!("[SpeedingAgent]: No lane found!");
                return false;
            }

            // Also try to search for active speed limit signs, they might override the default lane limit
            let mut speed_limit_sign_pattern = Pattern::new();
            speed_limit_sign_pattern.node(Node::new(speed_limit_type_index))
                .with_outgoing_relation(Relation::new(belongs_to_type_index))
                .to_node(Node::new(lane_type_index));

            if speed_limit_sign_pattern.find(Interval::eternity(), &graph) {
                let result_data = speed_limit_sign_pattern.get_result();

                if let KnowledgeGraphNode::SpeedLimitSign(sign) = result_data[0].node {
                    current_lane_speed_limit = sign.signSpeedLimit;
                    debug!("[SpeedingAgent]: Speed limit sign found {:?}", sign);
                }
            }

        }

        // Check if it makes sense to assess speed, lane might have zero speed limit
        if current_lane_speed_limit == 0 {
            error!("[SpeedingAgent]: Lane has no speed limit!");
        }

        // Read current acceleration from ego data
        // Do it by comparing speed between now and the last second
        let ego_data_vec: Vec<_> = read_ego.rev_take_into(|k, v| {
            (k.clone(), v.clone())
        }, SPEEDING_AGENT_ACCELERATION_WINDOW_LENGTH / 10);

        let mut ego_speed_now= 0.0;
        if let Component::EgoShared(data) = ego_data_vec.first().unwrap().1 {
            ego_speed_now = data.speed.into_inner();
        }
        let mut ego_speed_before= 0.0;
        if let Component::EgoShared(data) = ego_data_vec.last().unwrap().1 {
            ego_speed_before = data.speed.into_inner();
        }

        // Check if it makes sense to assess speed i.e. if ego speed has reasonable value
        if ms2kmh(ego_speed_now) < SPEEDING_AGENT_ENABLE_SPEED_THRESHOLD {
            debug!("[SpeedingAgent]: Too low ego speed!");
            return false;
        }

        // Extract the needed values and clamp them to a known range
        let ego_over_speed_limit = num::clamp(ms2kmh(ego_speed_now) as i32 - current_lane_speed_limit, 0, 100);
        // Its necessary to shift the real acceleration rate by +10 (to avoid negative numbers)
        let ego_acceleration = num::clamp((ego_speed_now - ego_speed_before) + 10.0, 0.0, 20.0);


        // We have all the data we need, lets do the speeding assessment
        // Get the fuzzy logic module
        let fuzzy_module = unpack_module::<FuzzyLogicModule>(&modules[1].1).unwrap();
        let mut fuzzy_engine = fuzzy_module.engine.clone();

        // Setup the fuzzy input variables
        fuzzy_engine.reset();
        let over_speed_limit = fz_input_var![
            ("down", "ok", vec![0.0, 20.0]),
            ("triangle", "slightly", vec![0.0, 15.0, 20.0]),
            ("triangle", "considerably", vec![5.0, 20.0, 35.0]),
            ("up", "severely", vec![30.0, 100.0])
        ];
        fuzzy_engine.add_input_variable("over_speed_limit", over_speed_limit, 0, 100);

        // Its necessary to shift the real acceleration rate by +10 (to avoid negative numbers)
        let acceleration = fz_input_var![
            ("down", "decelerating_a_lot", vec![0.0, 8.0]),
            ("triangle", "decelerating", vec![6.0, 8.0, 10.0]),
            ("triangle", "constant", vec![8.0, 10.0, 12.0]),
            ("triangle", "accelerating", vec![10.0, 12.0, 14.0]),
            ("up", "accelerating_a_lot", vec![12.0, 20.0])
        ];
        fuzzy_engine.add_input_variable("acceleration", acceleration, 0, 20);

        // Setup the fuzzy output variables
        let speeding_state = fz_output_var![
            ("down", "ok", vec![0.0, 5.0]),
            ("triangle", "slightly_above", vec![5.0, 10.0, 15.0]),
            ("triangle", "considerably_above", vec![15.0, 20.0, 25.0]),
            ("up", "severe_speeding", vec![25.0, 100.0])
        ];
        fuzzy_engine.add_output_variable("speeding_state", speeding_state, 0, 100);

        // Setup speeding rules
        let rules = vec![
            // Slight speeding while decelerating is fine
            ("if over_speed_limit is ok and acceleration is decelerating then speeding_state is ok"),
            ("if over_speed_limit is slightly and acceleration is decelerating then speeding_state is ok"),
            ("if over_speed_limit is considerably and acceleration is decelerating then speeding_state is ok"),
            ("if over_speed_limit is ok and acceleration is decelerating_a_lot then speeding_state is ok"),
            ("if over_speed_limit is slightly and acceleration is decelerating_a_lot then speeding_state is ok"),
            ("if over_speed_limit is considerably and acceleration is decelerating_a_lot then speeding_state is ok"),
            // Slight speeding while still accelerating is a real problem
            ("if over_speed_limit is ok and acceleration is accelerating then speeding_state is ok"),
            ("if over_speed_limit is slightly and acceleration is accelerating then speeding_state is considerably_above"),
            ("if over_speed_limit is considerably and acceleration is accelerating then speeding_state is severe_speeding"),
            ("if over_speed_limit is severely and acceleration is accelerating then speeding_state is severe_speeding"),
            // Standard speeding situations
            ("if over_speed_limit is ok and acceleration is constant then speeding_state is ok"),
            ("if over_speed_limit is slightly and acceleration is constant then speeding_state is slightly_above"),
            ("if over_speed_limit is considerably and acceleration is constant then speeding_state is considerably_above"),
            ("if over_speed_limit is severely and acceleration is constant then speeding_state is severe_speeding"),
        ];
        fuzzy_engine.add_rules(rules);

        // Defuzzification method
        fuzzy_engine.add_defuzzification_method("mom");

        // Calculate the result
        let inputs = fz_set_inputs![
            ("over_speed_limit", ego_over_speed_limit as f32),
            ("acceleration", ego_acceleration as f32)
        ];
        let result = fuzzy_engine.calculate(inputs);

        // Determine the result
        for item in vec![
            ((5.0..15.0), SpeedingState::SlightlyAboveLimit),
            ((15.0..25.0), SpeedingState::ConsiderablyAboveLimit),
            ((25.0..100.0), SpeedingState::SevereSpeeding)]
        {
            if item.0.contains(&result) {
                // Write the result
                let data = Component::Speeding(SpeedingData {
                    speed_limit: current_lane_speed_limit as usize,
                    ego_speed: OrderedFloat(ms2kmh(ego_speed_now)),
                    state: item.1
                });

                write_speeding.push(timestamp, data);
                write_speeding.refresh();

                warn!("[SpeedingAgent]: Speeding {:?}", item.1);
                return true;
            }
        }

        return false;
    };

    let speeding_agent = AgentEntity::new(AgentType::Periodic(1000))
        .set_source_components(vec![0, 1])
        .set_sink_component(5)
        .set_system_method(Arc::new(speeding_system));


    // SPEEDING FEEDBACK AGENT
    let speeding_feedback_system = |timestamp: GlobalTimestampMs,
                           read_handles: ComponentReadHandles,
                           write_handle: ComponentWriteHandle,
                           modules: Modules | -> bool {

        // Extract timestamp
        let timestamp = timestamp_value(timestamp);
        log_agent_run(timestamp, "SpeedingFeedbackAgent");

        // Get text to speech module
        let text2speech = unpack_module::<SpeechFeedbackModule>(&modules[0].1).unwrap();

        // Get handles and data
        let read_speeding = unpack_read_handle(&read_handles, 0);
        let mut write_speeding_feedback = unpack_write_handle(&write_handle);

        // Check if we have any feedback data, if not return
        if read_speeding.is_empty() {
            return false;
        }


        // Run the logic
        let target_length: usize = SPEEDING_AGENT_QUEUE_LENGTH_FOR_FEEDBACK;
        let last_feedback = write_speeding_feedback.clone().last(); // Cloning the handle gives read handle

        // If there is no feedback given yet assume zero last timestamp
        let last_feedback_timestamp;
        if last_feedback.is_some() {
            last_feedback_timestamp = last_feedback.unwrap().0;
        }
        else {
            last_feedback_timestamp = 0;
        }

        // If we encounter the same speeding marker x times, send feedback
        let vec: Vec<_> = read_speeding.rev_take_into(|k, v| {
            (k.clone(), v.clone())
        }, target_length);

        // Send feedback if all previous speeding records are new
        let sum = vec.iter().filter(|x| {
            let timestamp = x.0;
            timestamp > last_feedback_timestamp
        }).count();


        // Read the speeding state for feedback
        let mut speeding_state = Default::default();
        if let Component::Speeding (data) = read_speeding.last().unwrap().1 {
            speeding_state = data.state;
        }

        // Give feedback and write result to the component
        if sum == target_length {
            // Randomly pick corresponding feedback text
            let mut text = "";

            // All the options for speeding
            if speeding_state == SpeedingState::SlightlyAboveLimit {
                let slightly: Vec<_> = SPEEDING_AGENT_SLIGHTLY_TEXT
                    .choose_multiple(&mut rand::thread_rng(), 1)
                    .collect();
                text = slightly.first().unwrap();
            } else if speeding_state == SpeedingState::ConsiderablyAboveLimit {
                let considerably: Vec<_> = SPEEDING_AGENT_CONSIDERABLY_TEXT
                    .choose_multiple(&mut rand::thread_rng(), 1)
                    .collect();
                text = considerably.first().unwrap();
            } else if speeding_state == SpeedingState::SevereSpeeding {
                let severely: Vec<_> = SPEEDING_AGENT_SEVERELY_TEXT
                    .choose_multiple(&mut rand::thread_rng(), 1)
                    .collect();
                text = severely.first().unwrap();
            }

            // Log the feedback
            warn!("[SpeedingFeedbackAgent]: Feedback: {:?}", text);

            // Save given feedback
            let result = Component::SpeedingFeedback(SpeedingFeedbackData { value: text });
            write_speeding_feedback.push(timestamp, result);
            write_speeding_feedback.refresh();

            // Finally enqueue
            text2speech.enqueue_for_read(text);
            return true;
        }

        return false;
    };

    let speeding_feedback_agent = AgentEntity::new(AgentType::Observing)
        .set_source_components(vec![5])
        .set_sink_component(6)
        .set_system_method(Arc::new(speeding_feedback_system));


    // LANE CHANGE AGENT
    let lane_change_system = |timestamp: GlobalTimestampMs,
                              read_handles: ComponentReadHandles,
                              write_handle: ComponentWriteHandle,
                              _modules: Modules | -> bool {

        // Extract timestamp
        let timestamp = timestamp_value(timestamp);
        log_agent_run(timestamp, "LaneChangeAgent");

        let mut write_lane_change = unpack_write_handle(&write_handle);

        // Extract knowledge graph handle
        let read_graph = unpack_read_handle(&read_handles, 0);
        if let Component::Graph(data) = read_graph.first().unwrap().1 {
            let graph = data.get_graph_ref();

            // Search for lane pattern
            // Type indices
            let ego_type_index = KnowledgeGraphNode::Ego(Ego::default()).type_index();
            let lane_type_index = KnowledgeGraphNode::Lane(Lane::default()).type_index();
            let is_in_type_index = KnowledgeGraphRelation::isIn(isIn::default()).type_index();

            // Construct the pattern query
            let mut pattern = Pattern::new();
            pattern.node(Node::new(ego_type_index))
                .with_outgoing_relation(Relation::new(is_in_type_index))
                .to_node(Node::new(lane_type_index));

            // Intervals for query
            let first_lane_interval = Interval::from_eternity_to(
                num::clamp(timestamp as i64 - LANE_CHANGE_AGENT_DETECTION_WINDOW_LENGTH,
                           0,
                           i64::MAX) as u64);
            let second_lane_interval = Interval::new(
                num::clamp(timestamp as i64 - LANE_CHANGE_AGENT_DETECTION_WINDOW_LENGTH,
                           0,
                           i64::MAX) as u64, timestamp);

            // If first lane is found
            if pattern.find(first_lane_interval, &graph)
            {
                let first_lane_result_data = pattern.get_result();
                if let KnowledgeGraphNode::Lane(first_lane) = first_lane_result_data[0]
                    .outgoing_relations[0]
                    .to_node[0]
                    .node
                {
                    // And if second lane found
                    if pattern.find(second_lane_interval, &graph)
                    {
                        let second_lane_result_data = pattern.get_result();
                        if let KnowledgeGraphNode::Lane(second_lane) = second_lane_result_data[0]
                            .outgoing_relations[0]
                            .to_node[0]
                            .node
                        {
                            // Then its a lane change
                            let lane1_id = first_lane.laneId as usize;
                            let lane1_name = string_to_static_str(unpack_raw_string(&first_lane.laneName.raw));
                            let lane2_id = second_lane.laneId as usize;
                            let lane2_name = string_to_static_str(unpack_raw_string(&second_lane.laneName.raw));

                            let result = Component::LaneChange(LaneChangeData {
                                from_lane_id: lane1_id,
                                from_lane_name: lane1_name,
                                to_lane_id: lane2_id,
                                to_lane_name: lane2_name
                            });
                            write_lane_change.push(timestamp, result);
                            write_lane_change.refresh();

                            warn!("[LaneChangeAgent]: Lane change: from {:?} to {:?}", lane1_name, lane2_name);
                            return true;
                        }
                    }
                }
            }
        }

        return false;
    };

    let lane_change_agent = AgentEntity::new(AgentType::Periodic(1000))
        .set_source_components(vec![0])
        .set_sink_component(7)
        .set_system_method(Arc::new(lane_change_system));


    // TURN SIGNAL AGENT
    let turn_signal_system = |timestamp: GlobalTimestampMs,
                                        read_handles: ComponentReadHandles,
                                        write_handle: ComponentWriteHandle,
                                        _modules: Modules | -> bool {

        // Extract timestamp
        let timestamp = timestamp_value(timestamp);
        log_agent_run(timestamp, "TurnSignalAgent");

        // Get handles
        let read_ego = unpack_read_handle(&read_handles, 0);
        let mut write_turn_signal = unpack_write_handle(&write_handle);

        let mut ego = Default::default();
        if let Component::EgoShared(data) = read_ego.last().unwrap().1 {
            ego = data;
        }

        // Check turn signaling
        let turn_signal_raw = decode_turn_signal_state(ego.bits_from_car1);

        // Determine the state
        let mut turn_signal_state = TurnSignaling::None;
        if turn_signal_raw == -1 {
            turn_signal_state = TurnSignaling::ToLeft;
        } else if turn_signal_raw == 1 {
            turn_signal_state = TurnSignaling::ToRight;
        }


        // Write the result
        let result = Component::TurnSignal(TurnSignalData { turn_signal: turn_signal_state });
        write_turn_signal.push(timestamp, result);
        write_turn_signal.refresh();

        // Log the result
        if turn_signal_state != TurnSignaling::None {
            warn!("[TurnSignalAgent]: Signal: {:?}", turn_signal_state);
        }

        return true;
    };

    let turn_signal_agent = AgentEntity::new(AgentType::Periodic(1000))
        .set_source_components(vec![1])
        .set_sink_component(8)
        .set_system_method(Arc::new(turn_signal_system));


    // OVERTAKE AGENT
    let overtake_system = |timestamp: GlobalTimestampMs,
                                        read_handles: ComponentReadHandles,
                                        write_handle: ComponentWriteHandle,
                                        _modules: Modules | -> bool {

        // Extract timestamp
        let timestamp = timestamp_value(timestamp);
        log_agent_run(timestamp, "OvertakeAgent");

        let read_lane_change = unpack_read_handle(&read_handles, 1);
        let read_turn_signals = unpack_read_handle(&read_handles, 2);
        let mut write_overtake = unpack_write_handle(&write_handle);
        let read_overtake = write_overtake.clone();

        // At first read the last yielding record timestamp
        let mut last_overtake_timestamp = 0;
        if read_overtake.last().is_some() {
            if let Component::Overtake(data) = read_overtake.last().unwrap().1 {
                last_overtake_timestamp = data.overtake_timestamp;
            }
        }

        // Secondly check if we had two lane changes i.e. last two == back and forth
        let lane_changes: Vec<_> = read_lane_change.rev_take_into(|k, v| {
            (k.clone(), v.clone())
        }, 2);

        if lane_changes.len() < 2 { return false; }

        // Extract lane change and compare
        let mut first_lane_change = Default::default();
        if let Component::LaneChange(data) = lane_changes.first().unwrap().1 {
            first_lane_change = data;
        }
        let mut second_lane_change = Default::default();
        if let Component::LaneChange(data) = lane_changes.last().unwrap().1 {
            second_lane_change = data;
        }

        // Compare lane changes
        if first_lane_change.from_lane_id != second_lane_change.to_lane_id {
            return false;
        }


        // Extract knowledge graph handle
        let read_graph = unpack_read_handle(&read_handles, 0);
        if let Component::Graph(data) = read_graph.first().unwrap().1 {
            let graph = data.get_graph_ref();

            // Search for lane pattern
            // Type indices
            let ego_type_index = KnowledgeGraphNode::Ego(Ego::default()).type_index();
            let car_type_index = KnowledgeGraphNode::Car(Car::default()).type_index();
            let is_behind_type_index = KnowledgeGraphRelation::isBehind(isBehind::default()).type_index();
            let is_on_left_type_index = KnowledgeGraphRelation::isOnLeft(isOnLeft::default()).type_index();
            let is_in_front_type_index = KnowledgeGraphRelation::isInFront(isInFront::default()).type_index();

            // Construct the pattern queries
            let mut pattern_is_behind = Pattern::new();
            pattern_is_behind.node(Node::new(ego_type_index))
                .with_outgoing_relation(Relation::new(is_behind_type_index))
                .to_node(Node::new(car_type_index));

            let mut pattern_is_on_left = Pattern::new();
            pattern_is_on_left.node(Node::new(ego_type_index))
                .with_outgoing_relation(Relation::new(is_on_left_type_index))
                .to_node(Node::new(car_type_index));

            let mut pattern_is_in_front = Pattern::new();
            pattern_is_in_front.node(Node::new(ego_type_index))
                .with_outgoing_relation(Relation::new(is_in_front_type_index))
                .to_node(Node::new(car_type_index));

            // Interval for query, last 30 seconds
            let query_interval = Interval::new(
                num::clamp(timestamp as i64 - OVERTAKE_AGENT_DETECTION_WINDOW_LENGTH,
                           0,
                           i64::MAX) as u64, timestamp);

            // If is behind is found
            if pattern_is_behind.find(query_interval, &graph)
            {
                let is_behind_result_data = pattern_is_behind.get_result();
                let is_behind_timestamp = is_behind_result_data[0].outgoing_relations[0].timestamp;

                if pattern_is_on_left.find(query_interval, &graph)
                {
                    if pattern_is_in_front.find(query_interval, &graph)
                    {
                        let is_in_front_result_data = pattern_is_in_front.get_result();
                        let is_in_front_timestamp = is_in_front_result_data[0].outgoing_relations[0].timestamp;

                        if last_overtake_timestamp != is_in_front_timestamp {

                            // Check turn signals
                            let turn_signals: Vec<_> = read_turn_signals.rev_take_into(|k, v| {
                                (k.clone(), v.clone())
                            }, (OVERTAKE_AGENT_DETECTION_WINDOW_LENGTH / 1000) as usize);

                            // Blinked left and then right
                            let mut blinked_left = 0;
                            let mut blinked_right = 0;
                            for signal in turn_signals {
                                // Make sure the blinking happened after the ovetake started
                                if signal.0 < (timestamp - OVERTAKE_AGENT_TURN_SIGNALS_DETECTION_WINDOW_LENGTH as u64) {
                                    continue;
                                }

                                if let Component::TurnSignal(data) = signal.1 {
                                    if data.turn_signal == TurnSignaling::ToLeft {
                                        blinked_left += 1;
                                    } else if data.turn_signal == TurnSignaling::ToRight {
                                        blinked_right += 1;
                                    }
                                }
                            }

                            // Determine result
                            let mut blinking = OvertakeTurnSignal::None;
                            let threshold = OVERTAKE_AGENT_REQUIRED_BLINKING_DURATION / 1000;

                            if blinked_left >= threshold &&
                                blinked_right >= threshold {
                                blinking = OvertakeTurnSignal::Both;
                            } else if blinked_left >= threshold {
                                blinking = OvertakeTurnSignal::FirstToLeft;
                            } else if blinked_right >= threshold {
                                blinking = OvertakeTurnSignal::SecondToRight;
                            }

                            // Write the result
                            let result = Component::Overtake(OvertakeData {
                                overtake_timestamp: is_in_front_timestamp,
                                turn_signals: blinking,
                                duration: (is_in_front_timestamp - is_behind_timestamp) as usize
                            });
                            write_overtake.push(timestamp, result);
                            write_overtake.refresh();

                            // Log
                            warn!("[OvertakeAgent]: Overtake: {:?}", result);
                            return true;
                        }
                    }
                }
            }
        }

        return false;
    };

    let overtake_agent = AgentEntity::new(AgentType::Observing)
        .set_source_components(vec![0, 7, 8])
        .set_sink_component(9)
        .set_system_method(Arc::new(overtake_system));


    // OVERTAKE FEEDBACK AGENT
    let overtake_feedback_system = |timestamp: GlobalTimestampMs,
                                    read_handles: ComponentReadHandles,
                                    write_handle: ComponentWriteHandle,
                                    modules: Modules | -> bool {

        // Extract timestamp
        let timestamp = timestamp_value(timestamp);
        log_agent_run(timestamp, "OvertakeFeedbackAgent");

        // Get text to speech module
        let text2speech = unpack_module::<SpeechFeedbackModule>(&modules[0].1).unwrap();

        // Get handles and data
        let read_overtake = unpack_read_handle(&read_handles, 0);
        let mut write_overtake_feedback = unpack_write_handle(&write_handle);

        // Check if we have any feedback data, if not return
        if read_overtake.is_empty() {
            return false;
        }

        // Read the overtake details for feedback
        let mut overtake_turn_signals = Default::default();
        if let Component::Overtake (data) = read_overtake.last().unwrap().1 {
            overtake_turn_signals = data.turn_signals;
        }


        // Randomly pick corresponding feedback text
        let mut text = "";

        // All the options for turn signals
        if overtake_turn_signals == OvertakeTurnSignal::None {
            let none: Vec<_> = OVERTAKE_AGENT_NO_BLINK_TEXT
                .choose_multiple(&mut rand::thread_rng(), 1)
                .collect();
            text = none.first().unwrap();
        } else if overtake_turn_signals == OvertakeTurnSignal::Both {
            let both: Vec<_> = OVERTAKE_AGENT_TWO_BLINKS_TEXT
                .choose_multiple(&mut rand::thread_rng(), 1)
                .collect();
            text = both.first().unwrap();
        } else if overtake_turn_signals == OvertakeTurnSignal::FirstToLeft {
            let left: Vec<_> = OVERTAKE_AGENT_LEFT_BLINK_TEXT
                .choose_multiple(&mut rand::thread_rng(), 1)
                .collect();
            text = left.first().unwrap();
        } else if overtake_turn_signals == OvertakeTurnSignal::SecondToRight {
            let right: Vec<_> = OVERTAKE_AGENT_RIGHT_BLINK_TEXT
                .choose_multiple(&mut rand::thread_rng(), 1)
                .collect();
            text = right.first().unwrap();
        }

        // Log the feedback
        warn!("[OvertakeFeedbackAgent]: Feedback: {:?}", text);

        // Save given feedback
        let result = Component::OvertakeFeedback(OvertakeFeedbackData { value: text });
        write_overtake_feedback.push(timestamp, result);
        write_overtake_feedback.refresh();

        // Finally enqueue
        text2speech.enqueue_for_read(text);
        return true;
    };

    let overtake_feedback_agent = AgentEntity::new(AgentType::Observing)
        .set_source_components(vec![9])
        .set_sink_component(10)
        .set_system_method(Arc::new(overtake_feedback_system));


    // CAR YIELDING AGENT
    let car_yielding_system = |timestamp: GlobalTimestampMs,
                               read_handles: ComponentReadHandles,
                               write_handle: ComponentWriteHandle,
                               _modules: Modules | -> bool {

        // Extract timestamp
        let timestamp = timestamp_value(timestamp);
        log_agent_run(timestamp, "CarYieldingAgent");

        let read_turn_signals = unpack_read_handle(&read_handles, 1);
        let mut write_yielding = unpack_write_handle(&write_handle);
        let read_yielding = write_yielding.clone();

        // At first read the last yielding record timestamp
        let mut last_yielding_timestamp = 0;
        if read_yielding.last().is_some() {
            if let Component::CarYielding(data) = read_yielding.last().unwrap().1 {
                last_yielding_timestamp = data.yielding_timestamp;
            }
        }

        // Extract knowledge graph handle
        let read_graph = unpack_read_handle(&read_handles, 0);
        if let Component::Graph(data) = read_graph.first().unwrap().1 {
            let graph = data.get_graph_ref();

            // Search for yields pattern
            // Type indices
            let ego_type_index = KnowledgeGraphNode::Ego(Ego::default()).type_index();
            let lane_type_index = KnowledgeGraphNode::Lane(Lane::default()).type_index();
            let intersection_type_index = KnowledgeGraphNode::Intersection(Intersection::default()).type_index();
            let is_at_type_index = KnowledgeGraphRelation::isAt(isAt::default()).type_index();
            let yields_to_car_in_type_index = KnowledgeGraphRelation::yieldsToCarIn(yieldsToCarIn::default()).type_index();

            // Construct the pattern queries
            let mut intersection_pattern = Pattern::new();
            intersection_pattern.node(Node::new(ego_type_index))
                .with_outgoing_relation(Relation::new(is_at_type_index))
                .to_node(Node::new(intersection_type_index));

            let mut yield_pattern = Pattern::new();
            yield_pattern.node(Node::new(lane_type_index))
                .with_outgoing_relation(Relation::new(yields_to_car_in_type_index))
                .to_node(Node::new(lane_type_index));

            // Interval for query
            let interval = Interval::from_eternity_to(
                num::clamp(timestamp as i64 - 200,
                           0,
                           i64::MAX) as u64);

            // If any yielding
            if yield_pattern.find(interval, &graph) && intersection_pattern.find(interval, &graph)
            {
                let result_data = yield_pattern.get_result();
                if let KnowledgeGraphRelation::yieldsToCarIn(relation) = result_data[0]
                    .outgoing_relations[0].relation
                {
                    let relation_timestamp = result_data[0].outgoing_relations[0].timestamp;
                    let if_direction = decode_ego_direction(relation.ifDirection);

                    // TRAFFIC LIGHT INTERSECTION
                    if if_direction == CarDirection::Unknown && last_yielding_timestamp != relation_timestamp {
                        let ty = CarYieldingType::TrafficLights;
                        let result = Component::CarYielding(CarYieldingData {
                            yielding_timestamp: relation_timestamp,
                            yielding_type: ty,
                        });
                        write_yielding.push(timestamp, result);
                        write_yielding.refresh();

                        warn!("[CarYieldingAgent]: Intersection: {:?}", ty);
                        return true;
                    }

                    // UNCONTROLLED INTERSECTION
                    // Check where ego is going
                    // If no turn signals, assume straight
                    let mut ego_direction = CarDirection::Straight;
                    if !read_turn_signals.is_empty() {
                        if let Component::TurnSignal(data) = read_turn_signals.last().unwrap().1 {
                            if data.turn_signal == TurnSignaling::ToLeft {
                                ego_direction = CarDirection::Left;
                            } else if data.turn_signal == TurnSignaling::ToRight {
                                ego_direction = CarDirection::Right;
                            }
                        }
                    }

                    // Compare with ifDirection
                    if ego_direction == if_direction {
                        let ty = CarYieldingType::Uncontrolled;
                        let result = Component::CarYielding(CarYieldingData {
                            yielding_timestamp: relation_timestamp,
                            yielding_type: ty,
                        });
                        write_yielding.push(timestamp, result);
                        write_yielding.refresh();

                        warn!("[CarYieldingAgent]: Intersection: {:?}", ty);
                        return true;
                    }
                }
            }
        }

        return false;
    };

    let car_yielding_agent = AgentEntity::new(AgentType::Periodic(200))
        .set_source_components(vec![0, 8])
        .set_sink_component(11)
        .set_system_method(Arc::new(car_yielding_system));


    // CAR YIELDING FEEDBACK AGENT
    let car_yielding_feedback_system = |timestamp: GlobalTimestampMs,
                                        read_handles: ComponentReadHandles,
                                        write_handle: ComponentWriteHandle,
                                        modules: Modules | -> bool {

        // Extract timestamp
        let timestamp = timestamp_value(timestamp);
        log_agent_run(timestamp, "CarYieldingFeedbackAgent");

        // Get text to speech module
        let text2speech = unpack_module::<SpeechFeedbackModule>(&modules[0].1).unwrap();

        // Get handles and data
        let read_yielding = unpack_read_handle(&read_handles, 0);
        let mut write_yielding_feedback = unpack_write_handle(&write_handle);

        // Check if we have any feedback data, if not return
        if read_yielding.is_empty() {
            return false;
        }

        // Read the yielding details for feedback
        let mut yielding_type = Default::default();
        if let Component::CarYielding(data) = read_yielding.last().unwrap().1 {
            yielding_type = data.yielding_type;
        }


        // Randomly pick corresponding feedback text
        let mut text = "";

        // All the options for yielding
        if yielding_type == CarYieldingType::TrafficLights {
            let traffic_lights: Vec<_> = CAR_YIELDING_AGENT_RED_LIGHT_TEXT
                .choose_multiple(&mut rand::thread_rng(), 1)
                .collect();
            text = traffic_lights.first().unwrap();
        } else if yielding_type == CarYieldingType::TrafficLights {
            let uncontrolled_3way: Vec<_> = CAR_YIELDING_AGENT_NO_YIELD_TEXT
                .choose_multiple(&mut rand::thread_rng(), 1)
                .collect();
            text = uncontrolled_3way.first().unwrap();
        }

        // Log the feedback
        warn!("[CarYieldingFeedbackAgent]: Feedback: {:?}", text);

        // Save given feedback
        let result = Component::CarYieldingFeedback(CarYieldingFeedbackData { value: text });
        write_yielding_feedback.push(timestamp, result);
        write_yielding_feedback.refresh();

        // Finally enqueue
        text2speech.enqueue_for_read(text);
        return true;
    };

    let car_yielding_feedback_agent = AgentEntity::new(AgentType::Observing)
        .set_source_components(vec![11])
        .set_sink_component(12)
        .set_system_method(Arc::new(car_yielding_feedback_system));


    // PEDESTRIAN YIELDING AGENT
    let pedestrian_yielding_system = |timestamp: GlobalTimestampMs,
                               read_handles: ComponentReadHandles,
                               write_handle: ComponentWriteHandle,
                               _modules: Modules | -> bool {

        // Extract timestamp
        let timestamp = timestamp_value(timestamp);
        log_agent_run(timestamp, "PedestrianYieldingAgent");

        let mut write_yielding = unpack_write_handle(&write_handle);
        let read_yielding = write_yielding.clone();

        // At first read the last yielding record timestamp
        let mut last_yielding_timestamp = 0;
        if read_yielding.last().is_some() {
            if let Component::PedestrianYielding(data) = read_yielding.last().unwrap().1 {
                last_yielding_timestamp = data.yielding_timestamp;
            }
        }


        // Extract knowledge graph handle
        let read_graph = unpack_read_handle(&read_handles, 0);
        if let Component::Graph(data) = read_graph.first().unwrap().1 {
            let graph = data.get_graph_ref();

            // Search for yields pattern
            // Type indices
            let ego_type_index = KnowledgeGraphNode::Ego(Ego::default()).type_index();
            let crosswalk_type_index = KnowledgeGraphNode::Crosswalk(Crosswalk::default()).type_index();
            let passes_over_type_index = KnowledgeGraphRelation::passesOver(passesOver::default()).type_index();

            // Construct the pattern query
            let mut crosswalk_pattern = Pattern::new();
            crosswalk_pattern.node(Node::new(ego_type_index))
                .with_outgoing_relation(Relation::new(passes_over_type_index))
                .to_node(Node::new(crosswalk_type_index));

            // Interval for query
            let interval = Interval::from_eternity_to(
                num::clamp(timestamp as i64 - 200,
                           0,
                           i64::MAX) as u64);

            // If any yielding
            if crosswalk_pattern.find(interval, &graph)
            {
                let result_data = crosswalk_pattern.get_result();
                if let KnowledgeGraphNode::Crosswalk(node) = result_data[0]
                    .outgoing_relations[0]
                    .to_node[0]
                    .node
                {
                    let node_timestamp = result_data[0].outgoing_relations[0].to_node[0].timestamp;
                    let pedestrian_count = node.crosswalkPedestrianCount;

                    // CROSSWALK
                    if pedestrian_count > 0 && node_timestamp != last_yielding_timestamp {
                        let ty = PedestrianYieldingType::Crosswalk;
                        let result = Component::PedestrianYielding(PedestrianYieldingData {
                            yielding_timestamp: node_timestamp,
                            yielding_type: ty,
                        });
                        write_yielding.push(timestamp, result);
                        write_yielding.refresh();

                        warn!("[PedestrianYieldingAgent]: Crosswalk pedestrian count: {:?}", pedestrian_count/2);
                        return true;
                    }
                }
            }
        }

        return false;
    };

    let pedestrian_yielding_agent = AgentEntity::new(AgentType::Periodic(200))
        .set_source_components(vec![0])
        .set_sink_component(13)
        .set_system_method(Arc::new(pedestrian_yielding_system));


    // PEDESTRIAN YIELDING FEEDBACK AGENT
    let pedestrian_yielding_feedback_system = |timestamp: GlobalTimestampMs,
                                        read_handles: ComponentReadHandles,
                                        write_handle: ComponentWriteHandle,
                                        modules: Modules | -> bool {

        // Extract timestamp
        let timestamp = timestamp_value(timestamp);
        log_agent_run(timestamp, "PedestrianYieldingFeedbackAgent");

        // Get text to speech module
        let text2speech = unpack_module::<SpeechFeedbackModule>(&modules[0].1).unwrap();

        // Get handles and data
        let read_yielding = unpack_read_handle(&read_handles, 0);
        let mut write_yielding_feedback = unpack_write_handle(&write_handle);

        // Check if we have any feedback data, if not return
        if read_yielding.is_empty() {
            return false;
        }

        // Read the yielding details for feedback
        let mut yielding_type = Default::default();
        if let Component::PedestrianYielding(data) = read_yielding.last().unwrap().1 {
            yielding_type = data.yielding_type;
        }


        // Randomly pick corresponding feedback text
        let mut text = "";

        // All the options for yielding
        if yielding_type == PedestrianYieldingType::Crosswalk {
            let crosswalk: Vec<_> = PEDESTRIAN_YIELDING_AGENT_CROSSWALK_TEXT
                .choose_multiple(&mut rand::thread_rng(), 1)
                .collect();
            text = crosswalk.first().unwrap();
        }

        // Log the feedback
        warn!("[PedestrianYieldingFeedbackAgent]: Feedback: {:?}", text);

        // Save given feedback
        let result = Component::PedestrianYieldingFeedback(PedestrianYieldingFeedbackData { value: text });
        write_yielding_feedback.push(timestamp, result);
        write_yielding_feedback.refresh();

        // Finally enqueue
        text2speech.enqueue_for_read(text);
        return true;
    };

    let pedestrian_yielding_feedback_agent = AgentEntity::new(AgentType::Observing)
        .set_source_components(vec![13])
        .set_sink_component(14)
        .set_system_method(Arc::new(pedestrian_yielding_feedback_system));


    // Finally register all the entity agents
    platform.register_entity("incorrect_gear_agent",incorrect_gear_agent);
    platform.register_entity("incorrect_gear_feedback_agent",incorrect_gear_feedback_agent);
    platform.register_entity("speeding_agent",speeding_agent);
    platform.register_entity("speeding_feedback_agent",speeding_feedback_agent);
    platform.register_entity("lane_change_agent",lane_change_agent);
    platform.register_entity("turn_signal_agent", turn_signal_agent);
    platform.register_entity("overtake_agent",overtake_agent);
    platform.register_entity("overtake_feedback_agent", overtake_feedback_agent);
    platform.register_entity("car_yielding_agent", car_yielding_agent);
    platform.register_entity("car_yielding_feedback_agent", car_yielding_feedback_agent);
    platform.register_entity("pedestrian_yielding_agent", pedestrian_yielding_agent);
    platform.register_entity("pedestrian_yielding_feedback_agent", pedestrian_yielding_feedback_agent);

    // Introduce the system
    let text2speech = Box::new(SpeechFeedbackModule::new());
    text2speech.enqueue_for_read(SYSTEM_INTRODUCTION_TEXT1);
    text2speech.enqueue_for_read(SYSTEM_INTRODUCTION_TEXT2);

    // Kick start the heart
    info!("[Agent platform]: Agent platform started running...");
    platform.run();
}
