// Platform building blocks
use crate::ecs::*;
use crate::components::*;
use crate::modules::*;

// Imports
use config::*;

use shared_memory_interface::*;
use shared_memory_interface::types::VisualSharedMemory;
use shared_memory_interface::shared_memory::SharedMemory;
use shared_memory_interface::constants::EventType;

use knowledge_graph::graph::*;

use std::sync::{Arc};
use std::sync::atomic::{AtomicU64, Ordering};

use std::thread;
use std::thread::{JoinHandle};
use std::time::{Duration, Instant};

use crossbeam::channel::{unbounded, Sender, Receiver, tick};
use ordered_float::OrderedFloat;


// Platform
pub struct AgentPlatform {
    components: Vec<(&'static str, ComponentStorage<u64, Component>)>,
    entities: Vec<(&'static str, AgentEntity<u64, Component>)>,
    modules: Vec<(&'static str, Box<dyn Module>)>,

    #[allow(dead_code)]
    threads: Vec<JoinHandle<()>>,

    // usize is index of the component
    channels: Vec<(usize, Option<Sender<Instant>>, Option<Receiver<Instant>>, Option<Receiver<Instant>>)>,

    // Shared memory access
    graph_ptr: Arc<ShareableRefCell<*mut SharedMemory<KnowledgeGraph>>>,
    visual_ptr: Arc<ShareableRefCell<*mut SharedMemory<VisualSharedMemory>>>,
    current_timestamp: Arc<AtomicU64>
}
impl AgentPlatform {
    pub fn new() -> Self {
        AgentPlatform {
            modules: vec![],
            threads: vec![],
            components: vec![],
            entities: vec![],
            channels: vec![],

            graph_ptr: Arc::new(ShareableRefCell::new(std::ptr::null_mut())),
            visual_ptr: Arc::new(ShareableRefCell::new(std::ptr::null_mut())),
            current_timestamp: Arc::new(AtomicU64::new(0)),
        }
    }

    pub fn register_component(&mut self, name: &'static str, _component: Component) -> usize {
        let index = self.components.len();
        let storage = ComponentStorage::new();
        self.components.push((name, storage));
        return index;
    }

    pub fn get_component_storage(&self, name: &'static str) -> Option<&ComponentStorage<u64, Component>> {
        for storage in self.components.iter() {
            if storage.0 == name {
                return Some(&storage.1)
            }
        }
        return None;
    }

    pub fn register_entity(&mut self, name: &'static str, entity: AgentEntity<u64, Component>) -> usize {
        match entity.agent_type {
            AgentType::Periodic(period) => {
                let ticker = tick(Duration::from_millis(period));
                let (sender, receiver) = unbounded();
                let component_index = entity.sink_component.unwrap();
                self.channels.push((component_index, Some(sender), Some(receiver), Some(ticker)));
            }
            AgentType::Observing => {
                let (sender, receiver) = unbounded();
                let component_index = entity.sink_component.unwrap();
                self.channels.push((component_index, Some(sender), Some(receiver), None));
            }
        }

        let index = self.entities.len();
        self.entities.push((name, entity));
        return index;
    }

    pub fn register_module(&mut self, name: &'static str, module: Box<dyn Module>) -> usize {
        let index = self.modules.len();
        self.modules.push((name, module));
        return index;
    }

    pub fn initialize(&mut self) {
        // Create knowledge graph shared memory region
        self.graph_ptr = Arc::new(ShareableRefCell::new(initialize_shared_memory::
        <KnowledgeGraph>(KNOWLEDGE_GRAPH_SHARED_REGION_NAME,
                         KNOWLEDGE_GRAPH_SHARED_REGION_SIZE)));

        // Create visual shared memory region
        self.visual_ptr = Arc::new(ShareableRefCell::new(initialize_shared_memory::
        <VisualSharedMemory>(VISUAL_SHARED_REGION_NAME,
                             VISUAL_SHARED_REGION_SIZE)));


        // Create some auto-populated components as implicit data sources
        // KNOWLEDGE GRAPH component i.e. a pointer to the shared memory with graph
        self.register_component("graph", Component::Graph(GraphData::default()));
        let graph_storage = self.get_component_storage("graph").unwrap();
        let graph_write_handle = graph_storage.write_handle.clone();
        let mut graph_write = (*graph_write_handle).borrow_mut();

        let graph_raw = *(*self.graph_ptr).borrow_mut();
        let graph_ptr = unsafe { &mut *graph_raw };
        let graph_component = Component::Graph(GraphData::new(graph_ptr));

        graph_write.push(0, graph_component);
        graph_write.refresh();


        // Components derived from visual shared memory data
        // EGO DATA
        self.register_component("ego", Component::EgoShared(EgoData::default()));
        let ego_storage = self.get_component_storage("ego").unwrap();
        let ego_write_handle = ego_storage.write_handle.clone();

        // GAZE DATA
        self.register_component("gaze", Component::GazeShared(GazeData::default()));
        let gaze_storage = self.get_component_storage("gaze").unwrap();
        let gaze_write_handle = gaze_storage.write_handle.clone();

        // Visual shared memory reading thread
        let timestamp_handle = self.current_timestamp.clone();
        let visual_handle = self.visual_ptr.clone();

        thread::spawn(move || {
            loop {
                // Get visual shared memory info
                let visual_raw = *(*visual_handle).borrow_mut();
                let visual = unsafe { &mut *visual_raw }.get_data_ref();

                // At first update platform TIMESTAMP!
                // If its zero extract it from the knowledge graph
                let timestamp = visual.vehicle_data.idx as u64;
                timestamp_handle.store(timestamp, Ordering::SeqCst);

                // EGO DATA
                // Borrow write handle for ego data
                let mut ego_write = (*ego_write_handle).borrow_mut();

                // Fill in the component data
                let ego_data = EgoData {
                    position_x: OrderedFloat(visual.vehicle_data.position_x),
                    position_y: OrderedFloat(visual.vehicle_data.position_y),
                    position_z: OrderedFloat(visual.vehicle_data.position_z),
                    direction_x: OrderedFloat(visual.vehicle_data.direction_x),
                    direction_y: OrderedFloat(visual.vehicle_data.direction_y),
                    direction_z: OrderedFloat(visual.vehicle_data.direction_z),
                    direction_w: OrderedFloat(visual.vehicle_data.direction_w),
                    speed: OrderedFloat(visual.vehicle_data.speed),
                    speed_vector_x: OrderedFloat(visual.vehicle_data.speed_vector_x),
                    speed_vector_y: OrderedFloat(visual.vehicle_data.speed_vector_y),
                    speed_vector_z: OrderedFloat(visual.vehicle_data.speed_vector_z),
                    steering_wheel: OrderedFloat(visual.vehicle_data.steering_wheel),
                    gear: visual.vehicle_data.gear,
                    brake: OrderedFloat(visual.vehicle_data.brake),
                    clutch: OrderedFloat(visual.vehicle_data.clutch),
                    throttle: OrderedFloat(visual.vehicle_data.throttle),
                    rpm: OrderedFloat(visual.vehicle_data.rpm),
                    torque: OrderedFloat(visual.vehicle_data.torque),
                    bits_from_car1: visual.vehicle_data.bits_from_car1,
                };
                // Add component data
                let ego = Component::EgoShared(ego_data);
                ego_write.push(timestamp, ego);
                ego_write.refresh();

                // GAZE DATA
                // Borrow write handle for gaze data
                let mut gaze_write = (*gaze_write_handle).borrow_mut();

                let event = visual.event_buffer[visual.event_buffer_head as usize];
                let last_event = gaze_write.clone().last();

                if event.event_type == EventType::EyetrackerEvent.to_u32() {
                    if last_event.is_none() {
                        let gaze_data = GazeData { data: event.flag };
                        let gaze = Component::GazeShared(gaze_data);
                        gaze_write.push(timestamp, gaze);
                        gaze_write.refresh();
                    } else {
                        // Read the last event first to see if its the same one
                        let mut last_event_data = Default::default();
                        if let Component::GazeShared (data) = last_event.unwrap().1 {
                            last_event_data = data.data;
                        }

                        if last_event_data != event.flag {
                            let gaze_data = GazeData { data: event.flag };
                            let gaze = Component::GazeShared(gaze_data);
                            gaze_write.push(timestamp, gaze);
                            gaze_write.refresh();
                        }
                    }
                }

                // Wait for the speficied period of time
                let delay = Duration::from_millis(SHARED_MEMORY_UPDATE_FREQUENCY);
                thread::sleep(delay);
            }
        });
    }

    pub fn run(&mut self) {
        let mut threads = Vec::new();

        // Execute each agent's system method on a separate thread
        for (i, entity) in self.entities.iter().enumerate() {

            let agent_type = entity.1.agent_type.clone();
            let method = entity.1.system_method.clone();
            let source_indices = entity.1.source_components.clone().unwrap();
            let sink_index = entity.1.sink_component.clone().unwrap();

            // Extract corresponding source and sink components
            let sink_component = self.components[sink_index].1.write_handle.clone();
            let mut source_components = Vec::new();
            for index in source_indices.iter().cloned() {
                source_components.push(self.components[index].1.read_handle.clone());
            }

            // Determine sink channel message count (enough for every observing agent)
            let mut sink_channel_message_count = 0;
            for (_, ent) in self.entities.iter().enumerate() {
                for index in ent.1.source_components.clone().unwrap().iter() {
                    if *index == sink_index {
                        sink_channel_message_count += 1;
                    }
                }
            }


            // Extract channels
            let sink_channel = (sink_index, self.channels[i].1.clone());
            let mut source_channels: Vec::<(usize, Receiver<Instant>)> = Vec::new();

            match agent_type {
                AgentType::Observing => {
                    for index in source_indices.iter().cloned() {
                        for channel in &self.channels {
                            if channel.0 == index {
                                source_channels.push((index, channel.2.clone().unwrap()));
                            }
                        }
                    }
                }
                AgentType::Periodic(_) => {
                    for channel in &self.channels {
                        if channel.0 == sink_index {
                            source_channels.push((sink_index, channel.3.clone().unwrap()));
                        }
                    }
                }
            }

            // Modules
            let platform_modules = self.modules.clone();

            // Timestamp
            let timestamp = self.current_timestamp.clone();

            // Spawn thread for each agent
            let thread_handle = thread::spawn(move || {
                loop {
                    // Wait until you get notifications from all source components
                    // or from a tick channel
                    for channel in &source_channels {
                        channel.1.recv().unwrap();
                    }

                    // Clone modules for the agent
                    let modules = platform_modules.clone();

                    // Run the actual agent code
                    let result = method(timestamp.clone(),
                           source_components.clone(),
                           sink_component.clone(),
                           modules);

                    // Notify about new data, if agent returned true
                    if sink_channel.1.is_some() && result == true {
                        // Send all the messages required
                        for _ in 0..sink_channel_message_count {
                            sink_channel.1.as_ref().unwrap().send(Instant::now()).unwrap();
                        }
                    }
                }
            });
            threads.push(thread_handle);
        }

        // This will wait till infinity
        for thread in threads {
            thread.join().unwrap();
        }
    }
}
