// Heavily inspired by Rust Codegen: https://github.com/carllerche/codegen

use std::fmt;
use std::fmt::Write;

// Definition of  all types
#[derive(Clone, Debug)]
pub struct Scope {
    using: Vec<String>,
    items: Vec<Item>,
}

#[derive(Clone, Debug)]
pub struct Namespace {
    name: String,
    items: Vec<Item>,
}

#[derive(Clone, Debug)]
enum Item {
    Namespace(Namespace),
    Struct(Struct),
    Class(Class),
    Enum(Enum),
    Raw(String),
}

#[derive(Clone, Debug)]
pub struct Enum {
    name: String,
    modifiers: Vec<String>,
    options: Vec<(String, Option<String>)>,
}

#[derive(Clone, Debug)]
pub struct Struct {
    name: String,
    modifiers: Vec<String>,
    headers: Vec<Header>,
    attributes: Vec<Attribute>,
    signatures: Vec<MethodSignature>,
    methods: Vec<Method>,
}

#[derive(Clone, Debug)]
pub struct Class {
    name: String,
    modifiers: Vec<String>,
    headers: Vec<Header>,
    attributes: Vec<Attribute>,
    signatures: Vec<MethodSignature>,
    methods: Vec<Method>,
}

#[derive(Clone, Debug)]
struct Attribute {
    header: Option<String>,
    name: String,
    ty: String,
}

#[derive(Clone, Debug)]
pub struct MethodSignature {
    name: String,
    headers: Vec<Header>,
    modifiers: Vec<String>,
    return_type: String,
    parameters: Vec<Parameter>,
}

#[derive(Clone, Debug)]
pub struct Method {
    name: String,
    modifiers: Vec<String>,
    return_type: String,
    parameters: Vec<Parameter>,
    body: Vec<String>,
}

#[derive(Clone, Debug)]
pub struct Header {
    name: String,
    value: String,
    option: Option<String>,
    is_value_string: bool,
}

#[derive(Clone, Debug)]
pub struct Parameter {
    name: String,
    ty: String,
}


// Implementations
impl Scope {
    pub fn new() -> Self {
        Scope {
            using: vec![],
            items: vec![],
        }
    }

    pub fn add_using(&mut self, resource: &str) -> &mut Self {
        let using = String::from(resource);
        self.using.push(using);
        return self;
    }

    pub fn new_namespace(&mut self, name: &str) -> &mut Namespace {
        let namespace = Namespace::new(name);
        let item = Item::Namespace(namespace);
        self.items.push(item);

        return match self.items.last_mut().unwrap() {
            Item::Namespace(ref mut namespace) => namespace,
            _ => unreachable!()
        };
    }

    pub fn new_struct(&mut self, name: &str) -> &mut Struct {
        let strct = Struct::new(name);
        let item = Item::Struct(strct);
        self.items.push(item);

        return match self.items.last_mut().unwrap() {
            Item::Struct(ref mut strct) => strct,
            _ => unreachable!()
        };
    }

    pub fn new_class(&mut self, name: &str) -> &mut Class {
        let class = Class::new(name);
        let item = Item::Class(class);
        self.items.push(item);

        return match self.items.last_mut().unwrap() {
            Item::Class(ref mut class) => class,
            _ => unreachable!()
        };
    }

    pub fn new_enum(&mut self, name: &str) -> &mut Enum {
        let enm = Enum::new(name);
        let item = Item::Enum(enm);
        self.items.push(item);

        return match self.items.last_mut().unwrap() {
            Item::Enum(ref mut enm) => enm,
            _ => unreachable!()
        };
    }

    pub fn add_raw(&mut self, raw: &str) -> &mut Self {
        let raw = String::from(raw);
        let item = Item::Raw(raw);
        self.items.push(item);
        return self;
    }


    pub fn add_namespace(&mut self, namespace: Namespace) -> &mut Self {
        let item = Item::Namespace(namespace);
        self.items.push(item);
        return self;
    }

    pub fn add_struct(&mut self, strct: Struct) -> &mut Self {
        let item = Item::Struct(strct);
        self.items.push(item);

        return self;
    }

    pub fn add_class(&mut self, class: Class) -> &mut Self{
        let item = Item::Class(class);
        self.items.push(item);

        return self;
    }

    pub fn add_enum(&mut self, enm: Enum) -> &mut Self {
        let item = Item::Enum(enm);
        self.items.push(item);

        return self;
    }


    // Formatting
    pub fn to_string(&self) -> String {
        let mut result = String::new();
        self.to_code(&mut Formatter::new(&mut result)).unwrap();
        return result;
    }

    pub fn to_code(&self, formatter: &mut Formatter) -> fmt::Result {

        // Using directives
        for using in &self.using {
            write!(formatter, "using {};\n", using)?;
        }

        if !self.using.is_empty() {
            write!(formatter, "\n")?;
        }

        // Items
        for item in &self.items {
            match item {
                Item::Namespace(ref r) => r.to_code(formatter)?,
                Item::Struct(ref r) => r.to_code(formatter)?,
                Item::Class(ref r) => r.to_code(formatter)?,
                Item::Enum(ref r) => r.to_code(formatter)?,
                Item::Raw(ref r) => {
                    write!(formatter, "{}", r)?;
                }
            }
            write!(formatter, "\n")?;
        }

        return Ok(());
    }
}

impl Namespace {
    pub fn new(name: &str) -> Self {
        Namespace {
            name: String::from(name),
            items: vec![],
        }
    }

    pub fn new_namespace(&mut self, name: &str) -> &mut Namespace {
        let namespace = Namespace::new(name);
        let item = Item::Namespace(namespace);
        self.items.push(item);

        return match self.items.last_mut().unwrap() {
            Item::Namespace(ref mut namespace) => namespace,
            _ => unreachable!()
        };
    }

    pub fn new_struct(&mut self, name: &str) -> &mut Struct {
        let strct = Struct::new(name);
        let item = Item::Struct(strct);
        self.items.push(item);

        return match self.items.last_mut().unwrap() {
            Item::Struct(ref mut strct) => strct,
            _ => unreachable!()
        };
    }

    pub fn new_class(&mut self, name: &str) -> &mut Class {
        let class = Class::new(name);
        let item = Item::Class(class);
        self.items.push(item);

        return match self.items.last_mut().unwrap() {
            Item::Class(ref mut class) => class,
            _ => unreachable!()
        };
    }

    pub fn new_enum(&mut self, name: &str) -> &mut Enum {
        let enm = Enum::new(name);
        let item = Item::Enum(enm);
        self.items.push(item);

        return match self.items.last_mut().unwrap() {
            Item::Enum(ref mut enm) => enm,
            _ => unreachable!()
        };
    }

    pub fn add_raw(&mut self, raw: &str) -> &mut Self {
        let raw = String::from(raw);
        let item = Item::Raw(raw);
        self.items.push(item);

        return self;
    }

    pub fn add_namespace(&mut self, namespace: Namespace) -> &mut Self {
        let item = Item::Namespace(namespace);
        self.items.push(item);
        return self;
    }

    pub fn add_struct(&mut self, strct: Struct) -> &mut Self {
        let item = Item::Struct(strct);
        self.items.push(item);

        return self;
    }

    pub fn add_class(&mut self, class: Class) -> &mut Self{
        let item = Item::Class(class);
        self.items.push(item);

        return self;
    }

    pub fn add_enum(&mut self, enm: Enum) -> &mut Self {
        let item = Item::Enum(enm);
        self.items.push(item);

        return self;
    }

    pub fn to_code(&self, formatter: &mut Formatter) -> fmt::Result {

        // Name
        write!(formatter, "namespace {}", self.name)?;

        // Items
        if self.items.is_empty() {
            write!(formatter, " {{}}\n")?;
        } else {
            write!(formatter, " {{\n\n")?;
            formatter.indentation(| format | {
                for item in &self.items {
                    match item {
                        Item::Namespace(ref r) => r.to_code(format)?,
                        Item::Struct(ref r) => r.to_code(format)?,
                        Item::Class(ref r) => r.to_code(format)?,
                        Item::Enum(ref r) => r.to_code(format)?,
                        Item::Raw(ref r) => {
                            write!(format, "{}", r)?;
                        }
                    }
                    write!(format, "\n")?;
                }
                return Ok(());
            })?;
            write!(formatter, "}}\n")?;
        }

        return Ok(());
    }
}

impl Enum {
    pub fn new(name: &str) -> Self {
        Enum {
            name: String::from(name),
            modifiers: vec![],
            options: vec![],
        }
    }

    pub fn new_modifier(&mut self, name: &str) -> &mut Self {
        let modifier = String::from(name);
        self.modifiers.push(modifier);
        return self;
    }

    pub fn new_option(&mut self, name: &str) -> &mut Self {
        let option = String::from(name);
        self.options.push((option, None));
        return self;
    }

    pub fn new_option_with_value(&mut self, name: &str, value: &str) -> &mut Self {
        let option = String::from(name);
        let value = String::from(value);
        self.options.push((option, Some(value)));
        return self;
    }

    pub fn to_code(&self, formatter: &mut Formatter) -> fmt::Result {

        // Modifiers
        for modifier in &self.modifiers {
            write!(formatter, "{} ", modifier)?;
        }

        // Name
        write!(formatter, "enum {}", self.name)?;

        // Options
        if self.options.is_empty()
        {
            write!(formatter, " {{}}\n")?;
        } else {
            write!(formatter, " {{\n")?;
            formatter.indentation(|format| {
                for option in &self.options {
                    if option.1.is_some() {
                        let value = option.1.as_ref().unwrap();
                        write!(format, "{} = {},\n", option.0, value)?;
                    } else {
                        write!(format, "{},\n", option.0)?;
                    }
                }
                return Ok(());
            })?;
            write!(formatter, "}}\n")?;
        }

        return Ok(());
    }
}

impl Struct {
    pub fn new(name: &str) -> Self {
        Struct {
            name: String::from(name),
            modifiers: vec![],
            headers: vec![],
            attributes: vec![],
            signatures: vec![],
            methods: vec![],
        }
    }

    pub fn new_modifier(&mut self, name: &str) -> &mut Self {
        let modifier = String::from(name);
        self.modifiers.push(modifier);
        return self;
    }

    pub fn new_header(&mut self, name: &str, value: &str,
                      option: Option<&str>, is_value_string: bool) -> &mut Self {
        let header = Header::new(name, value, option, is_value_string);
        self.headers.push(header);
        return self;
    }

    pub fn new_attribute(&mut self, name: &str, ty: &str) -> &mut Self {
        let attribute = Attribute::new(name, ty);
        self.attributes.push(attribute);
        return self;
    }

    pub fn new_attribute_with_header(&mut self, name: &str, ty: &str, header: &str) -> &mut Self {
        let mut attribute = Attribute::new(name, ty);
        attribute.add_header(header);
        self.attributes.push(attribute);
        return self;
    }

    pub fn add_method_signature(&mut self, signature: MethodSignature) -> &mut MethodSignature {
        self.signatures.push(signature);
        return self.signatures.last_mut().unwrap();
    }

    pub fn add_method(&mut self, method: Method) -> &mut Method {
        self.methods.push(method);
        return self.methods.last_mut().unwrap();
    }

    // Format into code
    pub fn to_code(&self, formatter: &mut Formatter) -> fmt::Result {

        // Headers
        for header in &self.headers {
            header.to_code(formatter)?;
            write!(formatter, "\n")?;
        }

        // Modifiers
        for modifier in &self.modifiers {
            write!(formatter, "{} ", modifier)?;
        }

        // Name
        write!(formatter, "struct {}", self.name)?;

        // Body
        if self.signatures.is_empty() &&
            self.attributes.is_empty() &&
            self.methods.is_empty()
        {
            write!(formatter, " {{}}\n")?;
        } else {
            write!(formatter, " {{\n")?;
            formatter.indentation(| format | {

                // Attributes
                if !self.attributes.is_empty() {
                    for attribute in &self.attributes {
                        attribute.to_code(format)?;
                        write!(format, "\n")?;
                    }
                    write!(format, "\n")?;
                }

                // Signatures
                if !self.signatures.is_empty() {
                    for signature in &self.signatures {
                        signature.to_code(format)?;
                        write!(format, "\n\n")?;
                    }
                    write!(format, "\n")?;
                }

                // Methods
                if !self.methods.is_empty() {
                    for method in &self.methods {
                        method.to_code(format)?;
                        write!(format, "\n\n")?;
                    }
                    write!(format, "\n")?;
                }

                return Ok(());
            })?;

            write!(formatter, "}}\n")?;
        }

        return Ok(());
    }
}

impl Class {
    pub fn new(name: &str) -> Self {
        Class {
            name: String::from(name),
            modifiers: vec![],
            headers: vec![],
            attributes: vec![],
            signatures: vec![],
            methods: vec![],
        }
    }

    pub fn new_modifier(&mut self, name: &str) -> &mut Self {
        let modifier = String::from(name);
        self.modifiers.push(modifier);
        return self;
    }

    pub fn new_header(&mut self, name: &str, value: &str,
                      option: Option<&str>, is_value_string: bool) -> &mut Self {
        let header = Header::new(name, value, option, is_value_string);
        self.headers.push(header);
        return self;
    }

    pub fn new_attribute(&mut self, name: &str, ty: &str) -> &mut Self {
        let attribute = Attribute::new(name, ty);
        self.attributes.push(attribute);
        return self;
    }

    pub fn add_method_signature(&mut self, signature: MethodSignature) -> &mut MethodSignature {
        self.signatures.push(signature);
        return self.signatures.last_mut().unwrap();
    }

    pub fn add_method(&mut self, method: Method) -> &mut Method {
        self.methods.push(method);
        return self.methods.last_mut().unwrap();
    }

    // Format into code
    pub fn to_code(&self, formatter: &mut Formatter) -> fmt::Result {

        // Headers
        for header in &self.headers {
            header.to_code(formatter)?;
            write!(formatter, "\n")?;
        }

        // Modifiers
        for modifier in &self.modifiers {
            write!(formatter, "{} ", modifier)?;
        }

        // Name
        write!(formatter, "class {}", self.name)?;

        // Body
        if self.signatures.is_empty() &&
            self.attributes.is_empty() &&
            self.methods.is_empty()
        {
            write!(formatter, " {{}}\n")?;
        } else {
            write!(formatter, " {{\n")?;
            formatter.indentation(| format | {

                // Attributes
                if !self.attributes.is_empty() {
                    for attribute in &self.attributes {
                        attribute.to_code(format)?;
                        write!(format, "\n")?;
                    }
                    write!(format, "\n")?;
                }

                // Signatures
                if !self.signatures.is_empty() {
                    for signature in &self.signatures {
                        signature.to_code(format)?;
                        write!(format, "\n\n")?;
                    }
                    write!(format, "\n")?;
                }

                // Methods
                if !self.methods.is_empty() {
                    for method in &self.methods {
                        method.to_code(format)?;
                        write!(format, "\n\n")?;
                    }
                    write!(format, "\n")?;
                }

                return Ok(());
            })?;

            write!(formatter, "}}\n")?;
        }

        return Ok(());
    }
}

impl Attribute {
    pub fn new(name: &str, ty: &str) -> Self {
        Attribute {
            header: None,
            name: String::from(name),
            ty: String::from(ty),
        }
    }

    pub fn add_header(&mut self, header: &str) {
        self.header = Some(String::from(header));
    }

    pub fn to_code(&self, formatter: &mut Formatter) -> fmt::Result {
        if self.header.is_some() {
            write!(formatter, "{}\n", self.header.as_ref().unwrap())?;
        }
        write!(formatter, "public {} {};", self.ty, self.name)?;
        return Ok(());
    }
}

impl Parameter {
    pub fn new(name: &str, ty: &str) -> Self {
        Parameter {
            name: String::from(name),
            ty: String::from(ty),
        }
    }

    pub fn to_code(&self, formatter: &mut Formatter) -> fmt::Result {
        write!(formatter, "{} {}", self.ty, self.name)?;
        return Ok(());
    }
}

impl Header {
    pub fn new(name: &str, value: &str, option: Option<&str>, is_value_string: bool) -> Self {
        let mut header_option = None;
        if option.is_some() {
            header_option = Some(String::from(option.unwrap()));
        }

        Header {
            name: String::from(name),
            value: String::from(value),
            option: header_option,
            is_value_string,
        }
    }

    pub fn to_code(&self, formatter: &mut Formatter) -> fmt::Result {
        if !self.value.is_empty() {
            if self.is_value_string {
                if self.option.is_some() {
                    write!(formatter, "[{} (\"{}\", {})]", self.name, self.value,
                           self.option.as_ref().unwrap())?;
                }
                else {
                    write!(formatter, "[{} (\"{}\")]", self.name, self.value)?;
                }
            } else {
                write!(formatter, "[{} ({})]", self.name, self.value)?;
            }
        } else {
            write!(formatter, "[{}]", self.name)?;
        }

        return Ok(());
    }
}

impl MethodSignature {
    pub fn new(name: &str, return_type: &str) -> Self {
        MethodSignature {
            name: String::from(name),
            return_type: String::from(return_type),
            headers: vec![],
            modifiers: vec![],
            parameters: vec![],
        }
    }

    pub fn new_header(&mut self, name: &str, value: &str,
                      option: Option<&str>, is_value_string: bool) -> &mut Self {
        let header = Header::new(name, value, option, is_value_string);
        self.headers.push(header);
        return self;
    }

    pub fn new_header_with_option(&mut self, name: &str, value: &str,
                                  option: Option<&str>, is_value_string: bool) -> &mut Self {
        let header = Header::new(name, value, option, is_value_string);
        self.headers.push(header);
        return self;
    }

    pub fn new_modifier(&mut self, name: &str) -> &mut Self {
        let modifier = String::from(name);
        self.modifiers.push(modifier);
        return self;
    }

    pub fn new_parameter(&mut self, name: &str, ty: &str) -> &mut Self {
        let parameter = Parameter::new(name, ty);
        self.parameters.push(parameter);
        return self;
    }

    // Format into code
    pub fn to_code(&self, formatter: &mut Formatter) -> fmt::Result {

        // Headers
        for header in &self.headers {
            header.to_code(formatter)?;
            write!(formatter, "\n")?;
        }

        // Modifiers
        for modifier in &self.modifiers {
            write!(formatter, "{} ", modifier)?;
        }

        // Return type
        write!(formatter, "{} ", self.return_type)?;

        // Name
        write!(formatter, "{}", self.name)?;

        // Parameters
        write!(formatter, "(")?;
        if !self.parameters.is_empty() {
            for (i, parameter) in self.parameters.iter().enumerate() {

                parameter.to_code(formatter)?;

                if i != (self.parameters.len() - 1)
                {
                    write!(formatter, ", ")?;
                }
            }
        }
        write!(formatter, ");")?;

        return Ok(());
    }
}

impl Method{
    pub fn new(name: &str, return_type: &str) -> Self {
        Method{
            name: String::from(name),
            return_type: String::from(return_type),
            modifiers: vec![],
            parameters: vec![],
            body: vec![],
        }
    }

    pub fn new_modifier(&mut self, name: &str) -> &mut Self {
        let modifier = String::from(name);
        self.modifiers.push(modifier);
        return self;
    }

    pub fn new_parameter(&mut self, name: &str, ty: &str) -> &mut Self {
        let parameter = Parameter::new(name, ty);
        self.parameters.push(parameter);
        return self;
    }

    pub fn add_body_raw(&mut self, raw: &str) -> &mut Self {
        let raw_body = String::from(raw);
        self.body.push(raw_body);
        return self;
    }

    // Format into code
    pub fn to_code(&self, formatter: &mut Formatter) -> fmt::Result {

        // Modifiers
        for modifier in &self.modifiers {
            write!(formatter, "{} ", modifier)?;
        }

        // Return type
        write!(formatter, "{} ", self.return_type)?;

        // Name
        write!(formatter, "{}", self.name)?;

        // Parameters
        write!(formatter, "(")?;
        if !self.parameters.is_empty() {
            for (i, parameter) in self.parameters.iter().enumerate() {

                parameter.to_code(formatter)?;

                if i != (self.parameters.len() - 1)
                {
                    write!(formatter, ", ")?;
                }
            }
        }
        write!(formatter, ")")?;

        // Body
        if self.body.is_empty() {
            write!(formatter, " {{}}")?;
        } else {
            write!(formatter, "\n{{\n")?;
            formatter.indentation(| format | {
                for part in &self.body {
                    write!(format, "{}\n", part)?;
                }
                return Ok(())
            })?;
            write!(formatter, "}}")?;
        }

        return Ok(());
    }
}


// Formatting into C# source code
#[derive(Debug)]
pub struct Formatter<'a> {
    code: &'a mut String,
    indent: usize,
    spaces: usize,
}

impl<'a> Formatter<'a> {
    pub fn new(code: &'a mut String) -> Self {
        Formatter {
            code,
            indent: 4, // Default indent is 4 spaces
            spaces: 0, // Current number of spaces i.e. indent level
        }
    }

    fn is_start_of_line(&self) -> bool {
        return self.code.is_empty() || self.code.as_bytes().last() == Some(&b'\n');
    }

    fn fill_with_spaces(&mut self) {
        for _ in 0..self.spaces {
            self.code.push_str(" ");
        }
    }

    fn indentation<F, R>(&mut self, f: F) -> R
        where F: FnOnce(&mut Self) -> R
    {
        self.spaces += self.indent;
        let indented = f(self);
        self.spaces -= self.indent;

        return indented;
    }
}

impl<'a> fmt::Write for Formatter<'a> {
    fn write_str(&mut self, string: &str) -> fmt::Result {

        let mut first_line = true;
        let mut needs_indent = self.is_start_of_line();

        for line in string.lines() {

            if !first_line {
                self.code.push_str("\n");
            }
            first_line = false;


            if needs_indent &&
                !line.is_empty() &&
                line.as_bytes()[0] != b'\n'
            {
                self.fill_with_spaces();
            }
            needs_indent = true;

            self.code.push_str(line);
        }

        if string.as_bytes().last() == Some(&b'\n') {
            self.code.push_str("\n");
        }

        return Ok(());
    }
}
