use crate::inner::Inner;

use std::mem;
use std::sync;
use std::sync::atomic;
use std::sync::atomic::AtomicPtr;
use std::cell;
use std::marker::PhantomData;
use std::iter::FromIterator;
use std::borrow::Borrow;

// A handle that may be used to read from the eventually consistent vector
pub struct ReadHandle<K, V>
where
    K: Eq + Clone,
    V: Clone,
{
    pub inner: sync::Arc<AtomicPtr<Inner<K, V>>>,
    epoch: sync::Arc<sync::atomic::AtomicUsize>,
    my_epoch: sync::atomic::AtomicUsize,

    _not_sync_no_feature: PhantomData<cell::Cell<()>>,
}

impl<K, V> Clone for ReadHandle<K, V>
where
    K: Eq + Clone,
    V: Clone,
{
    fn clone(&self) -> Self {
        let epoch = sync::Arc::new(atomic::AtomicUsize::new(0));
        self.with_handle(|inner| {
            inner.register_epoch(&epoch);
        });
        ReadHandle {
            epoch: epoch,
            my_epoch: atomic::AtomicUsize::new(0),
            inner: sync::Arc::clone(&self.inner),
            _not_sync_no_feature: PhantomData,
        }
    }
}

pub fn new<K, V>(inner: Inner<K, V>) -> ReadHandle<K, V>
where
    K: Eq + Clone,
    V: Clone,
{
    // Tell writer about the epoch tracker
    let epoch = sync::Arc::new(atomic::AtomicUsize::new(0));
    inner.register_epoch(&epoch);

    let store = Box::into_raw(Box::new(inner));
    ReadHandle {
        epoch: epoch,
        my_epoch: atomic::AtomicUsize::new(0),
        inner: sync::Arc::new(AtomicPtr::new(store)),
        _not_sync_no_feature: PhantomData,
    }
}

impl<K, V> ReadHandle<K, V>
where
    K: Eq + Clone,
    V: Clone,
{
    fn with_handle<F, T>(&self, f: F) -> T
    where
        F: FnOnce(&Inner<K, V>) -> T,
    {
        // Update our epoch tracker.
        let epoch = self.my_epoch.fetch_add(1, atomic::Ordering::Relaxed);
        self.epoch.store(epoch + 1, atomic::Ordering::Release);

        // Ensure that the pointer read happens strictly after updating the epoch
        atomic::fence(atomic::Ordering::SeqCst);

        // Then, atomically read pointer, and use the vector being pointed to
        let r_handle = self.inner.load(atomic::Ordering::Acquire);
        let rs = unsafe { Box::from_raw(r_handle) };
        let res = f(&*rs);
        mem::forget(rs); // don't free the Box!

        // We've finished reading -- let the writer know
        self.epoch.store(
            (epoch + 1) | 1usize << (mem::size_of::<usize>() * 8 - 1),
            atomic::Ordering::Release,
        );

        res
    }

    // Returns the number of non-empty keys present in the map.
    pub fn len(&self) -> usize {
        self.with_handle(|inner| inner.data.len())
    }

    // Returns true if the map contains no elements.
    pub fn is_empty(&self) -> bool {
        self.with_handle(|inner| inner.data.is_empty())
    }

    // Returns the first element of the vector
    pub fn first(&self) -> Option<(K, V)>
    {
        self.with_handle(|inner| {
            let value = inner.data.first();
            if value.is_some() {
                return Some(value.unwrap().clone());
            }
            else {
                None
            }
        })
    }

    // Returns the last element of the vector
    pub fn last(&self) -> Option<(K, V)>
    {
        self.with_handle(|inner| {
            let value = inner.data.last();
            if value.is_some() {
                return Some(value.unwrap().clone());
            }
            else {
                None
            }
        })
    }

    // Returns the requested element of the vector
    pub fn get(&self, key: K) -> Option<(K, V)>
    {
        self.with_handle(move |inner| {
            inner.data.iter().find(|x| x.0 == key).cloned()
        })
    }

    pub fn get_and<Q: ?Sized, F, T>(&self, key: K, then: F) -> Option<T>
        where
            F: FnOnce(&(K,V)) -> T,
            K: Borrow<Q>,
            Q: Eq,
    {
        self.with_handle(move |inner| if !inner.is_ready() {
            None
        } else {
            let index = inner.data.iter().position(|x| x.0 == key);
            if index.is_some() {
                return Some(then(inner.data[index.unwrap()].borrow()));
            }
            else {
                return None;
            }
        })
    }

    /// Returns true if the vector contains any values for the specified key.
    ///
    /// The key may be any borrowed form of the map's key type, but `Hash` and `Eq` on the borrowed
    /// form *must* match those for the key type.
    pub fn contains_key(&self, key: K) -> bool
    {
        self.with_handle(move |inner| {
            let index = inner.data.iter().position(|x| x.0 == key);
            return index.is_some();
        })
    }


    // Read all values in the vector, and transform them into a new collection.
    // Be careful with this function! While the iteration is ongoing, any writer that tries to
    // refresh will block waiting on this reader to finish.
    pub fn for_each<F>(&self, mut f: F)
    where
        F: FnMut(&K, &V),
    {
        self.with_handle(move |inner| for (k, vs) in inner.data.iter() {
            f(k, vs)
        })
    }

    /// Read all values in the vector, and transform them into a new collection.
    pub fn map_into<Map, Collector, Target>(&self, mut f: Map) -> Collector
    where
        Map: FnMut(&K, &V) -> Target,
        Collector: FromIterator<Target>,
    {
        self.with_handle(move |inner| {
            Collector::from_iter(inner.data.iter().map(|(k, vs)| f(k, vs)))
        })
    }

    pub fn take_into<Map, Collector, Target>(&self, mut f: Map, count: usize) -> Collector
        where
            Map: FnMut(&K, &V) -> Target,
            Collector: FromIterator<Target>,
    {
        self.with_handle(move |inner| {
            Collector::from_iter(inner.data.iter().take(count).map(|(k, vs)| f(k, vs)))
        })
    }

    pub fn rev_take_into<Map, Collector, Target>(&self, mut f: Map, count: usize) -> Collector
        where
            Map: FnMut(&K, &V) -> Target,
            Collector: FromIterator<Target>,
    {
        self.with_handle(move |inner| {
            Collector::from_iter(inner.data.iter().rev().take(count).map(|(k, vs)| f(k, vs)))
        })
    }
}
