use super::Operation;
use crate::inner::Inner;
use crate::read::ReadHandle;

use std::sync::Arc;
use std::sync::atomic;


pub struct WriteHandle<K, V>
where
    K: Eq + Clone,
    V: Clone,
{
    w_handle: Option<Box<Inner<K, V>>>,
    oplog: Vec<Operation<K, V>>,
    swap_index: usize,
    r_handle: ReadHandle<K, V>,
    last_epochs: Vec<usize>,
    first: bool,
    second: bool,
}

pub fn new<K, V>(
    w_handle: Inner<K, V>,
    r_handle: ReadHandle<K, V>,
) -> WriteHandle<K, V>
where
    K: Eq + Clone,
    V: Clone,
{
    WriteHandle {
        w_handle: Some(Box::new(w_handle)),
        oplog: Vec::new(),
        swap_index: 0,
        r_handle: r_handle,
        last_epochs: Vec::new(),
        first: true,
        second: false,
    }
}

impl<K, V> WriteHandle<K, V>
where
    K: Eq + Clone,
    V: Eq + Clone,
{
    /// Refresh the handle used by readers so that pending writes are made visible.
    pub fn refresh(&mut self) {
        use std::thread;
        use std::mem;

        let epochs = Arc::clone(&self.w_handle.as_ref().unwrap().epochs);
        let epochs = epochs.lock().unwrap();
        self.last_epochs.resize(epochs.len(), 0);

        let mut iter = 0;
        let mut starti = 0;
        let high_bit = 1usize << (mem::size_of::<usize>() * 8 - 1);
        'retry: loop {
            // Read all and see if all have changed (which is likely)
            for (i, epoch) in epochs.iter().enumerate().skip(starti) {
                if self.last_epochs[i] & high_bit != 0 {
                    // Reader was not active right after last swap
                    // And therefore *must* only see new pointer
                    continue;
                }

                let now = epoch.load(atomic::Ordering::Acquire);
                if (now != self.last_epochs[i]) | (now & high_bit != 0) | (now == 0) {
                    // Reader must have seen last swap
                } else {
                    // Reader may not have seen swap
                    // Continue from this reader's epoch
                    starti = i;

                    // How eagerly should we retry?
                    if iter != 20 {
                        iter += 1;
                    } else {
                        thread::yield_now();
                    }

                    continue 'retry;
                }
            }

            // All the readers have left!
            // We can safely bring the w_handle up to date.
            let w_handle = self.w_handle.as_mut().unwrap();

            if self.second {
                let r_handle =
                    unsafe { Box::from_raw(self.r_handle.inner.load(atomic::Ordering::Relaxed)) };
                w_handle.data = r_handle.data.clone();
                mem::forget(r_handle);
            }

            // The w_handle map has not seen any of the writes in the oplog
            // The r_handle map has not seen any of the writes following swap_index
            if self.swap_index != 0 {
                // We can drain out the operations that only the w_handle map needs
                // NOTE: the if above is because drain(0..0) would remove 0
                for op in self.oplog.drain(0..self.swap_index) {
                    Self::apply_op(w_handle, op);
                }
            }
            // The rest have to be cloned because they'll also be needed by the r_handle map
            for op in self.oplog.iter().cloned() {
                Self::apply_op(w_handle, op);
            }
            // The w_handle map is about to become the r_handle, and can ignore the oplog
            self.swap_index = self.oplog.len();

            w_handle.mark_ready();

            // w_handle (the old r_handle) is now fully up to date!
            break;
        }

        // Prepare w_handle
        let w_handle = self.w_handle.take().unwrap();
        let w_handle = Box::into_raw(w_handle);

        // Swap in our w_handle, and get r_handle in return
        let r_handle = self.r_handle
            .inner
            .swap(w_handle, atomic::Ordering::Release);
        let r_handle = unsafe { Box::from_raw(r_handle) };

        // Ensure that the subsequent epoch reads aren't re-ordered to before the swap
        atomic::fence(atomic::Ordering::SeqCst);

        for (i, epoch) in epochs.iter().enumerate() {
            self.last_epochs[i] = epoch.load(atomic::Ordering::Acquire);
        }

        // NOTE: at this point, there are likely still readers using the w_handle we got
        self.w_handle = Some(r_handle);
        self.second = self.first;
        self.first = false;
    }

    fn add_op(&mut self, op: Operation<K, V>) {
        if !self.first {
            self.oplog.push(op);
        } else {
            // We know there are no outstanding w_handle readers, so we can modify it directly!
            let inner = self.w_handle.as_mut().unwrap();
            Self::apply_op(inner, op);
            // NOTE: since we didn't record this in the oplog, r_handle *must* clone w_handle
        }
    }

    /// Add the given value to the value-set of the given key.
    ///
    /// The updated value-set will only be visible to readers after the next call to `refresh()`.
    pub fn push(&mut self, k: K, v: V) {
        self.add_op(Operation::Push(k, v));
    }

    /// Replace the value-set of the given key with the given value.
    ///
    /// The new value will only be visible to readers after the next call to `refresh()`.
    pub fn update(&mut self, k: K, v: V) {
        self.add_op(Operation::Replace(k, v));
    }

    /// Clear the value-set of the given key, without removing it.
    ///
    /// The new value will only be visible to readers after the next call to `refresh()`.
    pub fn clear(&mut self, k: K) {
        self.add_op(Operation::Clear(k));
    }

    /// Remove the given value from the value-set of the given key.
    ///
    /// The updated value-set will only be visible to readers after the next call to `refresh()`.
    pub fn remove(&mut self, k: K, v: V) {
        self.add_op(Operation::Remove(k, v));
    }

    fn apply_op(inner: &mut Inner<K, V>, op: Operation<K, V>) {
        match op {
            Operation::Replace(key, value) => {
                let index = inner.data.iter().position(|x| x.0 == key);
                if index.is_some() {
                    inner.data[index.unwrap()] = (key, value);
                }
                else {
                    inner.data.push((key, value));
                }
            }
            Operation::Clear(key) => {
                inner.data.retain(|x| x.0 != key);
            }
            Operation::Push(key, value) => {
                inner.data.push((key, value));
            }
            Operation::Remove(key, _value) => {
                let index = inner.data.iter().position(|x| x.0 == key);
                if index.is_some() {
                    inner.data.remove(index.unwrap());
                }
            }
        }
    }
}

impl<K, V> Extend<(K, V)> for WriteHandle<K, V>
where
    K: Eq + Clone,
    V: Eq + Clone,
{
    fn extend<I: IntoIterator<Item = (K, V)>>(&mut self, iter: I) {
        for (k, v) in iter {
            self.push(k, v);
        }
    }
}

// Allow using write handle for reads
use std::ops::Deref;
impl<K, V> Deref for WriteHandle<K, V>
where
    K: Eq + Clone,
    V: Eq + Clone,
{
    type Target = ReadHandle<K, V>;
    fn deref(&self) -> &Self::Target {
        &self.r_handle
    }
}
