extern crate ecvector;

#[test]
fn it_works() {
    let x = ('x', 42);

    let (r, mut w) = ecvector::new();

    // the map is uninitialized, so all lookups should return None
    assert_eq!(r.first(), None);
    assert_eq!(r.last(), None);

    w.refresh();

    // after the first refresh, it is empty, but ready
    assert_eq!(r.first(), None);
    assert_eq!(r.last(), None);

    w.push(x.0, x.1);

    // it is empty even after an add (we haven't refresh yet)
    assert_eq!(r.first(), None);
    assert_eq!(r.last(), None);

    w.refresh();

    // but after the swap, the record is there!
    assert_eq!(r.first().unwrap(), x);
    assert_eq!(r.last().unwrap(), x);

    // non-existing records return None
    assert_eq!(r.contains_key(x.0), true);
}

#[test]
fn threads_fast() {
    threads_inner(false);
}
#[test]
fn threads_slow() {
    threads_inner(true);
}

fn threads_inner(slow: bool) {
    use std::time;
    use std::thread;

    let threads = 4;
    let mut n = 1000;
    if !slow {
        n *= 100;
    }
    let (r, mut w) = ecvector::new();

    let rs: Vec<_> = (0..threads)
        .map(|_| {
            let r = r.clone();
            thread::spawn(move || {
                // rustfmt
                for i in 0..n {
                    let i = i.into();
                    loop {
                        match r.get_and(i, |rs| {
                            if slow {
                                thread::sleep(time::Duration::from_millis(2));
                            }
                            rs.clone()
                        }) {
                            Some(rs) => {
                                assert_eq!(rs.0, i);
                                assert_eq!(rs.1, i);
                                break;
                            }
                            None => {
                                thread::yield_now();
                            }
                        }
                    }
                }
            })
        })
        .collect();

    for i in 0..n {
        w.push(i, i);
        w.refresh();
    }

    for r in rs {
        r.join().unwrap();
    }
}


#[test]
fn minimal_query() {
    let (r, mut w) = ecvector::new();

    w.push(1, "a");
    w.refresh();
    w.push(1, "b");

    assert_eq!(r.get_and(1, |rs| rs.0 ), Some(1));
    assert_eq!(r.get_and(1, |rs| rs.1 ), Some("a"));
}

#[test]
fn empty() {
    let (r, mut w) = ecvector::new();
    w.push(1, "a");
    w.push(1, "b");
    w.push(2, "c");
    w.clear(1);
    w.refresh();

    assert_eq!(r.get_and(1, |rs| rs.1), None);
    assert_eq!(r.get_and(2, |rs| rs.1), Some("c"));
}

/*
#[test]
fn empty_post_refresh() {
    let (r, mut w) = evmap::new();
    w.insert(1, "a");
    w.insert(1, "b");
    w.insert(2, "c");
    w.refresh();
    w.empty(1);
    w.refresh();

    assert_eq!(r.get_and(&1, |rs| rs.len()), None);
    assert_eq!(r.get_and(&2, |rs| rs.len()), Some(1));
    assert!(r.get_and(&2, |rs| rs.iter().any(|r| r == &"c")).unwrap());
}

#[test]
fn clear() {
    let (r, mut w) = evmap::new();
    w.insert(1, "a");
    w.insert(1, "b");
    w.insert(2, "c");
    w.clear(1);
    w.refresh();

    assert_eq!(r.get_and(&1, |rs| rs.len()), Some(0));
    assert_eq!(r.get_and(&2, |rs| rs.len()), Some(1));

    w.clear(2);
    w.refresh();

    assert_eq!(r.get_and(&1, |rs| rs.len()), Some(0));
    assert_eq!(r.get_and(&2, |rs| rs.len()), Some(0));

    w.empty(1);
    w.refresh();

    assert_eq!(r.get_and(&1, |rs| rs.len()), None);
    assert_eq!(r.get_and(&2, |rs| rs.len()), Some(0));
}

#[test]
fn replace() {
    let (r, mut w) = evmap::new();
    w.insert(1, "a");
    w.insert(1, "b");
    w.insert(2, "c");
    w.update(1, "x");
    w.refresh();

    assert_eq!(r.get_and(&1, |rs| rs.len()), Some(1));
    assert!(r.get_and(&1, |rs| rs.iter().any(|r| r == &"x")).unwrap());
    assert_eq!(r.get_and(&2, |rs| rs.len()), Some(1));
    assert!(r.get_and(&2, |rs| rs.iter().any(|r| r == &"c")).unwrap());
}

#[test]
fn replace_post_refresh() {
    let (r, mut w) = evmap::new();
    w.insert(1, "a");
    w.insert(1, "b");
    w.insert(2, "c");
    w.refresh();
    w.update(1, "x");
    w.refresh();

    assert_eq!(r.get_and(&1, |rs| rs.len()), Some(1));
    assert!(r.get_and(&1, |rs| rs.iter().any(|r| r == &"x")).unwrap());
    assert_eq!(r.get_and(&2, |rs| rs.len()), Some(1));
    assert!(r.get_and(&2, |rs| rs.iter().any(|r| r == &"c")).unwrap());
}

#[test]
fn map_into() {
    let (r, mut w) = evmap::new();
    w.insert(1, "a");
    w.insert(1, "b");
    w.insert(2, "c");
    w.refresh();
    w.insert(1, "x");

    use std::collections::HashMap;
    let copy: HashMap<_, Vec<_>> = r.map_into(|&k, vs| (k, vs.to_vec()));

    assert_eq!(copy.len(), 2);
    assert!(copy.contains_key(&1));
    assert!(copy.contains_key(&2));
    assert_eq!(copy[&1].len(), 2);
    assert_eq!(copy[&2].len(), 1);
    assert_eq!(copy[&1], vec!["a", "b"]);
    assert_eq!(copy[&2], vec!["c"]);
}

#[test]
fn foreach() {
    let (r, mut w) = evmap::new();
    w.insert(1, "a");
    w.insert(1, "b");
    w.insert(2, "c");
    w.refresh();
    w.insert(1, "x");

    r.for_each(|&k, vs| match k {
        1 => assert_eq!(vs, &*vec!["a", "b"]),
        2 => assert_eq!(vs, &*vec!["c"]),
        _ => unreachable!(),
    });
}
*/