use std::time::Duration;

#[derive(Clone, Copy, Debug)]
pub enum IntervalType {
    Open,
    RightOpen,
    LeftOpen,
    Closed
}

#[derive(Clone, Copy, Debug)]
pub enum IntervalRelation {
    Precedes,
    Meets,
    Overlaps,
    FinishedBy,
    Contains,
    Starts,
    Equals,
    StartedBy,
    During,
    Finishes,
    OverlappedBy,
    MetBy,
    PrecededBy
}

// The reference time point is the start of the system (0 timestamp)
// The time axis looks like this
// Now(10)...Past(9)...Past(8)...
#[derive(Clone, Copy, Debug)]
pub struct Interval {
    start: u64,
    end: u64,
    ty: IntervalType
}
impl Interval {
    // Basic factory methods
    pub fn from_duration(start: u64, duration: Duration) -> Interval {
        Interval {
            start,
            end: start + duration.as_millis() as u64,
            ty: IntervalType::Closed
        }
    }

    pub fn from_eternity_to(to: u64) -> Interval {
        Interval {
            start: 0,
            end: to,
            ty: IntervalType::Closed
        }
    }

    // Keep in mind that start timestamp is always lower than end timestamp
    pub fn new(start: u64, end: u64) -> Interval {
        Interval {
            start,
            end,
            ty: IntervalType::Closed
        }
    }

    pub fn eternity() -> Interval {
        Interval {
            start: 0,
            end: u64::max_value(),
            ty: IntervalType::Closed
        }
    }

    pub fn of_type(mut self, ty: IntervalType) -> Self{
        self.ty = ty;
        self
    }

    // Additional factory methods
    pub fn new_precedes(&self, interval: Interval, duration: Duration, gap: Duration) -> Interval {
        Interval {
            start: interval.start - gap.as_millis() as u64 - duration.as_millis() as u64,
            end: interval.start - gap.as_millis() as u64,
            ty: IntervalType::Closed
        }
    }

    pub fn new_meets(&self, interval: Interval, duration: Duration) -> Interval {
        Interval {
            start: interval.start - duration.as_millis() as u64,
            end: interval.start,
            ty: IntervalType::Closed
        }
    }

    pub fn new_finished_by(&self, interval: Interval, duration: Duration) -> Interval {
        Interval {
            start: interval.end - duration.as_millis() as u64,
            end: interval.end,
            ty: IntervalType::Closed
        }
    }

    pub fn new_starts(&self, interval: Interval, duration: Duration) -> Interval {
        Interval {
            start: interval.start,
            end: interval.start + duration.as_millis() as u64,
            ty: IntervalType::Closed
        }
    }

    pub fn new_equals(&self, interval: Interval) -> Interval {
        Interval {
            start: interval.start,
            end: interval.end,
            ty: IntervalType::Closed
        }
    }

    pub fn new_started_by(&self, interval: Interval, duration: Duration) -> Interval {
        Interval {
            start: interval.start,
            end: interval.start + duration.as_millis() as u64,
            ty: IntervalType::Closed
        }
    }

    pub fn new_finishes(&self, interval: Interval, duration: Duration) -> Interval {
        Interval {
            start: interval.end - duration.as_millis() as u64,
            end: interval.end,
            ty: IntervalType::Closed
        }
    }

    pub fn new_met_by(&self, interval: Interval, duration: Duration) -> Interval {
        Interval {
            start: interval.end,
            end: interval.end + duration.as_millis() as u64,
            ty: IntervalType::Closed
        }
    }

    pub fn new_preceded_by(&self, interval: Interval, duration: Duration, gap: Duration) -> Interval {
        Interval {
            start: interval.end + gap.as_millis() as u64,
            end: interval.end + gap.as_millis() as u64 + duration.as_millis() as u64,
            ty: IntervalType::Closed
        }
    }

    // Interval relations
    pub fn preceded_by(&self, other: Interval) -> bool {
        self.start < other.start &&
            self.start < other.end &&

            self.end < other.start &&
            self.end < other.end
    }

    pub fn met_by(&self, other: Interval) -> bool {
        self.start < other.start &&
            self.start < other.end &&

            self.end == other.start &&
            self.end < other.end
    }

    pub fn overlapped_by(&self, other: Interval) -> bool {
        self.start < other.start &&
            self.start < other.end &&

            self.end > other.start &&
            self.end < other.end
    }

    pub fn finishes(&self, other: Interval) -> bool {
        self.start < other.start &&
            self.start < other.end &&

            self.end > other.start &&
            self.end == other.end
    }

    pub fn during(&self, other: Interval) -> bool {
        self.start < other.start &&
            self.start < other.end &&

            self.end > other.start &&
            self.end > other.end
    }

    pub fn started_by(&self, other: Interval) -> bool {
        self.start == other.start &&
            self.start < other.end &&

            self.end > other.start &&
            self.end < other.end
    }

    pub fn equals(&self, other: Interval) -> bool {
        self.start == other.start &&
            self.start < other.end &&

            self.end > other.start &&
            self.end == other.end
    }

    pub fn starts(&self, other: Interval) -> bool {
        self.start == other.start &&
            self.start < other.end &&

            self.end > other.start &&
            self.end > other.end
    }

    pub fn contains(&self, other: Interval) -> bool {
        self.start > other.start &&
            self.start < other.end &&

            self.end > other.start &&
            self.end < other.end
    }

    pub fn finished_by(&self, other: Interval) -> bool {
        self.start > other.start &&
            self.start < other.end &&

            self.end > other.start &&
            self.end == other.end
    }

    pub fn overlaps(&self, other: Interval) -> bool {
        self.start > other.start &&
            self.start < other.end &&

            self.end > other.start &&
            self.end > other.end
    }

    pub fn meets(&self, other: Interval) -> bool {
        self.start > other.start &&
            self.start == other.end &&

            self.end > other.start &&
            self.end > other.end
    }

    pub fn precedes(&self, other: Interval) -> bool {
        self.start > other.start &&
            self.start > other.end &&

            self.end > other.start &&
            self.end > other.end
    }

    // Inclusion rules
    pub fn includes(&self, value: u64) -> bool {
        match self.ty {
            IntervalType::Open => { value > self.start && value < self.end }
            IntervalType::LeftOpen => { value > self.start && value <= self.end }
            IntervalType::RightOpen => { value >= self.start && value < self.end }
            IntervalType::Closed => { value >= self.start && value <= self.end }
        }
    }
}
