use derivative::*;
use phf::phf_map;

pub const KNOWLEDGE_GRAPH_NODE_TYPE_COUNT: usize = 17;
pub const KNOWLEDGE_GRAPH_RELATION_TYPE_COUNT: usize = 16;

#[derive(Clone, Copy, PartialEq, Default, Debug)]
#[repr(C)]
pub struct Vector3f {
    pub x: f32,
    pub y: f32,
    pub z: f32,
}

#[derive(Clone, Copy, PartialEq, Default, Debug)]
#[repr(C)]
pub struct Vector2f {
    pub x: f32,
    pub y: f32,
}

#[derive(Clone, Copy)]
#[repr(C)]
pub struct RawString {
    pub raw: [u8; 128],
}

impl Default for RawString {
    fn default() -> Self {
        return RawString { raw: [0; 128] };
    }
}

#[derive(Clone, Copy, PartialEq, Debug)]
#[repr(usize)]
pub enum KnowledgeGraphNode {
    Empty,
    Car(Car),
    Crosswalk(Crosswalk),
    DynamicObject(DynamicObject),
    Ego(Ego),
    Intersection(Intersection),
    Lane(Lane),
    NoEntrySign(NoEntrySign),
    Pedestrian(Pedestrian),
    PriorityRoadSign(PriorityRoadSign),
    Road(Road),
    RoadSign(RoadSign),
    SpeedLimitSign(SpeedLimitSign),
    StaticObject(StaticObject),
    StopSign(StopSign),
    TestClass1(TestClass1),
    TestClass2(TestClass2),
    YieldSign(YieldSign),
}

impl Default for KnowledgeGraphNode {
    fn default() -> Self {
        return KnowledgeGraphNode::Empty;
    }
}

impl KnowledgeGraphNode {
    pub fn type_index(&self) -> usize {
        return unsafe { *(self as *const Self as *const usize) };
    }
}

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct Car {
    pub carDirection: Vector3f,
    pub carId: i32,
    pub carPosition: Vector3f,
    pub carSpeed: f32,
}

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct Crosswalk {
    pub crosswalkId: i32,
    pub crosswalkPedestrianCount: i32,
    pub crosswalkPosition: Vector3f,
}

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct DynamicObject;

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct Ego;

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct Intersection {
    pub intersectionId: i32,
    pub intersectionPosition: Vector3f,
}

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct Lane {
    pub laneId: i32,
    #[derivative(Debug="ignore",PartialEq="ignore")]
    pub laneName: RawString,
    pub laneSpeedLimit: i32,
}

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct NoEntrySign;

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct Pedestrian {
    pub pedestrianDirection: Vector2f,
    pub pedestrianPosition: Vector2f,
    pub pedestrianSpeed: f32,
}

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct PriorityRoadSign;

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct Road {
    pub roadId: i32,
    #[derivative(Debug="ignore",PartialEq="ignore")]
    pub roadName: RawString,
}

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct RoadSign;

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct SpeedLimitSign {
    pub signSpeedLimit: i32,
}

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct StaticObject;

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct StopSign;

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct TestClass1 {
    pub testProperty1: i32,
}

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct TestClass2 {
    pub testProperty2: i32,
}

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct YieldSign;



#[derive(Clone, Copy, PartialEq, Debug)]
#[repr(usize)]
pub enum KnowledgeGraphRelation {
    Empty,
    belongsTo(belongsTo),
    connectsTo(connectsTo),
    contains(contains),
    has(has),
    isAt(isAt),
    isBehind(isBehind),
    isIn(isIn),
    isInFront(isInFront),
    isOn(isOn),
    isOnLeft(isOnLeft),
    isOnRight(isOnRight),
    isPartOf(isPartOf),
    passesOver(passesOver),
    testRelation(testRelation),
    walksOn(walksOn),
    yieldsToCarIn(yieldsToCarIn),
}

impl Default for KnowledgeGraphRelation {
    fn default() -> Self {
        return KnowledgeGraphRelation::Empty;
    }
}

impl KnowledgeGraphRelation {
    pub fn type_index(&self) -> usize {
        return unsafe { *(self as *const Self as *const usize) };
    }
}

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct belongsTo;

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct connectsTo;

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct contains;

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct has;

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct isAt;

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct isBehind;

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct isIn;

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct isInFront;

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct isOn;

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct isOnLeft;

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct isOnRight;

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct isPartOf;

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct passesOver;

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct testRelation {
    pub testProperty: f32,
}

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct walksOn;

#[derive(Clone, Copy, Default)]
#[derive(Derivative)]
#[derivative(Debug, PartialEq)]
#[repr(C)]
pub struct yieldsToCarIn {
    pub ifDirection: i32,
}

static SUBCLASSES: phf::Map<&'static str, &'static str> = phf_map! {

"Car" => "DynamicObject",

"Crosswalk" => "StaticObject",

"DynamicObject" => "",

"Ego" => "DynamicObject",

"Intersection" => "StaticObject",

"Lane" => "StaticObject",

"NoEntrySign" => "RoadSign",

"Pedestrian" => "DynamicObject",

"PriorityRoadSign" => "RoadSign",

"Road" => "StaticObject",

"RoadSign" => "StaticObject",

"SpeedLimitSign" => "RoadSign",

"StaticObject" => "",

"StopSign" => "RoadSign",

"TestClass1" => "",

"TestClass2" => "",

"YieldSign" => "RoadSign",

};

static RELATIONS: phf::Map<&'static str, [&'static str; 10]> = phf_map! {

"Car" => ["Ego","Ego","Ego","Ego","","","","","","",],

"Crosswalk" => ["","","","","","","","","","",],

"DynamicObject" => ["","","","","","","","","","",],

"Ego" => ["Intersection","Lane","Road","Crosswalk","","","","","","",],

"Intersection" => ["Lane","","","","","","","","","",],

"Lane" => ["Lane","Road","Lane","","","","","","","",],

"NoEntrySign" => ["","","","","","","","","","",],

"Pedestrian" => ["Lane","","","","","","","","","",],

"PriorityRoadSign" => ["","","","","","","","","","",],

"Road" => ["Lane","","","","","","","","","",],

"RoadSign" => ["Lane","","","","","","","","","",],

"SpeedLimitSign" => ["","","","","","","","","","",],

"StaticObject" => ["","","","","","","","","","",],

"StopSign" => ["","","","","","","","","","",],

"TestClass1" => ["TestClass2","","","","","","","","","",],

"TestClass2" => ["","","","","","","","","","",],

"YieldSign" => ["","","","","","","","","","",],

};