use std::collections::BTreeMap;
use std::sync::atomic::{AtomicUsize, Ordering, AtomicBool};
use std::alloc::{alloc, Layout};

use interval_algebra::Interval;

pub mod types;
use self::types::*;

pub mod iterators;
use self::iterators::*;

use config::*;


pub type ValueId = usize;
pub type InstanceId = usize;
pub type TypeIndex = usize;
pub type Timestamp = u64;

pub type NodeIndexRecord = (Timestamp,
                            ValueId,
                            InstanceId,
                            [Index; KNOWLEDGE_GRAPH_MAX_RELATIONS],
                            [Index; KNOWLEDGE_GRAPH_MAX_RELATIONS],
                            Marker);

pub type RelationIndexRecord = (Timestamp,
                                ValueId,
                                InstanceId,
                                Index,
                                Index,
                                Marker);

// Helper types for graph
#[derive(Clone, Copy, PartialEq, Debug)]
#[repr(C)]
pub struct Index {
    pub type_index: TypeIndex,
    pub instance_id: InstanceId,
}
impl Index {
    pub fn new(type_index: TypeIndex, instance_id: InstanceId) -> Self {
        Index {
            type_index,
            instance_id
        }
    }

    pub fn zeroed() -> Self {
        Index::new(0, 0)
    }

    pub fn is_empty(&self) -> bool {
        self.instance_id == 0 && self.type_index == 0
    }
}
impl Default for Index {
    fn default() -> Self {
        return Index {
            type_index: 0,
            instance_id: 0
        }
    }
}

#[derive(Clone, Copy, PartialEq, Debug)]
pub enum Marker {
    None,
    Valid,
    Removed
}
impl Default for Marker {
    fn default() -> Self {
        return Marker::None;
    }
}

#[derive(Clone, Debug)]
pub struct GraphIndices {
    node_index_tail: [usize; KNOWLEDGE_GRAPH_NODE_TYPE_COUNT + 1],
    nodes_tail: usize,
    relation_index_tail: [usize; KNOWLEDGE_GRAPH_RELATION_TYPE_COUNT + 1],
    relations_tail: usize,
    update_timestamp: Timestamp,
}
impl Default for GraphIndices {
    fn default() -> Self {
        GraphIndices {
            node_index_tail: [0; KNOWLEDGE_GRAPH_NODE_TYPE_COUNT + 1],
            nodes_tail: 0,
            relation_index_tail: [0; KNOWLEDGE_GRAPH_RELATION_TYPE_COUNT + 1],
            relations_tail: 0,
            update_timestamp: 0,
        }
    }
}


// Graph on linear arrays, no pointers == shared memory safe
pub struct KnowledgeGraph {
    // (u64, usize, usize, [usize], usize, [usize], usize, Marker) =>
    // => Timestamp, value id, instance id, from relations, from index, to relations, to index, marker
    node_index: [[NodeIndexRecord;
        KNOWLEDGE_GRAPH_NODE_INDEX_CAPACITY];
        KNOWLEDGE_GRAPH_NODE_TYPE_COUNT + 1],
    nodes: [KnowledgeGraphNode; KNOWLEDGE_GRAPH_NODE_CAPACITY],

    // (u64, usize, usize, usize, usize, Marker)
    // => Timestamp, value id, instance id, from index, to index, marker
    relation_index: [[RelationIndexRecord;
        KNOWLEDGE_GRAPH_RELATION_INDEX_CAPACITY];
        KNOWLEDGE_GRAPH_RELATION_TYPE_COUNT + 1],
    relations: [KnowledgeGraphRelation; KNOWLEDGE_GRAPH_RELATION_CAPACITY],

    indices_buffer: [GraphIndices; 2],
    current_buffer: AtomicUsize,
    autocommit: AtomicBool
}
impl Default for KnowledgeGraph {
    fn default() -> Self {
        KnowledgeGraph {
            // Zeros nodes
            node_index:[[(0, 0, 0,
                          [Index::default(); KNOWLEDGE_GRAPH_MAX_RELATIONS],
                          [Index::default(); KNOWLEDGE_GRAPH_MAX_RELATIONS],
                          Marker::None);
                KNOWLEDGE_GRAPH_NODE_INDEX_CAPACITY];
                KNOWLEDGE_GRAPH_NODE_TYPE_COUNT + 1],
            nodes: [KnowledgeGraphNode::default(); KNOWLEDGE_GRAPH_NODE_CAPACITY],

            // Zero relations
            relation_index: [[(0, 0, 0, Index::default(), Index::default(), Marker::None);
                KNOWLEDGE_GRAPH_RELATION_INDEX_CAPACITY];
                KNOWLEDGE_GRAPH_RELATION_TYPE_COUNT + 1],
            relations: [KnowledgeGraphRelation::default(); KNOWLEDGE_GRAPH_RELATION_CAPACITY],

            // Create buffers for indices
            indices_buffer: [GraphIndices::default(), GraphIndices::default()],
            current_buffer: AtomicUsize::new(0),
            autocommit: AtomicBool::new(true),
        }
    }
}

impl KnowledgeGraph {
    pub fn new_boxed() -> Box<KnowledgeGraph> {
        // Workaround to prevent STACK OVERFLOW while creating this huuuge data structure
        let layout = Layout::new::<KnowledgeGraph>();
        let mut graph = unsafe {
            let ptr = alloc(layout) as *mut KnowledgeGraph;
            std::ptr::write_bytes(ptr, 0, 1);
            Box::from_raw(ptr)
        };

        // Initialize and return
        graph.initialize();
        graph
    }


    // Iterator methods
    pub fn nodes_iter(&self, interval: Interval, type_index: usize) -> NodeIterator {
        let indices = self.load_indices_active();
        NodeIterator::new(self, indices, interval, type_index)
    }

    pub fn relations_iter(&self, interval: Interval, type_index: usize) -> RelationIterator {
        let indices = self.load_indices_active();
        RelationIterator::new(self, indices, interval, type_index)
    }


    // Transaction helpers
    fn load_indices(&self) -> GraphIndices {
        if self.autocommit.load(Ordering::SeqCst) {
            // Load indices from the ACTIVE buffer
            let index = self.current_buffer.load(Ordering::SeqCst);
            self.indices_buffer[index].clone()
        } else {
            // Load indices from the INACTIVE buffer
            let index = 1 - self.current_buffer.load(Ordering::SeqCst);
            self.indices_buffer[index].clone()
        }
    }

    fn load_indices_active(&self) -> GraphIndices {
        let index = self.current_buffer.load(Ordering::SeqCst);
        self.indices_buffer[index].clone()
    }

    fn save_indices(&mut self, indices: GraphIndices) {
        let index = 1 - self.current_buffer.load(Ordering::SeqCst);
        self.indices_buffer[index] = indices;

        // If autocommit, commit transaction
        if self.autocommit.load(Ordering::SeqCst) {
            self.commit_transaction();
        }
    }

    // The start of the transaction clears transaction buffer
    pub fn start_transaction(&mut self) {
        let index = self.current_buffer.load(Ordering::SeqCst);

        // Clone current indices to the inactive buffer
        let current_indices = self.indices_buffer[index].clone();
        self.indices_buffer[1 - index] = current_indices;
    }

    pub fn commit_transaction(&mut self) {
        // Swap the index buffers
        let new_index = 1 - self.current_buffer.load(Ordering::SeqCst);
        self.current_buffer.store(new_index, Ordering::SeqCst);
    }

    pub fn set_autocommit(&mut self, value: bool) {
        self.autocommit.store(value, Ordering::SeqCst);
    }


    // Actual graph methods
    pub fn initialize(&mut self) {
        self.set_autocommit(true);
    }

    pub fn is_empty(&self) -> bool {
        let indices = self.load_indices_active();
        indices.node_index_tail.iter().filter(|v| (**v > 0)).count() == 0
            && indices.relation_index_tail.iter().filter(|v| (**v > 0)).count() == 0
    }

    pub fn last_update(&self) -> Timestamp {
        let indices = self.load_indices_active();
        indices.update_timestamp
    }

    pub fn node_count(&self) -> usize {
        self.load_indices_active().node_index_tail.iter().sum()
    }

    pub fn relation_count(&self) -> usize {
        self.load_indices_active().relation_index_tail.iter().sum()
    }

    pub fn clear(&mut self) {
        // Clear indices
        let mut indices = self.load_indices();

        for value in indices.node_index_tail.iter_mut() {
            *value = 0;
        }
        indices.nodes_tail = 0;

        for value in indices.relation_index_tail.iter_mut() {
            *value = 0;
        }
        indices.relations_tail = 0;

        self.save_indices(indices);
    }


    pub fn add_node(&mut self, timestamp: Timestamp, instance_id: InstanceId,
                    node: KnowledgeGraphNode) -> Index {
        
        let mut indices = self.load_indices();
        let type_index = node.type_index();

        // First add node
        let node_id = indices.nodes_tail;
        self.nodes[node_id] = node;
        indices.nodes_tail += 1;

        // Then index it
        let index_record = (timestamp,
                            node_id,
                            instance_id,
                            [Index::default(); KNOWLEDGE_GRAPH_MAX_RELATIONS],
                            [Index::default(); KNOWLEDGE_GRAPH_MAX_RELATIONS],
                            Marker::Valid);

        let index_id = indices.node_index_tail[type_index];
        self.node_index[type_index][index_id] = index_record;
        indices.node_index_tail[type_index] += 1;

        // Potentially commit transaction
        indices.update_timestamp = timestamp;
        self.save_indices(indices);


        return Index::new(type_index, instance_id);
    }

    pub fn remove_node(&mut self, timestamp: Timestamp, index: Index) -> bool {
        let mut indices = self.load_indices();

        // Fetch last node index record
        let node_record = self.get_latest_node_record(index);

        // If nothing was found return, cannot remove
        if node_record.is_none() { return false; }


        // Clone node record and mark deletion
        let mut new_node_record = node_record.unwrap().clone();
        new_node_record.0 = timestamp;
        new_node_record.5 = Marker::Removed;

        // Remove the connecting relations as well
        let mut relations_to_remove = Vec::new();
        new_node_record.3.iter_mut().for_each(|index| {
            if !index.is_empty() {
                relations_to_remove.push(*index);
            }
        });
        new_node_record.4.iter_mut().for_each(|index| {
            if !index.is_empty() {
                relations_to_remove.push(*index);
            }
        });

        // Insert record about node removal
        let index_id = indices.node_index_tail[index.type_index];
        self.node_index[index.type_index][index_id] = new_node_record;
        indices.node_index_tail[index.type_index] += 1;


        // Potentially commit transaction
        indices.update_timestamp = timestamp;
        self.save_indices(indices);


        // Remove scheduled relations as well
        relations_to_remove.iter().for_each(|r| {
            self.remove_relation(timestamp, *r);
        });

        return true;
    }

    pub fn get_latest_node_record(&self, index: Index) -> Option<&NodeIndexRecord> {
        let indices = self.load_indices_active();

        self.node_index[index.type_index]
            .iter()
            .take(indices.node_index_tail[index.type_index])
            .rev()
            .find(|n| {
                if (**n).2 == index.instance_id { return true; }
                else { return false; }
            })
    }

    pub fn get_latest_node_value(&self, index: Index) -> Option<&KnowledgeGraphNode> {
        let record = self.get_latest_node_record(index);
        if record.is_none() { return None; }

        // If the latest record is marked as removed return None
        let record = record.unwrap();
        if record.5 == Marker::Removed { return None; }

        return Some(&self.nodes[record.1]);
    }

    pub fn get_nodes_of_type(&self, interval: Interval, type_index: usize)
            -> Vec<(Timestamp,
                    ValueId,
                    InstanceId,
                    Vec<Index>,
                    Vec<Index>,
                    Marker,
                    KnowledgeGraphNode)> {

        // Merge node additions and removals
        let mut nodes = BTreeMap::new();
        for n in self.nodes_iter(interval.clone(), type_index) {
            if n.5 == Marker::Valid {
                nodes.insert(n.2, n);
            } else {
                nodes.remove(&n.2);
            }
        }

        return nodes.values().cloned().collect();
    }

    pub fn get_all_nodes(&self, interval: Interval)
            -> Vec<(Timestamp,
                    ValueId,
                    InstanceId,
                    Vec<Index>,
                    Vec<Index>,
                    Marker,
                    KnowledgeGraphNode)> {

        let mut nodes = vec![];

        let type_count = KNOWLEDGE_GRAPH_NODE_TYPE_COUNT + 1;
        for type_index in 0..type_count {
            (&mut nodes).append(&mut self.get_nodes_of_type(interval.clone(), type_index))
        }

        return nodes;
    }

    // The returned tuple contains type_index and instance_id together!!!
    // BEWARE!!! Adding a relation also modifies nodes it connects
    // Therefore the indices of original nodes might not bet the most recent ones
    // Therefore, it is always necessary to do a search forward to see if there are any updates!!!
    pub fn add_relation(&mut self,
                        timestamp: Timestamp,
                        instance_id: InstanceId,
                        relation: KnowledgeGraphRelation,
                        from_node_index: Index,
                        to_node_index: Index) -> Index {

        let type_index = relation.type_index() as usize;
        let mut indices = self.load_indices();


        // Then add relation to the corresponding nodes i.e. crate NEW nodes!!!
        // Get the latest from and to node indices
        let latest_from_node_record_raw = self.get_latest_node_record(from_node_index);
        let latest_to_node_record_raw = self.get_latest_node_record(to_node_index);

        // If one of the nodes wasnt found, return
        if latest_from_node_record_raw.is_none() || latest_to_node_record_raw.is_none() { return Index::zeroed() }
        let mut latest_from_node_record = latest_from_node_record_raw.unwrap().clone();
        let mut latest_to_node_record = latest_to_node_record_raw.unwrap().clone();


        // Propagate info about new relation to nodes
        let index_relation = Index::new(type_index, instance_id);
        // Dont forget to update the timestamp!!!
        latest_from_node_record.0 = timestamp;
        let from_node_relation_index = KnowledgeGraph::find_empty_relation_cell_index(&latest_from_node_record.3, &index_relation);
        latest_from_node_record.3[from_node_relation_index as usize] = index_relation;

        latest_to_node_record.0 = timestamp;
        let to_node_relation_index = KnowledgeGraph::find_empty_relation_cell_index(&latest_to_node_record.4, &index_relation);
        latest_to_node_record.4[to_node_relation_index as usize] = index_relation;

        // Advance node index tails i.e. confirm node transaction
        let from_node_type = from_node_index.type_index;
        let to_node_type = to_node_index.type_index;


        self.node_index[from_node_type][indices.node_index_tail[from_node_type]] = latest_from_node_record;
        indices.node_index_tail[from_node_type] += 1;

        self.node_index[to_node_type][indices.node_index_tail[to_node_type]] = latest_to_node_record;
        indices.node_index_tail[to_node_type] += 1;


        // Finally add the relation
        let relation_id = indices.relations_tail;
        self.relations[relation_id] = relation;
        indices.relations_tail += 1;

        // Then index it
        let index_record = (timestamp,
                            relation_id,
                            instance_id,
                            Index::new(from_node_type, latest_from_node_record.2),
                            Index::new(to_node_type, latest_to_node_record.2),
                            Marker::Valid);

        let index_id = indices.relation_index_tail[type_index];
        self.relation_index[type_index][index_id] = index_record;
        indices.relation_index_tail[type_index] += 1;


        // Potentially commit transaction
        indices.update_timestamp = timestamp;
        self.save_indices(indices);

        return Index::new(type_index, instance_id);
    }

    pub fn remove_relation(&mut self, timestamp: Timestamp, index: Index) -> bool {
        let mut indices = self.load_indices();

        // Fetch last relation index record
        let relation_record_raw = self.get_latest_relation_record(index);

        // If nothing was found return, cannot remove
        if relation_record_raw.is_none() { return false; }

        let mut relation_record = relation_record_raw.unwrap().clone();
        relation_record.0 = timestamp;
        relation_record.5 = Marker::Removed;

        self.relation_index[index.type_index][indices.relation_index_tail[index.type_index]] = relation_record;
        indices.relation_index_tail[index.type_index] += 1;


        // Also mark removal of indices from the corresponding nodes
        // Find latest nodes
        let mut latest_from_node_record =
            self.get_latest_node_record(relation_record.3).unwrap().clone();
        let mut latest_to_node_record =
            self.get_latest_node_record(relation_record.4).unwrap().clone();

        // Modify indices
        // Dont forget to update the timestamp!!!
        latest_from_node_record.0 = timestamp;
        KnowledgeGraph::empty_relation_cell_by_value(&mut latest_from_node_record.3, &index);
        latest_to_node_record.0 = timestamp;
        KnowledgeGraph::empty_relation_cell_by_value(&mut latest_to_node_record.4, &index);


        // Save indices and advance tails
        let from_type_index = relation_record.3.type_index;  // From
        let to_type_index = relation_record.4.type_index;  // To

        self.node_index[from_type_index][indices.node_index_tail[from_type_index]] = latest_from_node_record;
        indices.node_index_tail[from_type_index] += 1;
        self.node_index[to_type_index][indices.node_index_tail[to_type_index]] = latest_to_node_record;
        indices.node_index_tail[to_type_index] += 1;


        // Potentially commit transaction
        indices.update_timestamp = timestamp;
        self.save_indices(indices);

        return true;
    }

    pub fn get_latest_relation_record(&self, index: Index) -> Option<&RelationIndexRecord> {
        let indices = self.load_indices_active();

        self.relation_index[index.type_index]
            .iter()
            .take(indices.relation_index_tail[index.type_index])
            .rev()
            .find(|n| {
                if (**n).2 == index.instance_id { return true; }
                else { return false; }
            })
    }

    pub fn get_latest_relation_value(&self, index: Index) -> Option<&KnowledgeGraphRelation> {
        let record = self.get_latest_relation_record(index);
        if record.is_none() { return None; }

        // If the latest record is marked as removed return None
        let record = record.unwrap();
        if record.5 == Marker::Removed { return None; }

        return Some(&self.relations[record.1]);
    }

    pub fn get_relations_of_type(&self, interval: Interval, type_index: usize)
            -> Vec<(Timestamp,
                    ValueId,
                    InstanceId,
                    Index,
                    Index,
                    Marker,
                    KnowledgeGraphRelation)> {

        // Merge node additions and removals
        let mut relations = BTreeMap::new();
        for r in self.relations_iter(interval.clone(), type_index) {
            if r.5 == Marker::Valid {
                relations.insert(r.2, r);
            } else {
                relations.remove(&r.2);
            }
        }

        return relations.values().cloned().collect();
    }

    pub fn get_all_relations(&self, interval: Interval)
            -> Vec<(Timestamp,
                    ValueId,
                    InstanceId,
                    Index,
                    Index,
                    Marker,
                    KnowledgeGraphRelation)> {

        let mut relations = vec![];

        let type_count = KNOWLEDGE_GRAPH_RELATION_TYPE_COUNT + 1;
        for type_index in 0..type_count {
            (&mut relations).append(&mut self.get_relations_of_type(interval.clone(), type_index));
        }

        return relations
    }

    // Helper methods for relation array
    fn find_empty_relation_cell_index(array: &[Index], index: &Index) -> isize {
        array.iter().enumerate().rev().fold(-1, |mut accumulator, (i, val)| {
            if val.type_index == 0 || *val == *index {
                accumulator = i as isize;
            }
            accumulator
        })
    }

    fn empty_relation_cell_by_value(array: &mut [Index], index: &Index) {
        let _cell = array.iter_mut().for_each(|val| {
            if *val == *index {
                *val = Index::zeroed();
            }
        });
    }
}
