use super::*;


pub struct NodeIterator<'a> {
    inner: &'a KnowledgeGraph,
    indices: GraphIndices,
    interval: Interval,
    type_index: TypeIndex,
    type_count: usize,
    index_column: usize
}
impl<'a> NodeIterator<'a> {
    pub fn new(inner: &'a KnowledgeGraph,
               indices: GraphIndices,
               interval: Interval,
               type_index: usize) -> Self {

        NodeIterator {
            inner,
            indices,
            interval,
            type_index,
            type_count: KNOWLEDGE_GRAPH_NODE_TYPE_COUNT + 1,
            index_column: 0
        }
    }
}
impl<'a> Iterator for NodeIterator<'a> {
    // timestamp, value id, instance id, from, to, marker, node
    type Item = (Timestamp, ValueId, InstanceId, Vec<Index>, Vec<Index>, Marker, KnowledgeGraphNode);

    fn next(&mut self) -> Option<Self::Item> {
        if self.type_index >= self.type_count {
            return None;
        }

        let mut index_record;
        loop {
            // Check if we have seen all records
            if self.index_column >= self.indices.node_index_tail[self.type_index] {
                return None;
            }

            // Fetch index record
            index_record = &self.inner.node_index[self.type_index][self.index_column];

            // Advance index
            self.index_column += 1;

            // Check if the value belongs to the interval
            if self.interval.includes(index_record.0) {
                break;
            }
        }

        // Forge the actual node element
        let timestamp = index_record.0;
        let value_id = index_record.1;
        let instance_id = index_record.2;
        let from = index_record.3.to_vec();
        let to = index_record.4.to_vec();
        let marker = index_record.5;

        let node = self.inner.nodes[value_id];
        return Some((timestamp, value_id, instance_id, from, to, marker, node));
    }
}


pub struct RelationIterator<'a> {
    inner: &'a KnowledgeGraph,
    indices: GraphIndices,
    interval: Interval,
    type_index: TypeIndex,
    type_count: usize,
    index_column: usize
}
impl<'a> RelationIterator<'a> {
    pub fn new(inner: &'a KnowledgeGraph,
               indices: GraphIndices,
               interval: Interval,
               type_index: TypeIndex) -> Self {

        RelationIterator {
            inner,
            indices,
            interval,
            type_index,
            type_count: KNOWLEDGE_GRAPH_RELATION_TYPE_COUNT + 1,
            index_column: 0
        }
    }
}
impl<'a> Iterator for RelationIterator<'a> {
    // value id, instance id, timestamp, from, to, marker, relation
    type Item = (Timestamp, ValueId, InstanceId, Index, Index, Marker, KnowledgeGraphRelation);

    fn next(&mut self) -> Option<Self::Item> {
        if self.type_index >= self.type_count {
            return None;
        }

        let mut index_record;
        loop {
            // Check if we have seen all records
            if self.index_column >= self.indices.relation_index_tail[self.type_index] {
                return None;
            }

            // Fetch index record
            index_record = &self.inner.relation_index[self.type_index][self.index_column];

            // Advance index
            self.index_column += 1;

            // Check if the value belongs to the interval
            if self.interval.includes(index_record.0) {
                break;
            }
        }

        // Get the actual relation
        let timestamp = index_record.0;
        let value_id = index_record.1;
        let instance_id = index_record.2;
        let from = index_record.3;
        let to = index_record.4;
        let marker = index_record.5;

        let node = self.inner.relations[value_id];
        return Some((timestamp, value_id, instance_id, from, to, marker, node));
    }
}
