use knowledge_graph::graph::*;
use knowledge_graph::graph::types::*;

#[test]
fn empty_graph_nodes() {
    let graph = KnowledgeGraph::new_boxed();
    assert_eq!(graph.node_count(), 0);
}

#[test]
fn empty_graph_relations() {
    let graph = KnowledgeGraph::new_boxed();
    assert_eq!(graph.relation_count(), 0);
}

#[test]
fn graph_add_nodes() {
    let mut graph = KnowledgeGraph::new_boxed();

    let class1 = TestClass1 { testProperty1: 1 };
    let node1 = KnowledgeGraphNode::TestClass1(class1);
    graph.add_node(1, 0, node1);

    assert_eq!(graph.node_count(), 1);

    let class2 = TestClass2 { testProperty2: 100 };
    let node2 = KnowledgeGraphNode::TestClass2(class2);
    graph.add_node(100, 0,node2);

    assert_eq!(graph.node_count(), 2);
}

#[test]
fn graph_fetch_nodes() {
    let mut graph = KnowledgeGraph::new_boxed();

    let class1 = TestClass1 { testProperty1: 1 };
    let node1 = KnowledgeGraphNode::TestClass1(class1);
    let node1_id = graph.add_node(1, 0, node1);

    let class2 = TestClass2 { testProperty2: 100 };
    let node2 = KnowledgeGraphNode::TestClass2(class2);
    let node2_id = graph.add_node(100, 0, node2);

    let returned1 = graph.get_latest_node_value(node1_id).unwrap();
    let returned2 = graph.get_latest_node_value(node2_id).unwrap();

    if let KnowledgeGraphNode::TestClass1(class) = returned1 {
        assert_eq!(class.testProperty1, class1.testProperty1);
    } else {
        assert!(false);
    }

    if let KnowledgeGraphNode::TestClass2(class) = returned2 {
        assert_eq!(class.testProperty2, class2.testProperty2);
    } else {
        assert!(false);
    }
}

#[test]
fn graph_add_relations() {
    let mut graph = KnowledgeGraph::new_boxed();

    let class1 = TestClass1 { testProperty1: 1 };
    let node1 = KnowledgeGraphNode::TestClass1(class1);
    let node1_id = graph.add_node(1,0,node1);

    let class2 = TestClass2 { testProperty2: 100 };
    let node2 = KnowledgeGraphNode::TestClass2(class2);
    let node2_id = graph.add_node(100,0,node2);


    let rel = testRelation { testProperty: 1.0 };
    let relation = KnowledgeGraphRelation::testRelation(rel);
    let _relation_id = graph.add_relation(200,0,relation,
                                        node1_id,
                                        node2_id);

    assert_eq!(graph.node_count(), 4);
    assert_eq!(graph.relation_count(), 1);
}

#[test]
fn graph_fetch_relations() {
    let mut graph = KnowledgeGraph::new_boxed();

    let class1 = TestClass1 { testProperty1: 1 };
    let node1 = KnowledgeGraphNode::TestClass1(class1);
    let node1_id = graph.add_node(1,0,node1);

    let class2 = TestClass2 { testProperty2: 100 };
    let node2 = KnowledgeGraphNode::TestClass2(class2);
    let node2_id = graph.add_node(100,0,node2);

    let rel1 = testRelation { testProperty: 1.0 };
    let relation1 = KnowledgeGraphRelation::testRelation(rel1);
    let relation1_id = graph.add_relation(200,0,relation1,
                                        node1_id,
                                        node2_id);


    let returned1 = graph.get_latest_relation_value(relation1_id).unwrap();

    if let KnowledgeGraphRelation::testRelation(relation) = returned1 {
        assert_eq!(relation.testProperty, rel1.testProperty);
    } else {
        assert!(false);
    }
}

#[test]
fn graph_return_only_newest_node() {
    let mut graph = KnowledgeGraph::new_boxed();

    let class1_1 = KnowledgeGraphNode::TestClass1(TestClass1 { testProperty1: 10 });
    let class1_2 = KnowledgeGraphNode::TestClass1(TestClass1 { testProperty1: 25 });
    let class1_3 = KnowledgeGraphNode::TestClass1(TestClass1 { testProperty1: 42 });

    let class2_1 = KnowledgeGraphNode::TestClass2(TestClass2 { testProperty2: 88 });

    let _relation1 = KnowledgeGraphRelation::testRelation(testRelation { testProperty: 12.0 });
    let _relation2 = KnowledgeGraphRelation::testRelation(testRelation { testProperty: 22.0 });

    let id1_1 = graph.add_node(1, 0,class1_1);
    let _id1_2 = graph.add_node(10, 0,class1_2);
    let _id1_3 = graph.add_node(15, 0,class1_3);
    let id2_1 = graph.add_node(16, 0,class2_1);

    assert_eq!(class1_3, *graph.get_latest_node_value(id1_1).unwrap());
    assert_eq!(class2_1, *graph.get_latest_node_value(id2_1).unwrap());
}

#[test]
fn graph_add_nodes_relations() {
    let mut graph = KnowledgeGraph::new_boxed();

    let class1_1 = KnowledgeGraphNode::TestClass1(TestClass1 { testProperty1: 10 });
    let class1_2 = KnowledgeGraphNode::TestClass1(TestClass1 { testProperty1: 25 });
    let class1_3 = KnowledgeGraphNode::TestClass1(TestClass1 { testProperty1: 42 });

    let class2_1 = KnowledgeGraphNode::TestClass2(TestClass2 { testProperty2: 88 });

    let relation1 = KnowledgeGraphRelation::testRelation(testRelation { testProperty: 12.0 });
    let relation2 = KnowledgeGraphRelation::testRelation(testRelation { testProperty: 22.0 });

    let id1_1 = graph.add_node(1, 0,class1_1);
    let id1_2 = graph.add_node(10, 0,class1_2);
    let id1_3 = graph.add_node(15, 0,class1_3);
    let id2_1 = graph.add_node(16, 0,class2_1);

    let id_rel1_1 = graph.add_relation(17,0,relation1,
                                       id1_1,id1_2);
    let _id_rel1_2 = graph.add_relation(18,0,relation2,
                                       id1_3, id1_2);

    assert_eq!(class1_3, *graph.get_latest_node_value(id1_1).unwrap());
    assert_eq!(class2_1, *graph.get_latest_node_value(id2_1).unwrap());
    assert_eq!(relation2, *graph.get_latest_relation_value(id_rel1_1).unwrap());
}

#[test]
fn graph_add_remove_nodes() {
    let mut graph = KnowledgeGraph::new_boxed();

    let class1_1 = KnowledgeGraphNode::TestClass1(TestClass1 { testProperty1: 10 });
    let class1_2 = KnowledgeGraphNode::TestClass1(TestClass1 { testProperty1: 25 });
    let class1_3 = KnowledgeGraphNode::TestClass1(TestClass1 { testProperty1: 42 });

    let class2_1 = KnowledgeGraphNode::TestClass2(TestClass2 { testProperty2: 88 });

    let _relation1 = KnowledgeGraphRelation::testRelation(testRelation { testProperty: 12.0 });
    let _relation2 = KnowledgeGraphRelation::testRelation(testRelation { testProperty: 22.0 });

    let id1_1 = graph.add_node(1, 0,class1_1);
    let _id1_2 = graph.add_node(10, 0,class1_2);
    let _id1_3 = graph.add_node(15, 0,class1_3);
    let id2_1 = graph.add_node(16, 0,class2_1);

    graph.remove_node(20, id1_1);
    graph.remove_node(20, id2_1);

    assert_eq!(None, graph.get_latest_node_value(id1_1));
    assert_eq!(None, graph.get_latest_node_value(id2_1));
}

#[test]
fn remove_relation_preserve_nodes() {
    let mut graph = KnowledgeGraph::new_boxed();

    let class1_1 = KnowledgeGraphNode::TestClass1(TestClass1 { testProperty1: 10 });
    let class1_2 = KnowledgeGraphNode::TestClass1(TestClass1 { testProperty1: 25 });
    let class1_3 = KnowledgeGraphNode::TestClass1(TestClass1 { testProperty1: 42 });

    let class2_1 = KnowledgeGraphNode::TestClass2(TestClass2 { testProperty2: 88 });

    let relation1 = KnowledgeGraphRelation::testRelation(testRelation { testProperty: 12.0 });
    let relation2 = KnowledgeGraphRelation::testRelation(testRelation { testProperty: 22.0 });

    let id1_1 = graph.add_node(1, 0,class1_1);
    let _id1_2 = graph.add_node(10, 0,class1_2);
    let _id1_3 = graph.add_node(15, 0,class1_3);
    let id2_1 = graph.add_node(16, 0,class2_1);

    let id_rel1_1 = graph.add_relation(17,0,relation1,
                                       id1_1,id2_1);
    let _id_rel1_2 = graph.add_relation(18,0,relation2,
                                       id1_1, id2_1);

    graph.remove_relation(20, id_rel1_1);

    assert_eq!(class1_3, *graph.get_latest_node_value(id1_1).unwrap());
    assert_eq!(class2_1, *graph.get_latest_node_value(id2_1).unwrap());
}

#[test]
fn remove_node_and_its_relations() {
    let mut graph = KnowledgeGraph::new_boxed();

    let class1_1 = KnowledgeGraphNode::TestClass1(TestClass1 { testProperty1: 10 });
    let class1_2 = KnowledgeGraphNode::TestClass1(TestClass1 { testProperty1: 25 });
    let class1_3 = KnowledgeGraphNode::TestClass1(TestClass1 { testProperty1: 42 });

    let class2_1 = KnowledgeGraphNode::TestClass2(TestClass2 { testProperty2: 88 });

    let relation1 = KnowledgeGraphRelation::testRelation(testRelation { testProperty: 12.0 });
    let relation2 = KnowledgeGraphRelation::testRelation(testRelation { testProperty: 22.0 });

    let id1_1 = graph.add_node(1, 0,class1_1);
    let _id1_2 = graph.add_node(10, 0,class1_2);
    let _id1_3 = graph.add_node(15, 0,class1_3);
    let id2_1 = graph.add_node(16, 0,class2_1);

    let id_rel1_1 = graph.add_relation(17,0,relation1,
                                       id1_1,id2_1);
    let id_rel1_2 = graph.add_relation(18,0,relation2,
                                       id1_1, id2_1);

    graph.remove_node(20, id1_1);

    assert_eq!(None, graph.get_latest_relation_value(id_rel1_1));
    assert_eq!(None, graph.get_latest_relation_value(id_rel1_2));
}
