use codegen::*;
use super::parser_types::*;

const MAX_RELATIONS: usize = 10;


pub fn generate_knowledge_graph_api(nodes: Vec<(String, RawNode)>,
                                    relations: Vec<(String, RawRelation)>) -> (String, String)
{
    let mut rust_scope = codegen::Scope::new();
    let mut csharp_scope = csharp_codegen::Scope::new();

    // CSharp prerequsities
    csharp_scope.add_using("System");
    csharp_scope.add_using("System.Runtime.InteropServices");
    let csharp_namespace = csharp_scope.new_namespace("AssessmentSystemApi");

    // Create csharp type enums
    let node_enm = csharp_namespace.new_enum("KnowledgeGraphNode");
    node_enm.new_modifier("public");
    node_enm.new_option_with_value("Empty", "0");
    for node in nodes.iter().enumerate() {
        node_enm.new_option_with_value((node.1).0.as_str(), format!("{}", node.0 + 1).as_str());
    }

    let relation_enm = csharp_namespace.new_enum("KnowledgeGraphRelation");
    relation_enm.new_modifier("public");
    relation_enm.new_option_with_value("Empty", "0");
    for relation in relations.iter().enumerate() {
        relation_enm.new_option_with_value((relation.1).0.as_str(), format!("{}", relation.0 + 1).as_str());
    }

    // Create graph index struct
    let strct = csharp_namespace.new_struct("TypeIndex");
    strct.new_modifier("public");
    strct.new_header("StructLayout", "LayoutKind.Sequential",
                     Some("CallingConvention = CallingConvention.Cdecl"),false);

    let mut strct_from_node = csharp_codegen::Method::new("OfGraphNode", "uint");
    strct_from_node.new_modifier("public").new_modifier("static");
    strct_from_node.new_parameter("node", "KnowledgeGraphNode");
    strct_from_node.add_body_raw("var type_index = (int)node;");
    strct_from_node.add_body_raw("return (uint)type_index;");
    strct.add_method(strct_from_node);

    let mut strct_from_relation = csharp_codegen::Method::new("OfGraphRelation", "uint");
    strct_from_relation.new_modifier("public").new_modifier("static");
    strct_from_relation.new_parameter("relation", "KnowledgeGraphRelation");
    strct_from_relation.add_body_raw("var type_index = (int)relation;");
    strct_from_relation.add_body_raw("return (uint)type_index;");
    strct.add_method(strct_from_relation);

    // Helper class for unpacking strings
    let strct2 = csharp_namespace.new_struct("String2Bytes");
    strct2.new_modifier("public");
    strct2.new_header("StructLayout", "LayoutKind.Sequential",
                     Some("CallingConvention = CallingConvention.Cdecl"),false);

    let mut strct2_pack = csharp_codegen::Method::new("Pack", "byte[]");
    strct2_pack.new_modifier("public").new_modifier("static");
    strct2_pack.new_parameter("value", "string");
    strct2_pack.add_body_raw("var bytes = System.Text.Encoding.UTF8.GetBytes(value);");
    strct2_pack.add_body_raw("var result = new byte[128];");
    strct2_pack.add_body_raw("bytes.CopyTo(result, 0);");
    strct2_pack.add_body_raw("return result;");
    strct2.add_method(strct2_pack);

    // Helper data types
    // C#
    let vector3 = csharp_namespace.new_struct("Vector3f");
    vector3.new_modifier("public");
    vector3.new_header("StructLayout","LayoutKind.Sequential",
                     Some("CallingConvention = CallingConvention.Cdecl"),false);
    vector3.new_attribute("x", "float");
    vector3.new_attribute("y", "float");
    vector3.new_attribute("z", "float");

    let vector2 = csharp_namespace.new_struct("Vector2f");
    vector2.new_modifier("public");
    vector2.new_header("StructLayout","LayoutKind.Sequential",
                        Some("CallingConvention = CallingConvention.Cdecl"),false);
    vector2.new_attribute("x", "float");
    vector2.new_attribute("y", "float");


    // HERE ARE THE C# TYPES
    // Create csharp node structs
    for node in &nodes {
        let strct = csharp_namespace.new_struct(node.0.as_str());
        strct.new_modifier("public");
        strct.new_header("StructLayout", "LayoutKind.Sequential",
                         Some("CallingConvention = CallingConvention.Cdecl"), false);

        for property in &node.1.properties {
            let ty = property.1.as_str();

            // Special handling for strings
            if ty == "string" {
                let header = format!("[MarshalAs(UnmanagedType.ByValArray, SizeConst = 128)]");
                strct.new_attribute_with_header(property.0.as_str(),
                                    convert_to_csharp_type(property.1.as_str()),
                                                header.as_str());
            } else {
                strct.new_attribute(property.0.as_str(),
                                    convert_to_csharp_type(property.1.as_str()));
            }
        }
    }

    // Create csharp relation structs
    for relation in &relations {
        let strct = csharp_namespace.new_struct(relation.0.as_str());
        strct.new_modifier("public");
        strct.new_header("StructLayout","LayoutKind.Sequential",
                         Some("CallingConvention = CallingConvention.Cdecl"),false);

        for property in &relation.1.properties {
            let ty = property.1.as_str();

            // Special handling for strings
            if ty == "string" {
                let header = format!("[MarshalAs(UnmanagedType.ByValArray, SizeConst = 128)]");
                strct.new_attribute_with_header(property.0.as_str(),
                                    convert_to_csharp_type(property.1.as_str()),
                                                header.as_str());
            } else {
                strct.new_attribute(property.0.as_str(),
                                    convert_to_csharp_type(property.1.as_str()));
            }
        }
    }


    let csharp_class = csharp_namespace.new_class("KnowledgeGraph");

    // Node manipulation
    for node in &nodes {
        let node_name = node.0.as_str();
        let node_name_lowercase = node.0.to_lowercase();

        // Add nodes
        let add_function_name = format!("add_{}_node", node.0).to_lowercase();

        // Rust
        let add_function = rust_scope.new_function(add_function_name.as_str())
            .attribute("no_mangle").vis("pub extern").ret("usize");

        add_function.arg("timestamp", "u64")
            .arg("instance_id", "usize")
            .arg(node_name_lowercase.as_str(), node_name)
            .arg("memory_ptr", "*mut SharedMemory<KnowledgeGraph>");

        add_function.line("let memory = unsafe { &mut *memory_ptr };")
            .line("let graph: &mut KnowledgeGraph = memory.get_data_mut_ref();")
            .line(format!("let node = KnowledgeGraphNode::{}({});", node_name, node_name_lowercase))
            .line(format!("let type_index = node.type_index();"))
            .line(format!("graph.add_node(timestamp, instance_id, node);"))
            .line(format!("return type_index;"));

        // C#
        let mut add_function_signature = csharp_codegen::MethodSignature::new(add_function_name.as_str(), "uint");
        add_function_signature.new_modifier("public static extern")
            .new_header("DllImport", "unity_interface",
                        Some("CallingConvention = CallingConvention.Cdecl"), true)
            .new_parameter("timestamp", "ulong")
            .new_parameter("instance_id", "uint")
            .new_parameter(node_name_lowercase.as_str(), node_name)
            .new_parameter("memory_ptr", "IntPtr");

        csharp_class.add_method_signature(add_function_signature);


        // Remove nodes
        let remove_function_name = format!("remove_{}_node", node.0).to_lowercase();

        // Rust
        let remove_function = rust_scope.new_function(remove_function_name.as_str())
            .attribute("no_mangle").vis("pub extern").ret("bool");

        remove_function.arg("timestamp", "u64")
            .arg("type_index", "usize")
            .arg("instance_id", "usize")
            .arg("memory_ptr", "*mut SharedMemory<KnowledgeGraph>");

        remove_function.line("let memory = unsafe { &mut *memory_ptr };")
            .line("let graph: &mut KnowledgeGraph = memory.get_data_mut_ref();")
            .line("let index = Index::new(type_index, instance_id);")
            .line(format!("return graph.remove_node(timestamp, index);"));

        // C#
        let mut remove_function_signature = csharp_codegen::MethodSignature::new(remove_function_name.as_str(), "bool");
        remove_function_signature.new_modifier("public static extern")
            .new_header("DllImport", "unity_interface",
                        Some("CallingConvention = CallingConvention.Cdecl"), true)
            .new_parameter("timestamp", "ulong")
            .new_parameter("type_index", "uint")
            .new_parameter("instance_id", "uint")
            .new_parameter("memory_ptr", "IntPtr");

        csharp_class.add_method_signature(remove_function_signature);
    }


    // Relation manipulation
    for relation in &relations {
        let relation_name = relation.0.as_str();
        let relation_name_lowercase = relation.0.to_lowercase();

        // Add relations
        let add_function_name = format!("add_{}_relation", relation.0).to_lowercase();

        // Rust
        let add_function = rust_scope.new_function(add_function_name.as_str())
            .attribute("no_mangle").vis("pub extern").ret("usize");

        add_function.arg("timestamp", "Timestamp")
            .arg("instance_id", "usize")
            .arg(relation_name_lowercase.as_str(), relation_name)
            .arg("from_node_type", "usize")
            .arg("from_node_instance_id", "usize")
            .arg("to_node_type", "usize")
            .arg("to_node_instance_id", "usize")
            .arg("memory_ptr", "*mut SharedMemory<KnowledgeGraph>");

        add_function.line("let memory = unsafe { &mut *memory_ptr };")
            .line("let graph: &mut KnowledgeGraph = memory.get_data_mut_ref();")
            .line(format!("let relation = KnowledgeGraphRelation::{}({});", relation_name, relation_name_lowercase))
            .line(format!("let type_index = relation.type_index();"))
            .line(format!("let from_node_index = Index::new(from_node_type, from_node_instance_id);"))
            .line(format!("let to_node_index = Index::new(to_node_type, to_node_instance_id);"))
            .line(format!("graph.add_relation(timestamp, instance_id, relation, from_node_index, to_node_index);"))
            .line(format!("return type_index;"));

        // C#
        let mut add_function_signature = csharp_codegen::MethodSignature::new(add_function_name.as_str(), "uint");
        add_function_signature.new_modifier("public static extern")
            .new_header("DllImport","unity_interface",
                        Some("CallingConvention = CallingConvention.Cdecl"), true)
            .new_parameter("timestamp", "ulong")
            .new_parameter("instance_id", "uint")
            .new_parameter(relation_name_lowercase.as_str(), relation_name)
            .new_parameter("from_node_type", "uint")
            .new_parameter("from_node_index_id", "uint")
            .new_parameter("to_node_type", "uint")
            .new_parameter("to_node_index_id", "uint")
            .new_parameter("memory_ptr", "IntPtr");

        csharp_class.add_method_signature(add_function_signature);


        // Remove relations
        let remove_function_name = format!("remove_{}_relation", relation.0).to_lowercase();

        // Rust
        let remove_function = rust_scope.new_function(remove_function_name.as_str())
            .attribute("no_mangle").vis("pub extern").ret("bool");

        remove_function.arg("timestamp", "u64")
            .arg("type_index", "usize")
            .arg("instance_id", "usize")
            .arg("memory_ptr", "*mut SharedMemory<KnowledgeGraph>");

        remove_function.line("let memory = unsafe { &mut *memory_ptr };")
            .line("let graph: &mut KnowledgeGraph = memory.get_data_mut_ref();")
            .line("let index = Index::new(type_index, instance_id);")
            .line(format!("return graph.remove_relation(timestamp, index);"));

        // C#
        let mut remove_function_signature = csharp_codegen::MethodSignature::new(remove_function_name.as_str(), "bool");
        remove_function_signature.new_modifier("public static extern")
            .new_header("DllImport","unity_interface",
                        Some("CallingConvention = CallingConvention.Cdecl"),true)
            .new_parameter("timestamp", "ulong")
            .new_parameter("type_index", "uint")
            .new_parameter("index_id", "uint")
            .new_parameter("memory_ptr", "IntPtr");

        csharp_class.add_method_signature(remove_function_signature);
    }

    return (rust_scope.to_string(), csharp_scope.to_string());
}


pub fn generate_rust_types(nodes: Vec<(String, RawNode)>,
                           relations: Vec<(String, RawRelation)>) -> String
{
    let mut scope = codegen::Scope::new();
    scope.import("derivative", "*");
    scope.import("phf", "phf_map");

    // Total node and relation count
    scope.new_constant("KNOWLEDGE_GRAPH_NODE_TYPE_COUNT", "usize",
                       nodes.len().to_string().as_str()).vis("pub");
    scope.new_constant("KNOWLEDGE_GRAPH_RELATION_TYPE_COUNT", "usize",
                       relations.len().to_string().as_str()).vis("pub");

    // Rust
    // Helper types
    let vector3 = scope.new_struct("Vector3f")
        .vis("pub")
        .repr("C")
        .derive("Clone")
        .derive("Copy")
        .derive("PartialEq")
        .derive("Default")
        .derive("Debug");
    vector3.pub_field("x", "f32");
    vector3.pub_field("y", "f32");
    vector3.pub_field("z", "f32");

    let vector2 = scope.new_struct("Vector2f")
        .vis("pub")
        .repr("C")
        .derive("Clone")
        .derive("Copy")
        .derive("PartialEq")
        .derive("Default")
        .derive("Debug");
    vector2.pub_field("x", "f32");
    vector2.pub_field("y", "f32");

    // Special raw string wrapper
    let raw_string = scope.new_struct("RawString")
        .vis("pub")
        .repr("C")
        .derive("Clone")
        .derive("Copy");
    raw_string.pub_field("raw", "[u8; 128]");

    // Implement default
    let mut impl_raw_string_default = Impl::new("RawString");
    impl_raw_string_default.impl_trait("Default")
        .new_fn("default")
        .ret("Self")
        .line("return RawString { raw: [0; 128] };");
    scope.push_impl(impl_raw_string_default);


    // Generate enum with all node types
    let node_enum = scope.new_enum("KnowledgeGraphNode")
        .vis("pub")
        .repr("usize")
        .derive("Clone")
        .derive("Copy")
        .derive("PartialEq")
        .derive("Debug");

    node_enum.new_variant("Empty");
    for node in nodes.iter().enumerate() {

        node_enum.new_variant((node.1).0.as_str())
            .tuple((node.1).0.as_str());
    }

    // Implement default
    let mut impl_node_default = Impl::new("KnowledgeGraphNode");
    impl_node_default.impl_trait("Default")
        .new_fn("default")
        .ret("Self")
        .line("return KnowledgeGraphNode::Empty;");
    scope.push_impl(impl_node_default);

    // Implement type index method
    scope.new_impl("KnowledgeGraphNode").new_fn("type_index")
        .arg_ref_self().vis("pub").ret("usize")
        .line("return unsafe { *(self as *const Self as *const usize) };");

    // Nodes
    for node in &nodes {
        let node_struct = scope.new_struct(node.0.as_str())
            .vis("pub")
            .repr("C")
            .derive("Clone")
            .derive("Copy")
            .derive("Default")
            .derivative("Debug, PartialEq");

        for attributes in &node.1.properties {
            // If the struct contains binary data ignore them for Eq and Debug
            let name = attributes.0.as_str();
            let ty = attributes.1.as_str();
            if ty == "string" {
                let derivative = "Debug=\"ignore\",PartialEq=\"ignore\"";
                node_struct.pub_field_with_derivative(name,convert_to_rust_type(ty),
                                                      derivative);
            } else {
                node_struct.pub_field(name,convert_to_rust_type(ty));
            }
        }
    }
    scope.raw("");


    // Generate enum with all relation types
    let relation_enum = scope.new_enum("KnowledgeGraphRelation")
        .vis("pub")
        .repr("usize")
        .derive("Clone")
        .derive("Copy")
        .derive("PartialEq")
        .derive("Debug");

    relation_enum.new_variant("Empty");
    for relation in relations.iter().enumerate() {
        relation_enum.new_variant((relation.1).0.as_str())
            .tuple((relation.1).0.as_str());
    }

    // Implement default
    let mut impl_relation_default = Impl::new("KnowledgeGraphRelation");
    impl_relation_default.impl_trait("Default")
        .new_fn("default")
        .ret("Self")
        .line("return KnowledgeGraphRelation::Empty;");
    scope.push_impl(impl_relation_default);

    // Implement type index method
    scope.new_impl("KnowledgeGraphRelation").new_fn("type_index")
        .arg_ref_self().vis("pub").ret("usize")
        .line("return unsafe { *(self as *const Self as *const usize) };");

    // Relations
    for relation in &relations {
        let relation_struct = scope.new_struct(relation.0.as_str())
            .vis("pub")
            .repr("C")
            .derive("Clone")
            .derive("Copy")
            .derive("Default")
            .derivative("Debug, PartialEq");

        for attributes in &relation.1.properties {
            // If the struct contains binary data ignore them for Eq and Debug
            let name = attributes.0.as_str();
            let ty = attributes.1.as_str();
            if ty == "string" {
                let derivative = "Debug=\"ignore\",PartialEq=\"ignore\"";
                relation_struct.pub_field_with_derivative(name, convert_to_rust_type(ty),
                                                      derivative);
            } else {
                relation_struct.pub_field(name, convert_to_rust_type(ty));
            }
        }
    }


    // Subclass relations
    scope.raw("static SUBCLASSES: phf::Map<&'static str, &'static str> = phf_map! {");

    for node in &nodes {
        let mut entry = String::new();
        entry.push_str("\"");
        entry.push_str(node.0.as_str());
        entry.push_str("\" => \"");

        entry.push_str(node.1.subclass_of.as_str());
        entry.push_str("\",");
        scope.raw(entry.as_str());
    }

    scope.raw("};");

    // Relation constraints
    let mut map_definition = String::new();
    map_definition.push_str("static RELATIONS: phf::Map<&'static str, [&'static str; ");
    map_definition.push_str(MAX_RELATIONS.to_string().as_str());
    map_definition.push_str("]> = phf_map! {");
    scope.raw(map_definition.as_str());

    for node in &nodes {
        let node_name = node.0.as_str();

        let mut entry = String::new();
        entry.push_str("\"");
        entry.push_str(node_name);
        entry.push_str("\" => ");

        entry.push_str("[");

        let mut elements = String::new();
        let mut element_count = 0;
        for relation in &relations {
            if relation.1.from == node_name {
                elements.push_str("\"");
                elements.push_str(relation.1.to.as_str());
                elements.push_str("\",");
                element_count += 1;
            }
        }
        let empty_elements = generate_empty_elements_string(MAX_RELATIONS - element_count);
        elements.push_str(empty_elements.as_str());

        entry.push_str(elements.as_str());
        entry.push_str("],");

        scope.raw(entry.as_str());
    }

    scope.raw("};");

    // Final code
    return scope.to_string();
}

fn convert_to_rust_type(datatype: &str) -> &str {

    let rust_type = match datatype {
        "float" => "f32",
        "double" => "f64",
        "boolean" => "bool",
        "decimal" => "i32",
        "integer" => "i32",
        "int" => "i32",
        "long" => "i64",
        "short" => "i16",
        "byte" => "i8",
        "unsignedInt" => "u32",
        "unsignedLong" => "u64",
        "unsignedShort" => "u64",
        "unsignedByte" => "u8",
        "Vector3" => "Vector3f",
        "Vector2" => "Vector2f",
        "string" => "RawString",
        _ => "Any",
    };

    return rust_type;
}

fn convert_to_csharp_type(datatype: &str) -> &str {

    let rust_type = match datatype {
        "float" => "float",
        "double" => "double",
        "boolean" => "bool",
        "decimal" => "decimal",
        "integer" => "int",
        "int" => "int",
        "long" => "long",
        "short" => "short",
        "byte" => "sbyte",
        "unsignedInt" => "uint",
        "unsignedLong" => "ulong",
        "unsignedShort" => "ushort",
        "unsignedByte" => "byte",
        "Vector3" => "Vector3f",
        "Vector2" => "Vector2f",
        "string" => "byte[]",
        _ => "void",
    };

    return rust_type;
}

fn generate_empty_elements_string(empty_elements_count: usize) -> String {
    let empty_element = "\"\",";

    let mut result = String::new();
    for _ in 0..empty_elements_count {
        result.push_str(empty_element);
    }

    return result;
}
