use std::fs;
use std::collections::HashMap;

use itertools::Itertools;

use rio_api::parser::TriplesParser;
use rio_turtle::TurtleParser;
use rio_turtle::TurtleError;

use super::parser_types::*;


pub fn parse(filename: &str) -> (Vec<(String, RawNode)>, Vec<(String, RawRelation)>) {

    // Read ontology file
    let file_content = fs::read_to_string(filename)
        .expect("Could not load Turtle ontology file!");

    // Create turtle parser
    let mut turtle_parser = TurtleParser::new(file_content.as_ref(), "")
        .expect("Could not parse Turtle file!");

    let mut nodes: HashMap<String, RawNode> = HashMap::new();
    let mut relations: HashMap<String, RawRelation> = HashMap::new();
    let mut node_properties: HashMap<String, RawNodeProperty> = HashMap::new();


    // Parse classes and relations
    turtle_parser.parse_all(&mut |triple| {

        let subject: String = extract_type(triple.subject.to_string());
        let predicate: String = extract_type(triple.predicate.to_string());
        let object: String = extract_type(triple.object.to_string());

        let is_relation_attribute = remove_lt_gt(triple.predicate.to_string())
            .starts_with("urn");

        // Extract and save relations
        if object == "ObjectProperty"
        {
            let relation = RawRelation::new();
            relations.insert(subject.clone(), relation);
        }

        // From and to
        else if predicate == "domain" && relations.contains_key(subject.as_str())
        {
            relations.get_mut(subject.as_str()).unwrap().from = object.clone();
        }

        else if predicate == "range" && relations.contains_key(subject.as_str())
        {
            relations.get_mut(subject.as_str()).unwrap().to = object.clone();
        }

        // Relation attributes
        else if is_relation_attribute
        {
            relations.get_mut(subject.as_str()).unwrap().properties
                .push((predicate.clone(), object.clone()));
        }

        // Extract and save classes
        else if object == "Class"
        {
            let node = RawNode::new();
            nodes.insert(subject.clone(), node);
        }

        // Subclasses (pseudo relations)
        else if predicate == "subClassOf"
        {
            nodes.get_mut(subject.as_str()).unwrap().subclass_of = object.clone();
        }

        // Class properties
        else if object == "DatatypeProperty"
        {
            let property = RawNodeProperty::new();
            node_properties.insert(subject.clone(), property);
        }

        else if predicate == "domain" && node_properties.contains_key(subject.as_str())
        {
            node_properties.get_mut(subject.as_str()).unwrap().belongs_to = object.clone();
        }

        else if predicate == "range" && node_properties.contains_key(subject.as_str())
        {
            node_properties.get_mut(subject.as_str()).unwrap().data_type = object.clone();
        }

        return Ok(()) as Result<(), TurtleError>;

    }).expect("Error during Turtle triples parsing!");


    // Map node properties to nodes
    // Remember to sort the properties, the order can change because of hashmap
    for properties in node_properties.iter()
        .sorted_by(|a, b| Ord::cmp(&a.0, &b.0))
        .map(|v| ((*v.0).clone(), (*v.1).clone()))
    {
        if nodes.get_mut(properties.1.belongs_to.as_str()).is_none()
        {
            continue;
        }

        nodes.get_mut(properties.1.belongs_to.as_str()).unwrap().properties
            .push((properties.0, properties.1.data_type));
    }

    // Convert HashMaps to vectors in order to get consistent iterators
    // Remember to sort the nodes, the order can change because of hashmap
    let nodes_vec: Vec<(String, RawNode)> = nodes.iter()
        .sorted_by(|a, b| Ord::cmp(&a.0, &b.0))
        .map(|v| ((*v.0).clone(), (*v.1).clone())).collect();

    // Remember to sort the relations, the order can change because of hashmap
    let relations_vec: Vec<(String, RawRelation)> = relations.iter()
        .sorted_by(|a, b| Ord::cmp(&a.0, &b.0))
        .map(|v| ((*v.0).clone(), (*v.1).clone())).collect();

    return (nodes_vec, relations_vec);
}

fn extract_type(string: String) -> String
{
    remove_lt_gt(string).split('#').last().unwrap().to_string()
}

fn remove_lt_gt(string: String) -> String
{
    string.replace("<", "").replace(">", "")
}