// Description logic
// ABox axioms - knowledge about instances (individuals)
// > Concept assertions - assign instance to a class i.e. something is an instance of that class
// > Role assertions - describe relations between instances
// > Unique name assumption - each instance can be uniquely identified by name
// > Individual equality - two names refer to the same instance
// TBox axioms - knowledge about classes (concepts)
// > Concept inclusion - one class is a subclass of other
// > Concept equivalence - when two classes are the same
// RBox axioms - knowledge about relations (roles)
// > Role inclusion - one role is a subrole of other
// > Role equivalence - when two roles are the same
// > Role composition
// > Disjoint roles
// Constructors - way of defining more specific concepts, roles and individuals

// Signatures
// > Concept (class)
// > Role (Property)
// > Individual (Instance)

// > AND, OR, NOT, FOR ALL, THERE EXISTS, IS_A (i.e. inheritance, hierarchy)

use knowledge_graph::graph::*;
use knowledge_graph::graph::types::*;

use interval_algebra::Interval;


pub struct Pattern {
    node: Option<Node>,
}

impl Pattern {
    pub fn new() -> Self {
        Pattern {
            node: None,
        }
    }

    pub fn node(&mut self, node: Node) -> &mut Node {
        self.node = Some(node);
        self.node.as_mut().unwrap()
    }

    pub fn get_root(&self) -> Option<&Node> {
        if self.node.is_some() {
            self.node.as_ref()
        }
        else {
            None
        }
    }

    pub fn get_result(&self) -> Vec<NodeResult> {
        if self.node.is_some() {
            return self.node.as_ref().unwrap().get_results().clone()
        } else {
            return vec![];
        }
    }

    pub fn find(&mut self, interval: Interval, graph: &KnowledgeGraph) -> bool {
        // Do lookup
        if self.node.is_some() {
            return self.node.as_mut().unwrap().find(interval, graph);
        } else {
            return false;
        }
    }
}

#[derive(Clone, Debug)]
pub struct NodeResult {
    pub index: Index,
    pub timestamp: Timestamp,
    pub node: KnowledgeGraphNode,

    pub incoming_relations: Vec<RelationResult>,
    pub outgoing_relations: Vec<RelationResult>,
}
impl NodeResult {
    pub fn new(index: Index, timestamp: Timestamp, node: KnowledgeGraphNode) -> Self {
        NodeResult {
            index,
            timestamp,
            node,
            incoming_relations: vec![],
            outgoing_relations: vec![],
        }
    }
}

#[derive(Clone, Debug)]
pub struct Node {
    type_index: TypeIndex,
    relation_type: RelationType,
    incoming_relations: Vec<Relation>,
    outgoing_relations: Vec<Relation>,

    // Fetched results
    results: Vec<NodeResult>,
}
impl Node  {
    pub fn new(type_index: TypeIndex) -> Self {
        Node {
            type_index,
            relation_type: RelationType::Unspecified,
            incoming_relations: vec![],
            outgoing_relations: vec![],
            results: vec![],
        }
    }

    pub fn clear(&mut self) {
        self.results.clear();
    }

    pub fn get_results(&self) -> &Vec<NodeResult> {
        &self.results
    }

    // Finds given instance
    pub fn find(&mut self, interval: Interval, graph: &KnowledgeGraph) -> bool {
        self.clear();

        // If no nodes are found, fail right way
        let nodes = graph.get_nodes_of_type(interval, self.type_index);
        if nodes.len() == 0 {
            return false;
        }

        // Iterate over found nodes
        'outer: for candidate_node in nodes.iter() {
            let mut node_result = NodeResult::new(Index::new(self.type_index, candidate_node.2),
                candidate_node.0,
                candidate_node.6);

            // Check if the candidate has all the required outgoing relations
            let outgoing_relations = &candidate_node.3;

            for out in self.outgoing_relations.iter_mut() {
                if out.find(interval, graph) {
                    // At least one candidate needs to satisfy the relation requirements
                    let mut at_least_one = false;

                    for res in out.get_results() {
                        // Check if the relation really outcomes from this node
                        if outgoing_relations.contains(&(*res).index) {
                            // Insert result
                            node_result.outgoing_relations.push(res.clone());

                            at_least_one = true;
                        }
                    }
                    if !at_least_one {
                        continue 'outer;
                    }
                } else {
                    continue 'outer;
                }
            }

            // Check if the candidate has all the required incoming relations
            let incoming_relations = &candidate_node.4;

            for inc in self.incoming_relations.iter_mut() {
                if inc.find(interval, graph) {
                    // At least one candidate needs to satisfy the relation requirements
                    let mut at_least_one = false;

                    for res in inc.get_results() {
                        // Check if the relation really incomes to this node
                        if incoming_relations.contains(&(*res).index) {
                            // Insert result
                            node_result.incoming_relations.push(res.clone());

                            at_least_one = true;
                        }
                    }
                    if !at_least_one {
                        continue 'outer;
                    }
                } else {
                    continue 'outer;
                }
            }

            // Candidate passed all tests, lets add it
            self.results.push(node_result);
        }

        return self.results.len() > 0;
    }

    pub fn with_outgoing_relation(&mut self, mut relation: Relation) -> &mut Relation {
        relation.relation_type = RelationType::Outgoing;
        self.outgoing_relations.push(relation);
        self.outgoing_relations.last_mut().unwrap()
    }

    pub fn with_incoming_relation(&mut self, mut relation: Relation) -> &mut Relation {
        relation.relation_type = RelationType::Incoming;
        self.incoming_relations.push(relation);
        self.incoming_relations.last_mut().unwrap()
    }
}

#[derive(Clone, Copy, Debug)]
pub enum RelationType {
    Unspecified,
    Incoming,
    Outgoing
}

#[derive(Clone, Debug)]
pub struct RelationResult {
    pub index: Index,
    pub relation_type: RelationType,
    pub timestamp: Timestamp,
    pub relation: KnowledgeGraphRelation,

    pub from_node: Vec<NodeResult>,
    pub to_node: Vec<NodeResult>,
}
impl RelationResult {
    pub fn new(index: Index, timestamp: Timestamp, relation_type: RelationType, relation: KnowledgeGraphRelation) -> Self {
        RelationResult {
            index,
            relation_type,
            timestamp,
            relation,
            from_node: vec![],
            to_node: vec![],
        }
    }
}

#[derive(Clone, Debug)]
pub struct Relation {
    type_index: TypeIndex,
    relation_type: RelationType,
    from_node: Option<Node>,
    to_node: Option<Node>,

    // Fetched results
    results: Vec<RelationResult>,
}
impl Relation {
    pub fn new(type_index: TypeIndex) -> Self {
        Relation {
            type_index,
            relation_type: RelationType::Unspecified,
            from_node: None,
            to_node: None,
            results: vec![]
        }
    }

    pub fn clear(&mut self) {
        self.results.clear();
    }

    pub fn get_from(&self) -> Option<&Node> {
        self.from_node.as_ref()
    }

    pub fn get_to(&self) -> Option<&Node> {
        self.to_node.as_ref()
    }

    pub fn get_results(&self) -> &Vec<RelationResult> {
        &self.results
    }

    // Finds given instance
    pub fn find(&mut self, interval: Interval, graph: &KnowledgeGraph) -> bool {
        self.clear();

        // If no relations are found, fail right way
        let relations = graph.get_relations_of_type(interval, self.type_index);
        if relations.len() == 0 {
            return false;
        }

        // Iterate over found relations
        for candidate_relation in relations.iter() {
            let mut relation_result = RelationResult::new(Index::new(self.type_index,candidate_relation.2),
                                    candidate_relation.0,
                                    self.relation_type,
                                    candidate_relation.6);

            // Check the existence of the start/end nodes
            match self.relation_type {
                RelationType::Incoming => {
                    let from_node = self.from_node.as_mut().unwrap();
                    if !from_node.find(interval, graph) {
                        continue;
                    }
                    relation_result.from_node = from_node.get_results().clone();
                },
                RelationType::Outgoing => {
                    let to_node = self.to_node.as_mut().unwrap();
                    if !to_node.find(interval, graph) {
                        continue;
                    }
                    relation_result.to_node = to_node.get_results().clone();
                },
                _ => {},
            }

            // Candidate passed all tests, lets add it
            self.results.push(relation_result);
        }

        return self.results.len() > 0;
    }

    pub fn from_node(&mut self, node: Node) -> &mut Node {
        self.from_node = Some(node);
        self.from_node.as_mut().unwrap()
    }

    pub fn to_node(&mut self, node: Node) -> &mut Node {
        self.to_node = Some(node);
        self.to_node.as_mut().unwrap()
    }
}


pub fn hello_world() {

}