use std::str;
use std::{thread, time};

mod memory_interface;

pub mod constants;
use constants::*;

pub mod types;
use types::*;

pub mod shared_memory;
use shared_memory::*;

use config::*;


// Shared memory API
pub fn initialize_shared_memory<T>(region_name: &str, region_size: u32) -> *mut SharedMemory<T>
{
    // Create shared memory region
    let memory_ptr: *mut SharedMemory<T> = unsafe {
        std::mem::transmute(Box::new(SharedMemory::<T>::new(region_name, region_size)))
    };

    let memory = unsafe { &mut *memory_ptr };

    // Initialize knowledge graph memory
    let result = memory.initialize();
    if !result {
        return std::ptr::null_mut::<SharedMemory<T>>();
    }

    // Check if region exists
    if !memory.already_exists() {
        memory.zero_data();
    }

    return memory_ptr;
}

pub fn release_shared_memory<T>(memory_ptr: *mut SharedMemory<T>) {
    // Free shared memory region
    let memory = unsafe { &mut *memory_ptr };

    if memory.is_initialized() {
        memory.release();
    }

    // Release struct by letting it go out of scope
    let _memory : Box<SharedMemory<T>> = unsafe { std::mem::transmute(memory_ptr) };
}


// Sanity test methods
pub fn visual_shared_memory_sanity_test() {
    let mut memory = SharedMemory::<VisualSharedMemory>::new(VISUAL_SHARED_REGION_NAME,
                                                             VISUAL_SHARED_REGION_SIZE);
    memory.initialize();

    loop {
        let ten_millis = time::Duration::from_millis(3000);
        thread::sleep(ten_millis);

        dump_visual_shared_memory(memory.get_data_ref());
    }
}

fn dump_visual_shared_memory(visual_shared_memory: &VisualSharedMemory) {

    let vehicle_data = visual_shared_memory.vehicle_data;
    println!("[Shared memory]: Dump");
    println!("CarPosition: {}, {}, {}", vehicle_data.position_x,
             vehicle_data.position_y,
             vehicle_data.position_z);
    println!("SpeedVector: {}, {}, {}", vehicle_data.speed_vector_x,
             vehicle_data.speed_vector_y,
             vehicle_data.speed_vector_z);
    println!("Gear: {}", vehicle_data.gear);
    println!("Brake: {}", vehicle_data.brake);
    println!("Clutch: {}", vehicle_data.clutch);
    println!("Throttle: {}", vehicle_data.throttle);
    println!("RPM: {}", vehicle_data.rpm);
    println!("Speed: {}", vehicle_data.speed);

    let build_name_raw = visual_shared_memory.unity_build_name;
    println!("BuildName: {}", str::from_utf8(&build_name_raw).unwrap());

    let buffer = RingBuffer::from(visual_shared_memory.event_buffer,
                         visual_shared_memory.event_buffer_head as usize,
                         visual_shared_memory.event_buffer_tail as usize);

    let eye_event = EventType::EyetrackerEvent.to_u32();
    println!("Event: {:?}", buffer.buffer[buffer.head_index]);

    for i in buffer.tail_index..buffer.head_index {
        if buffer.buffer[i].event_type == eye_event {
            println!("EYE TRACKING: {:?}", buffer.buffer[i]);
        }
    }
    println!("Head: {}", buffer.head_index);
    println!("Tail: {}", buffer.tail_index);
}
