// Vehicle data struct
#[derive(Clone, Copy)]
#[repr(C)]
pub struct VehicleData
{
    // Car position
    pub position_x: f32,
    pub position_y: f32,
    pub position_z: f32,

    // Timestamp
    pub idx: i32,

    // Car quaternion rotation
    pub direction_x: f32,
    pub direction_y: f32,
    pub direction_z: f32,
    pub direction_w: f32,

    // Speed
    pub speed_vector_x: f32,
    pub speed_vector_y: f32,
    pub speed_vector_z: f32,

    // Gear
    pub gear: i32,

    offset_yaw: f32, // Not used
    offset_pitch: f32, // Not used
    offset_roll: f32, // Not used

    // Car data
    pub steering_wheel: f32,
    pub brake: f32,
    pub clutch: f32,
    pub throttle: f32,

    // Command flag
    flag: i32,

    // Engine data
    pub rpm: f32,
    pub torque: f32,
    sound_event_engine: i32,
    pub speed: f32,

    // Mask data for sound effects at each wheel
    sound_l1: u16,
    sound_l2: u16,
    sound_r1: u16,
    sound_r2: u16,

    // Timestamp for all objects in dynamic part of packets
    frame_idx: i32,
    // Used to transfer data to the display channel
    pub park_brake: f32,

    // Data from CarSim
    // Slip angles
    alpha_l1: f32,
    alpha_l2: f32,
    alpha_r1: f32,
    alpha_r2: f32,

    // Longitudinal slip values
    kappa_l1: f32,
    kappa_l2: f32,
    kappa_r1: f32,
    kappa_r2: f32,

    // RPM values for each wheel
    avy_l1: f32,
    avy_l2: f32,
    avy_r1: f32,
    avy_r2: f32,

    pub bits_from_car1: u32,
    pub bits_from_car2: u32,
    bits_from_simulation1: u32,
    bits_from_simulation2: u32,
}

// Collision data
#[derive(Clone, Copy)]
#[repr(C)]
pub struct HyperPlane3
{
    pub x: f32,
    pub y: f32,
    pub z: f32,
    dummy: f32,
}

#[derive(Clone, Copy)]
#[repr(C)]
pub struct CollisionData
{
    pub wheel_collision: [HyperPlane3; 4]
}

#[derive(Clone, Copy)]
#[repr(C)]
pub struct CarWheelCollision
{
    pub l1: [f32; 16],
    pub r1: [f32; 16],
    pub l2: [f32; 16],
    pub r2: [f32; 16],
}

#[derive(Clone, Copy)]
#[repr(C)]
pub struct CarWheelMU
{
    pub l1: f32,
    pub l2: f32,
    pub r1: f32,
    pub r2: f32,
}

// Can bus structs
#[derive(Clone, Copy)]
#[repr(C)]
pub struct CanMask
{
    pub mask: [u32; 64],
}

#[derive(Clone, Copy)]
#[repr(C)]
pub struct CanPacket
{
    pub id: u32,
    pub flag: u32,
    pub data: [u8; 8],
}

#[derive(Clone, Copy)]
#[repr(C)]
pub struct CanBase
{
    pub packet_time: u64,
    pub delta_time: f32,
    pub dummy: u32,

    pub can_packet: CanPacket,
}

// Synced objects struct - 24 bytes
#[derive(Clone, Copy)]
#[repr(C)]
pub struct SyncObject
{
    position_x: f32,
    position_y: f32,
    position_z: f32,

    // Bit 0-1 selects the quaternion that is biggest (and not part of the array)
    // Bit 2    1 == active object, 0 = not active
    // Bit 3-7 not used
    // Bit 8-15 counts the number of times this object is read from unity. Reset when a new packet is arrived.
    // If the counter reaches 0xff the object is flagged as timed out
    flag: u16,
    id: u16,

    // Compressed quaternion
    rotation: [i16; 3],

    // Model
    model_id: u16
}

// Low precision - 10 bytes
#[derive(Clone, Copy)]
#[repr(C)]
pub struct SyncObjectLP
{
    pos_x: i16,
    pos_y: i16,
    pos_z: i16,

    // The 3 smallest values of a quaternion  +- 1/sqrt(2) as 10 bit fixed point
    quaternion: u16,
    alignment: u8, // WEIRD ALIGNMENT!!!
}

// Synced object car
#[derive(Clone, Copy)]
#[repr(C)]
pub struct SyncObjectCar
{
    car: SyncObject,
    wheels: [SyncObjectLP; 4],
}

#[derive(Clone, Copy)]
#[repr(C)]
pub struct SyncObjectSpeed
{
    // Speed in m/s
    dx: f32,
    dy: f32,
    dz: f32,

    // Acceleration in m/s^2
    ddx: f32,
    ddy: f32,
    ddz: f32,

    // Time in ms from start of simulation for this value
    timestamp: u32
}

#[derive(Clone, Copy)]
#[repr(C)]
pub struct ExternalIoData
{
    // Input part
    pub sensor_in: [f32; 16],
    pub analog_in_raw: [i16; 16],
    pub input_bits: [u32; 7],

    pub motion_status: u32,
    pub input_i_raw: [i32; 16],
    pub actuator_length: [f32; 6],
    input_dummy: [u32; 10],

    // Output part
    pub sensor_out: [f32; 16],
    pub analog_out_raw: [i16; 16],
    pub output_bits: [u32; 7],

    pub motion_command: u32,

    pub motion_x: f32,
    pub motion_y: f32,
    pub motion_z: f32,

    pub motion_yaw: f32,
    pub motion_pitch: f32,
    pub motion_roll: f32,

    output_dummy: [u32; 26],
}

#[derive(Clone, Copy, Default, Debug)]
#[repr(C)]
pub struct SyncPositionEvent
{
    pub flag: u32,
    pub event_type: u32,
    pub data: f32,

    pub x: f32,
    pub y: f32,
    pub z: f32,
}

// Main shared memory struct
#[derive(Clone, Copy)]
#[repr(C)]
pub struct VisualSharedMemory
{
    dummy_spin_to_visual: i32,
    car_qt_watchdog: i32,
    visual_system_type: i32,
    ego_model_type: i32,

    // Vehicle data
    pub vehicle_data: VehicleData,

    dummy_spin_from_visual: i32,
    unity_watchdog: i32,
    dummy21: i32,
    dummy22: i32,

    // Old and new collision data
    collision_data: CollisionData,
    car_wheel_collision: CarWheelCollision,
    car_wheel_mu: CarWheelMU,

    // Unity vehicle data
    vehicle_unity_data: VehicleData,

    dummy_spin_can_lock: i32,
    dummy31: i32,
    dummy32: i32,
    dummy33: i32,

    // Can bus data
    pub updated: CanMask,
    pub cdata: [CanBase; 2048],

    dummy_spin_sync_unity: i32,

    // Synced cars
    pub synced_cars_mask: [u32; 128],
    pub synced_cars: [SyncObjectCar; 4096],

    pub unity_build_id: i32,
    dummy41: i32,
    dummy42: i32,

    pub unity_build_name: [u8; 256],

    // Synced objects
    pub synced_objects_mask: [u32; 256],
    pub synced_objects: [SyncObject; 8192],

    pub active_cars_mask: [u32; 128],
    pub active_objects_mask: [u32; 256],

    // Car speed
    pub car_speed: [SyncObjectSpeed; 4096],

    dummy_spin_external_io: i32,
    dummy51: i32,
    dummy52: i32,
    dummy53: i32,

    pub exio: ExternalIoData,

    // Event ring buffers
    pub dummy_event_buffer_lock: i32,
    pub event_buffer_head: i32,
    pub event_buffer_tail: i32,
    pub event_buffer: [SyncPositionEvent; 1024],

    pub dummy_event_copy_buffer_lock: i32,
    pub event_copy_buffer_head: i32,
    pub event_copy_buffer_tail: i32,
    pub event_copy_buffer: [SyncPositionEvent; 1024],

    pub dummy_event_buffer2_lock: i32,
    pub event_buffer2_head: i32,
    pub event_buffer2_tail: i32,
    pub event_buffer2: [SyncPositionEvent; 1024],

    // The rest of shared memory data is not interesting for the assessment system...
}
impl VisualSharedMemory {
    pub fn new() -> Self {
        Self {
            dummy_spin_to_visual: 0,
            car_qt_watchdog: 0,
            visual_system_type: 0,
            ego_model_type: 0,
            vehicle_data: VehicleData {
                position_x: 0.0,
                position_y: 0.0,
                position_z: 0.0,
                idx: 0,
                direction_x: 0.0,
                direction_y: 0.0,
                direction_z: 0.0,
                direction_w: 0.0,
                speed_vector_x: 0.0,
                speed_vector_y: 0.0,
                speed_vector_z: 0.0,
                gear: 0,
                offset_yaw: 0.0,
                offset_pitch: 0.0,
                offset_roll: 0.0,
                steering_wheel: 0.0,
                brake: 0.0,
                clutch: 0.0,
                throttle: 0.0,
                flag: 0,
                rpm: 0.0,
                torque: 0.0,
                sound_event_engine: 0,
                speed: 0.0,
                sound_l1: 0,
                sound_l2: 0,
                sound_r1: 0,
                sound_r2: 0,
                frame_idx: 0,
                park_brake: 0.0,
                alpha_l1: 0.0,
                alpha_l2: 0.0,
                alpha_r1: 0.0,
                alpha_r2: 0.0,
                kappa_l1: 0.0,
                kappa_l2: 0.0,
                kappa_r1: 0.0,
                kappa_r2: 0.0,
                avy_l1: 0.0,
                avy_l2: 0.0,
                avy_r1: 0.0,
                avy_r2: 0.0,
                bits_from_car1: 0,
                bits_from_car2: 0,
                bits_from_simulation1: 0,
                bits_from_simulation2: 0
            },
            dummy_spin_from_visual: 0,
            unity_watchdog: 0,
            dummy21: 0,
            dummy22: 0,
            collision_data: CollisionData { wheel_collision: [HyperPlane3 {
                x: 0.0,
                y: 0.0,
                z: 0.0,
                dummy: 0.0
            }; 4] },
            car_wheel_collision: CarWheelCollision {
                l1: [0.0; 16],
                r1: [0.0; 16],
                l2: [0.0; 16],
                r2: [0.0; 16]
            },
            car_wheel_mu: CarWheelMU {
                l1: 0.0,
                l2: 0.0,
                r1: 0.0,
                r2: 0.0
            },
            vehicle_unity_data: VehicleData {
                position_x: 0.0,
                position_y: 0.0,
                position_z: 0.0,
                idx: 0,
                direction_x: 0.0,
                direction_y: 0.0,
                direction_z: 0.0,
                direction_w: 0.0,
                speed_vector_x: 0.0,
                speed_vector_y: 0.0,
                speed_vector_z: 0.0,
                gear: 0,
                offset_yaw: 0.0,
                offset_pitch: 0.0,
                offset_roll: 0.0,
                steering_wheel: 0.0,
                brake: 0.0,
                clutch: 0.0,
                throttle: 0.0,
                flag: 0,
                rpm: 0.0,
                torque: 0.0,
                sound_event_engine: 0,
                speed: 0.0,
                sound_l1: 0,
                sound_l2: 0,
                sound_r1: 0,
                sound_r2: 0,
                frame_idx: 0,
                park_brake: 0.0,
                alpha_l1: 0.0,
                alpha_l2: 0.0,
                alpha_r1: 0.0,
                alpha_r2: 0.0,
                kappa_l1: 0.0,
                kappa_l2: 0.0,
                kappa_r1: 0.0,
                kappa_r2: 0.0,
                avy_l1: 0.0,
                avy_l2: 0.0,
                avy_r1: 0.0,
                avy_r2: 0.0,
                bits_from_car1: 0,
                bits_from_car2: 0,
                bits_from_simulation1: 0,
                bits_from_simulation2: 0
            },
            dummy_spin_can_lock: 0,
            dummy31: 0,
            dummy32: 0,
            dummy33: 0,
            updated: CanMask { mask: [0; 64] },
            cdata: [CanBase {
                packet_time: 0,
                delta_time: 0.0,
                dummy: 0,
                can_packet: CanPacket {
                    id: 0,
                    flag: 0,
                    data: [0; 8]
                }
            }; 2048],
            dummy_spin_sync_unity: 0,
            synced_cars_mask: [0; 128],
            synced_cars: [SyncObjectCar { car: SyncObject {
                position_x: 0.0,
                position_y: 0.0,
                position_z: 0.0,
                flag: 0,
                id: 0,
                rotation: [0; 3],
                model_id: 0
            }, wheels: [SyncObjectLP {
                pos_x: 0,
                pos_y: 0,
                pos_z: 0,
                quaternion: 0,
                alignment: 0
            }; 4] }; 4096],
            unity_build_id: 0,
            dummy41: 0,
            dummy42: 0,
            unity_build_name: [0; 256],
            synced_objects_mask: [0; 256],
            synced_objects: [SyncObject {
                position_x: 0.0,
                position_y: 0.0,
                position_z: 0.0,
                flag: 0,
                id: 0,
                rotation: [0; 3],
                model_id: 0
            }; 8192],
            active_cars_mask: [0; 128],
            active_objects_mask: [0; 256],
            car_speed: [SyncObjectSpeed {
                dx: 0.0,
                dy: 0.0,
                dz: 0.0,
                ddx: 0.0,
                ddy: 0.0,
                ddz: 0.0,
                timestamp: 0
            }; 4096],
            dummy_spin_external_io: 0,
            dummy51: 0,
            dummy52: 0,
            dummy53: 0,
            exio: ExternalIoData {
                sensor_in: [0.0; 16],
                analog_in_raw: [0; 16],
                input_bits: [0; 7],
                motion_status: 0,
                input_i_raw: [0; 16],
                actuator_length: [0.0; 6],
                input_dummy: [0; 10],
                sensor_out: [0.0; 16],
                analog_out_raw: [0; 16],
                output_bits: [0; 7],
                motion_command: 0,
                motion_x: 0.0,
                motion_y: 0.0,
                motion_z: 0.0,
                motion_yaw: 0.0,
                motion_pitch: 0.0,
                motion_roll: 0.0,
                output_dummy: [0; 26]
            },
            dummy_event_buffer_lock: 0,
            event_buffer_head: 0,
            event_buffer_tail: 0,
            event_buffer: [SyncPositionEvent {
                flag: 0,
                event_type: 0,
                data: 0.0,
                x: 0.0,
                y: 0.0,
                z: 0.0
            }; 1024],
            dummy_event_copy_buffer_lock: 0,
            event_copy_buffer_head: 0,
            event_copy_buffer_tail: 0,
            event_copy_buffer: [SyncPositionEvent {
                flag: 0,
                event_type: 0,
                data: 0.0,
                x: 0.0,
                y: 0.0,
                z: 0.0
            }; 1024],
            dummy_event_buffer2_lock: 0,
            event_buffer2_head: 0,
            event_buffer2_tail: 0,
            event_buffer2: [SyncPositionEvent {
                flag: 0,
                event_type: 0,
                data: 0.0,
                x: 0.0,
                y: 0.0,
                z: 0.0
            }; 1024],
        }
    }
}

// WARNING: This struct contains some implemented logic as well
#[repr(C)]
pub struct RingBuffer<T: Default + Clone + Copy>
{
    pub buffer: [T; 1024],
    pub head_index: usize,
    pub tail_index: usize,
}

impl<T: Default + Clone + Copy> RingBuffer<T> {
    pub fn new() -> Self {
        RingBuffer {
            buffer: [Default::default(); 1024],
            head_index: 0,
            tail_index: 0,
        }
    }

    pub fn from(buffer: [T; 1024], head: usize, tail: usize) -> Self {
        RingBuffer {
            buffer,
            head_index: head,
            tail_index: tail,
        }
    }

    pub fn clear(&mut self) {
        self.head_index = 0;
        self.tail_index = 0;
    }

    pub fn push(&mut self, value: T) -> bool {
        // Determine new head
        let mut new_index = self.head_index + 1;
        if new_index >= 1024
        {
            new_index = 0;
        }

        // Check if buffer is full
        if new_index == self.tail_index
        {
            return false;
        }

        // Push new value
        self.buffer[new_index] = value;
        self.head_index = new_index;
        return true;
    }

    pub fn pop(&mut self) -> Option<T> {
        // The buffer is empty
        if self.head_index == self.tail_index {
            return None;
        }

        let value = self.buffer[self.tail_index];

        // Push new value
        self.tail_index += 1;
        if self.tail_index >= 1024
        {
            self.tail_index = 0;
        }

        return Some(value);
    }
}
