﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

using AssessmentSystemApi;
using System;

public class DetectionAgents : MonoBehaviour
{
    // Knowledge graph shared memory
    private System.IntPtr graph = System.IntPtr.Zero;

    [Header("CAR DETECTION AGENT")]
    [Header("Trigger with hysteresis i.e. two triggers")]
    public ObjectDetectionTrigger carEnterTrigger;
    public ObjectDetectionTrigger carExitTrigger;
    [ReadOnly]
    public int carLayer = 17;
    [ReadOnly]
    public string carTag = "TrafficCar";

    [Header("Cars in range")]
    public float carsInRangeUpdateFrequencyMs = 200f;
    [Space(10)]
    [ReadOnly]
    public List<GameObject> carsInRange = new List<GameObject>();
    private List<CarInRange> carsInRangeCached = new List<CarInRange>();


    [Header("CAR RELATIONS AGENT")]
    public float egoWidthOffset = 2f;
    public float egoLengthOffset = 3f;
    [Range(0f, 1f)]
    public float facingSameDirectionThreshold = 0.5f;


    [Header("LANE DETECTION AGENT")]
    public LayerMask laneObjectsLayerMask = 1 << 24;
    public Vector3 detectionRayOffset = new Vector3(0, 2.0f, 0);
    public float detectionRayLegth = 10.0f;
    public float detectionFrequencyMs = 100.0f;
    public float lastLaneReplacementIntervalMs = 1000.0f;
    public int laneFallbackSpeedLimit = 50;

    [Header("Possible lanes to be in")]
    [ReadOnly]
    public LaneTag[] candidateLanes;

    [ReadOnly]
    public LaneTag previousLane;
    [ReadOnly]
    public LaneTag currentLane;
    [ReadOnly]
    public bool laneRemoved;


    [Header("ROAD SIGN DETECTION AGENT")]
    public ObjectDetectionTrigger signEnterTrigger;
    [ReadOnly]
    public int signLayer = 26;
    [ReadOnly]
    public string signTag = "RoadSign";
    [ReadOnly]
    public List<GameObject> activeRoadSigns = new List<GameObject>();


    [Header("CROSSWALK DETECTION AGENT")]
    [Header("Trigger with hysteresis i.e. two triggers")]
    public ObjectDetectionTrigger crosswalkEnterTrigger;
    public ObjectDetectionTrigger crosswalkExitTrigger;

    [ReadOnly]
    public int crosswalkLayer = 27;
    [ReadOnly]
    public string crosswalkTag = "Crosswalk";

    [ReadOnly]
    public List<GameObject> crosswalksInRange = new List<GameObject>();


    [Header("INTERSECTION DETECTION AGENT")]
    [Header("Trigger with hysteresis i.e. two triggers")]
    public ObjectDetectionTrigger intersectionEnterTrigger;
    public ObjectDetectionTrigger intersectionExitTrigger;

    [ReadOnly]
    public int yieldLaneIdOffset = 100;
    [ReadOnly]
    public int intersectionLayer = 29;
    public float yieldUpdateFrequencyMs = 100.0f;


    [ReadOnly]
    public List<GameObject> intersectionsInRange = new List<GameObject>();
    [ReadOnly]
    private LaneTag currentIntersectionEnterLane;
    [ReadOnly]
    public List<LaneTag> activeYields = new List<LaneTag>();

    private List<GameObject> solvedIntersectionsInRange = new List<GameObject>();


    [Header("DEBUG")]
    public bool printDetectedLanes = false;
    public bool printDetectedCars = false;
    public bool printDetectedCrosswalks = false;
    public bool printDetectedIntersections = false;
    public bool printDetectedYields = false;


    // Start is called before the first frame update
    void Start()
    {
        // Assign callbacks
        carEnterTrigger.OnObjectEnterEvent += CarTriggerEnter;
        carExitTrigger.OnObjectExitEvent += CarTriggerExit;

        signEnterTrigger.OnObjectEnterEvent += SignTriggerEnter;

        crosswalkEnterTrigger.OnObjectEnterEvent += CrosswalkTriggerEnter;
        crosswalkExitTrigger.OnObjectExitEvent += CrosswalkTriggerExit;

        intersectionEnterTrigger.OnObjectEnterEvent += IntersectionTriggerEnter;
        intersectionExitTrigger.OnObjectExitEvent += IntersectionTriggerExit;

        // Initialize knowledge graph
        graph = SharedMemory.initialize_knowledge_graph();
        SharedMemory.reset_knowledge_graph(graph);
        KnowledgeGraph.add_ego_node(0, 0, new Ego(), graph);  // Ego acts as the graph root

        // Start lane detection agent
        StartCoroutine(LaneDetectionAgent());

        // Start car relations agent
        StartCoroutine(CarRelationsAgent());

        // Start intersection agent
        StartCoroutine(IntersectionAgent());
    }

    void OnDestroy()
    {
        if (graph != System.IntPtr.Zero)
        {
            SharedMemory.reset_knowledge_graph(graph);
            SharedMemory.release_knowledge_graph(graph);
            graph = System.IntPtr.Zero;
        }
    }


    // Helper methods
    private ulong GetCurrentTimestampMs()
    {
        var timestamp = DllData.externalGo.FromExternalModel().idx;
        if (timestamp == 0)
        {
            // Takes care of timestamping when no visual shared memory is connected (only for dbg purposes)
            return (ulong) (Time.realtimeSinceStartup * 1000f);
        }
        else
        {
            return (ulong) timestamp;
        }
    }

    private Vector3f Vector3ToVector3f(Vector3 vector)
    {
        return new Vector3f { x = vector.x, y = vector.y, z = vector.z };
    }


    // CAR DETECTION AGENT
    // Trigger enter and exit
    private void CarTriggerEnter(GameObject go)
    {
        if (go.layer == carLayer &&
            go.tag == carTag)
        {
            // Add car to the cars in range list
            var car = new CarInRange();
            car.Id = (uint)go.GetComponent<Way.TrafficSystem.AICarController>().CarID;
            car.Rigidbody = go.GetComponent<Rigidbody>();
            car.GameObject = go;

            carsInRangeCached.Add(car);
            carsInRange.Add(go);
        }
    }

    private void CarTriggerExit(GameObject go)
    {
        if (go.layer == carLayer &&
            go.tag == carTag)
        {
            if (carsInRange.Contains(go))
            {
                var id = go.GetComponent<Way.TrafficSystem.AICarController>().CarID;
                carsInRangeCached.RemoveAll((item) => { return item.Id == id; });
                carsInRange.Remove(go);

                // Remove from knowledge graph too
                var timestamp = (ulong)GetCurrentTimestampMs();
                var carNodeType = TypeIndex.OfGraphNode(KnowledgeGraphNode.Car);
                KnowledgeGraph.remove_car_node(timestamp, carNodeType, (uint)id, graph);
            }
        }
    }

    // CAR RELATIONS AGENT
    private IEnumerator CarRelationsAgent()
    {
        while (true)
        {
            yield return new WaitForSeconds(carsInRangeUpdateFrequencyMs / 1000f);

            var timestamp = GetCurrentTimestampMs();
            var carNodeType = TypeIndex.OfGraphNode(KnowledgeGraphNode.Car);
            var egoNodeType = TypeIndex.OfGraphNode(KnowledgeGraphNode.Ego);

            foreach (var go in carsInRangeCached)
            {
                // Update car data
                var carId = go.Id;
                var carDirection = go.GameObject.transform.TransformDirection(Vector3.forward);
                var carPosition = go.GameObject.transform.position;
                var carSpeed = go.Rigidbody.velocity.magnitude;

                var car = new Car { carDirection = Vector3ToVector3f(carDirection), carPosition = Vector3ToVector3f(carPosition), carSpeed = carSpeed };
                KnowledgeGraph.add_car_node(timestamp, carId, car, graph);


                // Detect in front or behind relations
                // Ego data
                var egoPosition = transform.parent.position;
                var egoDirection = Vector3.zero;

                // Difference between runtime EGO and E36 car facing direction
                if (Application.isEditor)
                {
                    egoDirection = transform.parent.TransformDirection(Vector3.forward);
                }
                else
                {
                    egoDirection = transform.parent.TransformDirection(Vector3.right);
                }


                if (CarRelationUtils.FacingSameDirection(carDirection, egoDirection, facingSameDirectionThreshold))
                {
                    var relation = CarRelationUtils.CarPositionRelativeToEgo(egoPosition, egoDirection, carPosition, egoWidthOffset, egoLengthOffset);

                    if (relation == CarRelationUtils.CarPosition.Front || relation == CarRelationUtils.CarPosition.FrontLeft || relation == CarRelationUtils.CarPosition.FrontRight)
                    {
                        KnowledgeGraph.add_isinfront_relation(timestamp, carId, new isInFront(), egoNodeType, 0, carNodeType, carId, graph);
                    }
                    else if (relation == CarRelationUtils.CarPosition.Behind || relation == CarRelationUtils.CarPosition.BehindLeft || relation == CarRelationUtils.CarPosition.BehindRight)
                    {
                        KnowledgeGraph.add_isbehind_relation(timestamp, carId, new isBehind(), egoNodeType, 0, carNodeType, carId, graph);
                    }
                    else if (relation == CarRelationUtils.CarPosition.Right)
                    {
                        KnowledgeGraph.add_isonright_relation(timestamp, carId, new isOnRight(), egoNodeType, 0, carNodeType, carId, graph);
                    }
                    else if (relation == CarRelationUtils.CarPosition.Left)
                    {
                        KnowledgeGraph.add_isonleft_relation(timestamp, carId, new isOnLeft(), egoNodeType, 0, carNodeType, carId, graph);
                    }

                    if (printDetectedCars) { Debug.LogError("CAR: Ego is " + relation + " the car."); }
                }
            }
        }
    }


    // LANE DETECTION AGENT
    private IEnumerator LaneDetectionAgent()
    {
        var timer = 0.0f;

        // Run detection until stopped explicitly
        while (true)
        {
            yield return new WaitForSeconds(detectionFrequencyMs / 1000f);
            timer += detectionFrequencyMs;

            var timestamp = GetCurrentTimestampMs();
            var laneNodeType = TypeIndex.OfGraphNode(KnowledgeGraphNode.Lane);
            var roadNodeType = TypeIndex.OfGraphNode(KnowledgeGraphNode.Road);
            var egoNodeType = TypeIndex.OfGraphNode(KnowledgeGraphNode.Ego);

            var newCurrentLanes = new List<LaneTag>();

            // Cast a ray down to the road 
            var rayOrigin = transform.position + detectionRayOffset;
            var hits = Physics.RaycastAll(rayOrigin, Vector3.down, detectionRayLegth, laneObjectsLayerMask, QueryTriggerInteraction.Ignore);

            // Does the ray intersect any lane objects?
            if (hits.Length > 0)
            {
                // Just one lane, obvious
                if (hits.Length == 1)
                {
                    // Fetch the lane and check if its different to the current one
                    var parent = hits[0].transform.parent;
                    var tag = parent.GetComponent<LaneTag>();

                    if (currentLane != tag)
                    {
                        // Remove old nodes first
                        if (currentLane != null)
                        {
                            // Lane
                            var oldLaneId = (uint)currentLane.id;
                            KnowledgeGraph.remove_lane_node(timestamp, laneNodeType, oldLaneId, graph);

                            // Intersection or road as well
                            var oldRoadTag = currentLane.transform.parent.GetComponent<RoadTag>();
                            if (oldRoadTag != null)
                            {
                                var roadId = (uint)oldRoadTag.id;
                                KnowledgeGraph.remove_road_node(timestamp, roadNodeType, roadId, graph);
                            }
                        }


                        // Add new lane to the knowledge graph
                        // Lane
                        var laneId = (uint)tag.id;
                        var laneSpeedLimit = tag.speedLimit;
                        KnowledgeGraph.add_lane_node(timestamp, laneId, new Lane { laneId = (int) laneId, laneName = String2Bytes.Pack(tag.name), laneSpeedLimit = laneSpeedLimit }, graph);
                        KnowledgeGraph.add_isin_relation(timestamp, laneId, new isIn(), egoNodeType, 0, laneNodeType, laneId, graph);

                        // Road it belongs to
                        var roadTag = parent.parent.GetComponent<RoadTag>();
                        if (roadTag != null)
                        {
                            var roadId = (uint)roadTag.id;
                            KnowledgeGraph.add_road_node(timestamp, roadId, new Road { roadId = (int)roadId, roadName = String2Bytes.Pack(roadTag.name) }, graph);
                            KnowledgeGraph.add_ison_relation(timestamp, roadId, new isOn(), egoNodeType, 0, roadNodeType, roadId, graph);
                        }


                        // Check possible replacement of buffered previous lane
                        if (candidateLanes.Length == 0)
                        {
                            previousLane = tag;
                        }
                        else if (timer > lastLaneReplacementIntervalMs)
                        {
                            previousLane = tag;
                            timer = 0;
                        }

                        // Update current lane (which implies deleting all active road signs)
                        DeleteAllRoadSigns();
                        currentLane = tag;
                        laneRemoved = false;

                        newCurrentLanes.Add(tag);

                        if (printDetectedLanes) { Debug.LogError("LANE: " + tag.gameObject.name); }
                    }
                }
                // More lanes
                else
                {
                    // Check possible outcomes
                    foreach (var hit in hits)
                    {
                        var parent = hit.transform.parent;
                        var tag = parent.GetComponent<LaneTag>();

                        if (previousLane != null && previousLane.outgoingLanes.Contains(tag) || (previousLane != null && previousLane == tag))
                        {
                            newCurrentLanes.Add(tag);
                        }
                    }
                }
            }
            else
            {
                if (!laneRemoved && currentLane != null)
                {
                    // Remove lane from knowledge graph too
                    var laneId = (uint)currentLane.id;
                    KnowledgeGraph.remove_lane_node(timestamp, laneNodeType, laneId, graph);

                    // Intersection or road as well
                    var roadTag = currentLane.transform.parent.GetComponent<RoadTag>();
                    if (roadTag != null)
                    {
                        var roadId = (uint)roadTag.id;
                        KnowledgeGraph.remove_road_node(timestamp, roadNodeType, roadId, graph);
                    }

                    DeleteAllRoadSigns();
                    currentLane = null;
                    previousLane = null;
                    laneRemoved = true;
                }
            }

            // Update detected lanes
            candidateLanes = newCurrentLanes.ToArray();
        }
    }

    // ROAD SIGN DETECTION AGENT
    // Trigger enter
    private void SignTriggerEnter(GameObject go)
    {
        if (go.layer == signLayer &&
            go.tag == signTag &&
            !activeRoadSigns.Contains(go))
        {
            var timestamp = GetCurrentTimestampMs();

            var roadSign = go.GetComponent<RoadSignTag>();
            if (roadSign.belongsTo == currentLane)
            {
                AddRoadSign(timestamp, roadSign, go, currentLane);
            }
        }
    }

    // Deletion when ego leaves the current lane
    private void DeleteAllRoadSigns()
    {
        var timestamp = GetCurrentTimestampMs();
        foreach (var sign in activeRoadSigns)
        {
            // Deal with all road sign types separately
            var roadSign = sign.GetComponent<RoadSignTag>();
            DeleteRoadSign(timestamp, roadSign);
        }

        activeRoadSigns.Clear();
    }

    private void AddRoadSign(ulong timestamp, RoadSignTag roadSign, GameObject roadSignGo, LaneTag lane)
    {
        var laneNodeType = TypeIndex.OfGraphNode(KnowledgeGraphNode.Lane);
        var laneId = (uint)currentLane.id;
        var signId = (uint)roadSign.id;

        // Deal with all road sign types
        // NO ENTRY SIGN
        if (roadSign.type == RoadSignType.NoEntrySign)
        {
            var signNodeType = TypeIndex.OfGraphNode(KnowledgeGraphNode.NoEntrySign);
            KnowledgeGraph.add_noentrysign_node(timestamp, signId, new NoEntrySign(), graph);
            KnowledgeGraph.add_belongsto_relation(timestamp, signId, new belongsTo(), signNodeType, signId, laneNodeType, laneId, graph);
            activeRoadSigns.Add(roadSignGo);
        }
        // PRIORITY ROAD SIGN
        else if (roadSign.type == RoadSignType.PriorityRoadSign)
        {
            var signNodeType = TypeIndex.OfGraphNode(KnowledgeGraphNode.PriorityRoadSign);
            KnowledgeGraph.add_priorityroadsign_node(timestamp, signId, new PriorityRoadSign(), graph);
            KnowledgeGraph.add_belongsto_relation(timestamp, signId, new belongsTo(), signNodeType, signId, laneNodeType, laneId, graph);
            activeRoadSigns.Add(roadSignGo);
        }
        // SPEED LIMIT SIGN
        if (roadSign.type == RoadSignType.SpeedLimitSign)
        {
            // Speed limit value
            var speedLimit = int.Parse(roadSign.GetProperty("limit"));

            var signNodeType = TypeIndex.OfGraphNode(KnowledgeGraphNode.SpeedLimitSign);
            KnowledgeGraph.add_speedlimitsign_node(timestamp, signId, new SpeedLimitSign { signSpeedLimit = speedLimit }, graph);
            KnowledgeGraph.add_belongsto_relation(timestamp, signId, new belongsTo(), signNodeType, signId, laneNodeType, laneId, graph);
            activeRoadSigns.Add(roadSignGo);
        }
        // STOP SIGN
        else if (roadSign.type == RoadSignType.StopSign)
        {
            var signNodeType = TypeIndex.OfGraphNode(KnowledgeGraphNode.StopSign);
            KnowledgeGraph.add_stopsign_node(timestamp, signId, new StopSign(), graph);
            KnowledgeGraph.add_belongsto_relation(timestamp, signId, new belongsTo(), signNodeType, signId, laneNodeType, laneId, graph);
            activeRoadSigns.Add(roadSignGo);
        }
        // YIELD SIGN
        else if (roadSign.type == RoadSignType.YieldSign)
        {
            var signNodeType = TypeIndex.OfGraphNode(KnowledgeGraphNode.YieldSign);
            KnowledgeGraph.add_yieldsign_node(timestamp, signId, new YieldSign(), graph);
            KnowledgeGraph.add_belongsto_relation(timestamp, signId, new belongsTo(), signNodeType, signId, laneNodeType, laneId, graph);
            activeRoadSigns.Add(roadSignGo);
        }
    }

    private void DeleteRoadSign(ulong timestamp, RoadSignTag roadSign)
    {
        var signId = (uint)roadSign.id;

        // NO ENTRY SIGN
        if (roadSign.type == RoadSignType.NoEntrySign)
        {
            var signNodeType = TypeIndex.OfGraphNode(KnowledgeGraphNode.NoEntrySign);
            KnowledgeGraph.remove_noentrysign_node(timestamp, signNodeType, signId, graph);
        }
        // PRIORITY ROAD
        else if (roadSign.type == RoadSignType.PriorityRoadSign)
        {
            var signNodeType = TypeIndex.OfGraphNode(KnowledgeGraphNode.PriorityRoadSign);
            KnowledgeGraph.remove_priorityroadsign_node(timestamp, signNodeType, signId, graph);
        }
        // SPEED LIMIT SIGN
        else if (roadSign.type == RoadSignType.SpeedLimitSign)
        {
            var signNodeType = TypeIndex.OfGraphNode(KnowledgeGraphNode.SpeedLimitSign);
            KnowledgeGraph.remove_speedlimitsign_node(timestamp, signNodeType, signId, graph);
        }
        // STOP SIGN
        else if (roadSign.type == RoadSignType.StopSign)
        {
            var signNodeType = TypeIndex.OfGraphNode(KnowledgeGraphNode.StopSign);
            KnowledgeGraph.remove_stopsign_node(timestamp, signNodeType, signId, graph);
        }
        // YIELD SIGN
        else if (roadSign.type == RoadSignType.YieldSign)
        {
            var signNodeType = TypeIndex.OfGraphNode(KnowledgeGraphNode.YieldSign);
            KnowledgeGraph.remove_yieldsign_node(timestamp, signNodeType, signId, graph);
        }
    }

    
    // CROSSWALK DETECTION AGENT
    // Trigger enter and exit
    private void CrosswalkTriggerEnter(GameObject go)
    {
        if (go.layer == crosswalkLayer &&
            go.tag == crosswalkTag)
        {
            crosswalksInRange.Add(go);

            // Add crosswalk node
            var timestamp = GetCurrentTimestampMs();
            var crosswalkNodeType = TypeIndex.OfGraphNode(KnowledgeGraphNode.Crosswalk);
            var egoNodeType = TypeIndex.OfGraphNode(KnowledgeGraphNode.Ego);

            var tag = go.GetComponent<CrosswalkTag>();
            var crosswalkId = (uint)tag.id;
            var crosswalkPosition = Vector3ToVector3f(go.transform.position);

            KnowledgeGraph.add_crosswalk_node(timestamp, crosswalkId, new Crosswalk { crosswalkId = (int)crosswalkId, crosswalkPosition = crosswalkPosition, crosswalkPedestrianCount = tag.pedestrianCount }, graph);
            KnowledgeGraph.add_passesover_relation(timestamp, crosswalkId, new passesOver(), egoNodeType, 0, crosswalkNodeType, crosswalkId, graph);

            if (printDetectedCrosswalks) { Debug.LogError("CROSSWALK: " + tag.gameObject.name); }
        }
    }

    private void CrosswalkTriggerExit(GameObject go)
    {
        if (go.layer == crosswalkLayer &&
            go.tag == crosswalkTag)
        {
            crosswalksInRange.Remove(go);

            // Remove from knowledge graph too
            var timestamp = (ulong)GetCurrentTimestampMs();
            var crosswalkId = (uint)go.GetComponent<CrosswalkTag>().id;
            var crosswalkNodeType = TypeIndex.OfGraphNode(KnowledgeGraphNode.Crosswalk);

            KnowledgeGraph.remove_crosswalk_node(timestamp, crosswalkNodeType, crosswalkId, graph);
        }
    }


    // INTERSECTION DETECTION AGENT
    // Trigger enter and exit
    private void IntersectionTriggerEnter(GameObject go)
    {
        if (go.layer == intersectionLayer && go.name == "ego-trigger")
        {
            currentIntersectionEnterLane = currentLane;
            intersectionsInRange.Add(go.transform.parent.gameObject);

            var timestamp = GetCurrentTimestampMs();
            var intersectionId = (uint)go.transform.parent.GetComponent<IntersectionTag>().id;
            var intersectionPosition = Vector3ToVector3f(go.transform.position);
            var egoNodeType = TypeIndex.OfGraphNode(KnowledgeGraphNode.Ego);
            var intersectionNodeType = TypeIndex.OfGraphNode(KnowledgeGraphNode.Intersection);

            KnowledgeGraph.add_intersection_node(timestamp, intersectionId, new Intersection { intersectionId = (int)intersectionId, intersectionPosition = intersectionPosition }, graph);
            KnowledgeGraph.add_isat_relation(timestamp, intersectionId, new isAt(), egoNodeType, 0, intersectionNodeType, intersectionId, graph);

            if (printDetectedIntersections) { Debug.LogError("INTERSECTION: " + go.transform.parent.name); }
        }
    }

    private void IntersectionTriggerExit(GameObject go)
    {
        if (go.layer == intersectionLayer && go.name == "ego-trigger")
        {
            currentIntersectionEnterLane = null;
            intersectionsInRange.Remove(go.transform.parent.gameObject);

            // Remove from knowledge graph too
            var timestamp = (ulong) GetCurrentTimestampMs();
            var intersectionId = (uint)go.transform.parent.GetComponent<IntersectionTag>().id;
            var intersectionNodeType = TypeIndex.OfGraphNode(KnowledgeGraphNode.Intersection);

            KnowledgeGraph.remove_intersection_node(timestamp, intersectionNodeType, intersectionId, graph);
            DeleteAllActiveYields();

            if (printDetectedIntersections) { Debug.LogError("INTERSECTION_REMOVE: " + go.transform.parent.name); }
        }
    }

    private int TurnDirection2Bitmask(TurnDirection direction)
    {
        if (direction == TurnDirection.Right)
        {
            return 1;
        }
        else if (direction == TurnDirection.Straight)
        {
            return 2;
        }
        else if (direction == TurnDirection.Left)
        {
            return 4;
        }
        else
        {
            return 0;
        }
    }

    private void DeleteAllActiveYields()
    {
        var timestamp = GetCurrentTimestampMs();
        var laneNodeType = TypeIndex.OfGraphNode(KnowledgeGraphNode.Lane);

        foreach (var yield in activeYields)
        {
            var laneId = (uint)yield.id;
            laneId += (uint)yieldLaneIdOffset;

            KnowledgeGraph.remove_lane_node(timestamp, laneNodeType, laneId, graph);

            if (printDetectedYields) { Debug.LogError("YIELDS_REMOVE: " + yield); }
        }

        solvedIntersectionsInRange.Clear();
        activeYields.Clear();
    }

    private IEnumerator IntersectionAgent()
    {
        // Run detection until stopped explicitly
        while (true)
        {
            yield return new WaitForSeconds(yieldUpdateFrequencyMs / 1000f);

            // Determine active yielding
            foreach (var go in intersectionsInRange)
            {
                var tag = go.GetComponent<IntersectionTag>();
                var timestamp = GetCurrentTimestampMs();
                var laneNodeType = TypeIndex.OfGraphNode(KnowledgeGraphNode.Lane);

                // 3 Way uncontrolled intersection
                if (tag.type == IntersectionType.Uncontrolled && tag.size == IntersectionSize._3Way)
                {
                    var yieldToList = tag.DefaultHasToYieldTo(currentLane);
                    foreach (var yieldTo in yieldToList)
                    {
                        var yieldBitMask = TurnDirection2Bitmask(yieldTo.Item1);
                        var lane = yieldTo.Item2;

                        var lane_id = (uint)lane.id;
                        lane_id += (uint)yieldLaneIdOffset;

                        if (printDetectedYields) { Debug.LogError("UNCONTROLLED 3-WAY: " + tag.gameObject.name); }

                        KnowledgeGraph.add_lane_node(timestamp, lane_id, new Lane { laneId = (int)lane_id, laneName = String2Bytes.Pack(lane.name), laneSpeedLimit = lane.speedLimit }, graph);
                        KnowledgeGraph.add_yieldstocarin_relation(timestamp, lane_id, new yieldsToCarIn { ifDirection = yieldBitMask }, laneNodeType, (uint)currentLane.id, laneNodeType, lane_id, graph);

                        activeYields.Add(lane);
                    }
                }

                // Traffic light controlled intersection
                var trafficLightsTag = go.GetComponent<TrafficLightsTag>();
                if (tag.type == IntersectionType.TrafficLights && trafficLightsTag != null)
                {
                    if (!solvedIntersectionsInRange.Contains(go))
                    {
                        var active = trafficLightsTag.GetActiveLanes();
                        if (currentLane != null && !active.Contains(currentLane) && currentIntersectionEnterLane == currentLane)
                        {
                            if (printDetectedYields) { Debug.LogError("TRAFFIC_LIGHT_CONTROLLED: " + tag.gameObject.name); }

                            if (active.Count == 0)
                            {
                                var lane = new LaneTag();
                                lane.id = 0;
                                lane.name = "untitled";
                                lane.speedLimit = laneFallbackSpeedLimit;

                                var laneId = (uint)lane.id;
                                laneId += (uint)yieldLaneIdOffset;

                                KnowledgeGraph.add_lane_node(timestamp, laneId, new Lane { laneId = (int)laneId, laneName = String2Bytes.Pack(lane.name), laneSpeedLimit = lane.speedLimit }, graph);
                                KnowledgeGraph.add_yieldstocarin_relation(timestamp, laneId, new yieldsToCarIn { ifDirection = 0 }, laneNodeType, (uint)currentLane.id, laneNodeType, laneId, graph);

                                solvedIntersectionsInRange.Add(go);
                                activeYields.Add(lane);
                            }

                            foreach (var lane in active)
                            {
                                var laneId = (uint)lane.id;
                                laneId += (uint)yieldLaneIdOffset;

                                KnowledgeGraph.add_lane_node(timestamp, laneId, new Lane { laneId = (int)laneId, laneName = String2Bytes.Pack(lane.name), laneSpeedLimit = lane.speedLimit }, graph);
                                KnowledgeGraph.add_yieldstocarin_relation(timestamp, laneId, new yieldsToCarIn { ifDirection = 0 }, laneNodeType, (uint)currentLane.id, laneNodeType, laneId, graph);

                                solvedIntersectionsInRange.Add(go);
                                activeYields.Add(lane);
                            }
                        }
                    }
                }
            }
        }
    }
}