﻿using System.Collections.Generic;
using PathCreation;
using PathCreation.Utility;
using UnityEngine;

public class LaneTagCreator : PathSceneTool {
    [Header ("Lane settings")]
    public float laneWidth = 0.5f;
    [Range (0, 0.5f)]
    public float depth = 0.15f;
    public bool flattenSurface;

    [Header ("Material settings")]
    public Color topColor = new Color(1f, 0f, 0f);
    public Color bottomColor = new Color(0.5f, 0.5f, 0.5f);
    [Range(0, 1.0f)]
    public float alpha = 0.5f;
    public bool hide = false;

    [Header("Lane collider")]
    public float offset = -0.2f;
    public float height = 0.1f;

    Material topMaterial;
    Material bottomMaterial;

    [SerializeField, HideInInspector]
    GameObject meshHolder;

    MeshFilter meshFilter;
    MeshRenderer meshRenderer;
    Mesh mesh;

    Mesh colliderMesh;

    public void ForceUpdate()
    {
        PathUpdated();
    }

    protected override void PathUpdated () {
        if (pathCreator != null) {
            AssignMeshComponents ();
            AssignColliders();
            AssignMaterials ();
            CreateRoadMesh ();
        }
    }

    void CreateRoadMesh()
    {
        Vector3[] verts = new Vector3[path.NumPoints * 8];
        Vector2[] uvs = new Vector2[verts.Length];
        Vector3[] normals = new Vector3[verts.Length];

        int numTris = 2 * (path.NumPoints - 1) + ((path.isClosedLoop) ? 2 : 0);
        int[] roadTriangles = new int[numTris * 3];
        int[] underRoadTriangles = new int[numTris * 3];
        int[] sideOfRoadTriangles = new int[numTris * 2 * 3];

        int vertIndex = 0;
        int triIndex = 0;

        // Vertices for the top of the road are layed out:
        // 0  1
        // 8  9
        // and so on... So the triangle map 0,8,1 for example, defines a triangle from top left to bottom left to bottom right.
        int[] triangleMap = { 0, 8, 1, 1, 8, 9 };
        int[] sidesTriangleMap = { 4, 6, 14, 12, 4, 14, 5, 15, 7, 13, 15, 5 };

        bool usePathNormals = !(path.space == PathSpace.xyz && flattenSurface);

        for (int i = 0; i < path.NumPoints; i++)
        {
            Vector3 localUp = (usePathNormals) ? Vector3.Cross(path.GetTangent(i), path.GetNormal(i)) : path.up;
            Vector3 localRight = (usePathNormals) ? path.GetNormal(i) : Vector3.Cross(localUp, path.GetTangent(i));

            // Find position to left and right of current path vertex
            Vector3 vertSideA = path.GetPoint(i) - localRight * Mathf.Abs(laneWidth);
            Vector3 vertSideB = path.GetPoint(i) + localRight * Mathf.Abs(laneWidth);

            // Add top of road vertices
            verts[vertIndex + 0] = vertSideA;
            verts[vertIndex + 1] = vertSideB;
            // Add bottom of road vertices
            verts[vertIndex + 2] = vertSideA - localUp * depth;
            verts[vertIndex + 3] = vertSideB - localUp * depth;

            // Duplicate vertices to get flat shading for sides of road
            verts[vertIndex + 4] = verts[vertIndex + 0];
            verts[vertIndex + 5] = verts[vertIndex + 1];
            verts[vertIndex + 6] = verts[vertIndex + 2];
            verts[vertIndex + 7] = verts[vertIndex + 3];

            // Set uv on y axis to path time (0 at start of path, up to 1 at end of path)
            uvs[vertIndex + 0] = new Vector2(0, path.times[i]);
            uvs[vertIndex + 1] = new Vector2(1, path.times[i]);

            // Top of road normals
            normals[vertIndex + 0] = localUp;
            normals[vertIndex + 1] = localUp;
            // Bottom of road normals
            normals[vertIndex + 2] = -localUp;
            normals[vertIndex + 3] = -localUp;
            // Sides of road normals
            normals[vertIndex + 4] = -localRight;
            normals[vertIndex + 5] = localRight;
            normals[vertIndex + 6] = -localRight;
            normals[vertIndex + 7] = localRight;

            // Set triangle indices
            if (i < path.NumPoints - 1 || path.isClosedLoop)
            {
                for (int j = 0; j < triangleMap.Length; j++)
                {
                    roadTriangles[triIndex + j] = (vertIndex + triangleMap[j]) % verts.Length;
                    // reverse triangle map for under road so that triangles wind the other way and are visible from underneath
                    underRoadTriangles[triIndex + j] = (vertIndex + triangleMap[triangleMap.Length - 1 - j] + 2) % verts.Length;
                }
                for (int j = 0; j < sidesTriangleMap.Length; j++)
                {
                    sideOfRoadTriangles[triIndex * 2 + j] = (vertIndex + sidesTriangleMap[j]) % verts.Length;
                }

            }

            vertIndex += 8;
            triIndex += 6;
        }

        mesh.Clear();
        mesh.vertices = verts;
        mesh.uv = uvs;
        mesh.normals = normals;
        mesh.subMeshCount = 3;
        mesh.SetTriangles(roadTriangles, 0);
        mesh.SetTriangles(underRoadTriangles, 1);
        mesh.SetTriangles(sideOfRoadTriangles, 2);
        mesh.RecalculateBounds();


        // Generate one more mesh for the collider
        int vertIndex2 = 0;
        for (int i = 0; i < path.NumPoints; i++)
        {
            Vector3 localUp = (usePathNormals) ? Vector3.Cross(path.GetTangent(i), path.GetNormal(i)) : path.up;
            Vector3 localRight = (usePathNormals) ? path.GetNormal(i) : Vector3.Cross(localUp, path.GetTangent(i));

            Vector3 vertSideA = path.GetPoint(i) - localRight * Mathf.Abs(laneWidth);
            Vector3 vertSideB = path.GetPoint(i) + localRight * Mathf.Abs(laneWidth);

            verts[vertIndex2 + 2] = vertSideA - localUp * height;
            verts[vertIndex2 + 3] = vertSideB - localUp * height;

            vertIndex2 += 8;
        }

        // Offset mesh for collider
        for (int i = 0; i < verts.Length; i++)
        {
            verts[i].y += offset;
        }

        colliderMesh.Clear();
        colliderMesh.vertices = verts;
        colliderMesh.uv = uvs;
        colliderMesh.normals = normals;
        colliderMesh.subMeshCount = 3;
        colliderMesh.SetTriangles(roadTriangles, 0);
        colliderMesh.SetTriangles(underRoadTriangles, 1);
        colliderMesh.SetTriangles(sideOfRoadTriangles, 2);
        colliderMesh.RecalculateBounds();
    }

    // Add MeshRenderer and MeshFilter components to this gameobject if not already attached
    void AssignMeshComponents () {

        if (meshHolder == null) {
            meshHolder = new GameObject ("LaneTagMesh");
        }

        meshHolder.transform.parent = transform;
        meshHolder.transform.rotation = Quaternion.identity;
        meshHolder.transform.position = Vector3.zero;
        meshHolder.transform.localScale = Vector3.one;

        // Ensure mesh renderer and filter components are assigned
        if (!meshHolder.gameObject.GetComponent<MeshFilter> ()) {
            meshHolder.gameObject.AddComponent<MeshFilter> ();
        }
        if (!meshHolder.GetComponent<MeshRenderer> ()) {
            meshHolder.gameObject.AddComponent<MeshRenderer> ();
        }

        meshRenderer = meshHolder.GetComponent<MeshRenderer> ();
        meshFilter = meshHolder.GetComponent<MeshFilter> ();
        if (mesh == null) {
            mesh = new Mesh ();
        }
        meshFilter.sharedMesh = mesh;
    }

    // Add MeshRenderer and MeshFilter components to this gameobject
    void AssignColliders()
    {
        if (colliderMesh == null)
        {
            colliderMesh = new Mesh();
        }

        if (meshHolder != null)
        {
            var collider = meshHolder.GetComponent<MeshCollider>();
            if (collider == null)
            {
                collider = meshHolder.AddComponent<MeshCollider>();
            }
            else
            {
                collider.sharedMesh = null;
                collider.sharedMesh = colliderMesh;
            }
        }
    }

    void AssignMaterials () {

        if (topMaterial == null)
        {
            topMaterial = new Material(Shader.Find("Standard"));
            topMaterial.ToFadeMode();
        }

        if (bottomMaterial == null)
        {
            bottomMaterial = new Material(Shader.Find("Standard"));
            bottomMaterial.ToFadeMode();
        }

        var alphaValue = alpha;
        if (hide) {
            alphaValue = 0;
        }

        // Set colors and transparency
        topMaterial.color = new Color(topColor.r, topColor.g, topColor.b, alphaValue);
        bottomMaterial.color = new Color(bottomColor.r, bottomColor.g, bottomColor.b, alphaValue);

        // Set materials
        meshRenderer.sharedMaterials = new Material[] { topMaterial, bottomMaterial, bottomMaterial };
    }
}