﻿using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public enum IntersectionType { Uncontrolled, Controlled, TrafficLights, Unspecified }
public enum IntersectionSize { _3Way, _4Way, Unspecified }
public enum TurnDirection { Left, Straight, Right }

public class IntersectionTag : MonoBehaviour
{
    [UniqueIdentifier]
    public int id = -1;
    public string name = "Untitled";
    public IntersectionType type = IntersectionType.Unspecified;
    public IntersectionSize size = IntersectionSize.Unspecified;

    public ObjectDetectionTrigger carTrigger;

    // Yielding rules (aka default)
    [Serializable]
    public struct YieldRule
    {
        public LaneTag fromLane;
        public LaneTag ifLeftYieldsTo;
        public LaneTag ifStraightYieldsTo;
        public LaneTag ifRightYieldsTo;
    }
    [Header("DEFAULT YIELDING RULES")]
    public YieldRule[] defaultYieldingRules;


    [Header("CARS AT INTERSECTION")]
    public float trackUpdateFrequencyMs = 200.0f;
    public LayerMask laneObjectsLayerMask = 1 << 24;
    public Vector3 detectionRayOffset = new Vector3(0, 2.0f, 0);
    public float detectionRayLegth = 10.0f;

    [ReadOnly]
    public int carLayer = 17;
    [ReadOnly]
    public string carTag = "TrafficCar";

    [ReadOnly]
    public List<GameObject> carsAtIntersection = new List<GameObject>();
    public List<LaneTag> occupiedLanes = new List<LaneTag>();


    public List<(TurnDirection, LaneTag)> DefaultHasToYieldTo(LaneTag laneTag)
    {
        var list = new List<(TurnDirection, LaneTag)>();
        foreach (var rule in defaultYieldingRules)
        {
            if (rule.fromLane == laneTag)
            {
                // If turning left from the given lane
                if (rule.ifLeftYieldsTo != null && occupiedLanes.Contains(rule.ifLeftYieldsTo))
                {
                    list.Add((TurnDirection.Left, rule.ifLeftYieldsTo));
                }
                // If going straight from the given lane
                else if (rule.ifStraightYieldsTo != null && occupiedLanes.Contains(rule.ifStraightYieldsTo))
                {
                    list.Add((TurnDirection.Straight, rule.ifStraightYieldsTo));
                }
                // If going straight from the given lane
                else if (rule.ifRightYieldsTo != null && occupiedLanes.Contains(rule.ifRightYieldsTo))
                {
                    list.Add((TurnDirection.Right, rule.ifRightYieldsTo));
                }

                return list;
            }
        }

        return list;
    }


    // Run detection of the cars in the crossroad (for yielding)
    private void Start()
    {
        // Assign callbacks
        carTrigger.OnObjectEnterEvent += CarTriggerEnter;
        carTrigger.OnObjectExitEvent += CarTriggerExit;

        StartCoroutine(CarLaneDetection());
    }

    private void CarTriggerEnter(GameObject go)
    {
        if (go.layer == carLayer &&
            go.tag == carTag)
        {
            carsAtIntersection.Add(go);
        }
    }

    private void CarTriggerExit(GameObject go)
    {
        if (go.layer == carLayer &&
            go.tag == carTag)
        {
            carsAtIntersection.Remove(go);
        }
    }

    private IEnumerator CarLaneDetection()
    {
        while (true)
        {
            yield return new WaitForSeconds(trackUpdateFrequencyMs / 1000f);

            var newOccupiedLanes = new List<LaneTag>();
            foreach (var car in carsAtIntersection)
            {
                // Raycast to detect lane
                var rayOrigin = car.transform.position + detectionRayOffset;
                var hits = Physics.RaycastAll(rayOrigin, Vector3.down, detectionRayLegth, laneObjectsLayerMask, QueryTriggerInteraction.Ignore);

                if (hits.Length == 1)
                {
                    // Fetch the lane
                    var parent = hits[0].transform.parent;
                    var tag = parent.GetComponent<LaneTag>();
                    newOccupiedLanes.Add(tag);
                }
            }
            occupiedLanes = newOccupiedLanes;
        }
    }
}